
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! Initialisation of the adgali library.

/*! \file
 * \ingroup video
 * This file initialise all the adgali
 * internals.
 */

#include "config.h"
#include "agl_private.h"


//! Initialisation of the adgali library.

/*! this function initialise all the adgali
 * internals.
 * \return nothing
 */

int
agl_init(agl_handle * phandle)
{
	agl_container *handle;

	agl_rvassert(phandle != NULL, -1);

	if (agl_initparachute() < 0)
		return -1;

	*phandle = (agl_handle) handle = (agl_container *) agl_malloc(sizeof(agl_container));
	if (*phandle == NULL)
	{
		agl_memerr();
		return -1;
	}

	if (agl_sublayerinit(handle) < 0)
	{
		agl_error("sublayer initialization failed");
		agl_free(handle);
		return -1;
	}
	if (agl_modinit(handle) < 0)
	{
		agl_error("module initialization failed");
		agl_free(handle);
		return -1;
	}
	if (agl_cacheinit(handle) < 0)
	{
		agl_error("cache initialization failed");
		agl_free(handle);
		return -1;
	}
	return 1;
}

//! Sets the video mode for the adgali display.

/*! This function sets the video modes for adgali.
 * This function could be internal for the future.
 * \return nothing
 */

void
agl_setvideo(void)
{
}

//! gets out of adgali and free all the resources.

/*! Free all the resources and gets out.
 * Could be automatically called in the future when
 * we exiting the program
 * \return nothing
 */

int
agl_exit(agl_handle handle)
{
	int ret;

	ret = 1;

	agl_rvassert(handle != NULL, -1);

	if (agl_cacheexit(handle) < 0)
	{
		agl_error("cache initialization failed");
		agl_free(handle);
		return -1;
	}
	if (agl_modexit((agl_container *) handle) < 0)
	{
		agl_error("module exit failed");
		ret = -1;
	}
	if (agl_sublayerexit(handle) < 0)
	{
		agl_error("sublayer exit failed");
		ret = -1;
	}
	agl_free(handle);
	if (agl_memgarbage(handle) < 0)
	{
		agl_error("memory garbage failed");
		ret = -1;
	}

	agl_exitparachute();

	return ret;
}
