
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#include <string.h>
#include "config.h"
#include "agl_private.h"

int
agl_defnsubrecur(agl_elemtype * pdef, int list)
{
	agl_elemtype *pcur;

	agl_rvassert(pdef != NULL, -1);

	pcur = pdef;
	while (pcur != NULL)
	{
		pdef = pcur;
		switch (pdef->vtype & AGL_STYPEALL)
		{
		case AGL_STYPESTR:
			if (!(pdef->vtype & AGL_STYPESTATIC))
				agl_free(pdef->vu.str);
			break;
		case AGL_STYPEOBJREF:
			if (!(pdef->vtype & AGL_STYPESTATIC))
			{
				agl_free(pdef->vu.ref.name);
				agl_free(pdef->vu.ref.mod);
			}
			break;
		case AGL_STYPEDBL:
		case AGL_STYPEINT:
		case AGL_STYPEUSER:
		case AGL_STYPENONE:
			break;
		case AGL_STYPELIST:
			if (agl_defnsubrecur(pdef->vu.list, 1) <= 0)
				return -1;
		default:
			agl_error2("unknown type %d", pdef->vtype & AGL_STYPEALL);
			return -1;
		}
		switch (pdef->itype & AGL_STYPEALL)
		{
		case AGL_STYPESTR:
			if (!(pdef->vtype & AGL_STYPESTATIC))
				agl_free(pdef->iu.str);
			break;
		case AGL_STYPEOBJREF:
			if (!(pdef->vtype & AGL_STYPESTATIC))
			{
				agl_free(pdef->iu.ref.name);
				agl_free(pdef->iu.ref.mod);
			}
			break;
		case AGL_STYPEDBL:
		case AGL_STYPEINT:
		case AGL_STYPEUSER:
		case AGL_STYPENONE:
			break;
		case AGL_STYPELIST:
			if (agl_defnsubrecur(pdef->iu.list, 1) <= 0)
				return -1;
		default:
			agl_error2("unknown type %d", pdef->itype & AGL_STYPEALL);
			return -1;
		}
		if (list)
		{
			pcur = pdef->pnext;
			agl_free(pdef);
		}
	}

	return 1;
}

int
agl_defnvcmp(agl_elemtype * pdef, agl_elemtype * pv)
{
	agl_elemtype *pel, *pelv, *peli;
	int ret;

	agl_rvassert(pdef != NULL, -1);
	agl_rvassert(pv != NULL, -1);

	agl_rvassert((pv->vtype & AGL_STYPEALL) == AGL_STYPENONE, 0);

	pel = pdef;

	if ((pel->vtype & AGL_STYPEALL) == (pv->vtype & AGL_STYPEALL))
	{
		switch (pv->vtype & AGL_STYPEALL)
		{
		case AGL_STYPESTR:
			if (strcmp(pel->vu.str, pv->vu.str) != 0)
				return 0;
			break;
		case AGL_STYPEOBJREF:
			if (strcmp(pel->vu.ref.name, pv->vu.ref.name) != 0)
				return 0;
			if (strcmp(pel->vu.ref.mod, pv->vu.ref.mod) != 0)
				return 0;
			break;
		case AGL_STYPEDBL:
			if (pel->vu.dbl != pv->vu.dbl)
				return 0;
			break;
		case AGL_STYPEINT:
			if (pel->vu.num != pv->vu.num)
				return 0;
			break;
		case AGL_STYPELIST:
			pelv = pv->iu.list;

			while (pelv != NULL)
			{
				ret = agl_defnindex(pel->iu.list, pelv, &peli);
				if (ret < 0)
					return -1;
				if (ret == 1)
				{
					ret = agl_defnvcmp(pelv, peli);

					if (ret <= 0)
						return ret;
				}
				else
					return 0;
			}

			break;
		case AGL_STYPEUSER:
			if (pel->vu.user != pv->vu.user)
				return 0;
			break;
		default:
			agl_error2("unknown type %d", pv->vtype & AGL_STYPEALL);
			return -1;
		}

	}

	// don't test index, they are the same (if this is a list elem, agl_defnindex find this
	// element with the good index

	return 1;
}

int
agl_defnicmp(agl_elemtype * pdef, agl_elemtype * pv)
{
	agl_elemtype *pel;
	int found;

	agl_rvassert(pdef != NULL, -1);
	agl_rvassert(pv != NULL, -1);

	agl_rvassert((pv->vtype & AGL_STYPEALL) == AGL_STYPENONE, 0);

	pel = pdef;

	while (pel != NULL)
	{
		if ((pel->itype & AGL_STYPEALL) == (pv->itype & AGL_STYPEALL))
		{
			switch (pv->itype & AGL_STYPEALL)
			{
			case AGL_STYPESTR:
				if (strcmp(pel->iu.str, pv->iu.str) != 0)
					return 0;
				break;
			case AGL_STYPEOBJREF:
				if (strcmp(pel->vu.ref.name, pv->vu.ref.name) != 0)
					return 0;
				if (strcmp(pel->vu.ref.mod, pv->vu.ref.mod) != 0)
					return 0;
				break;
			case AGL_STYPEDBL:
				if (pel->iu.dbl != pv->iu.dbl)
					return 0;
				break;
			case AGL_STYPEINT:
				if (pel->iu.num != pv->iu.num)
					return 0;
				break;
			case AGL_STYPELIST:
				found = agl_defnicmp(pel->iu.list, pv->iu.list);
				if (found != 1)
					return found;
				break;
			case AGL_STYPEUSER:
				if (pel->iu.user != pv->iu.user)
					return 0;
				break;
			default:
				agl_error2("unknown type %d", pv->itype & AGL_STYPEALL);
				return -1;
			}

		}
		pel = pel->pnext;
	}
	return 1;
}


int
agl_defnindex(agl_elemtype * pdef, agl_elemtype * pv, agl_elemtype ** presp)
{
	agl_elemtype *pel;
	int idx, found;

	agl_rvassert(pdef != NULL, -1);
	agl_rvassert(pv != NULL, -1);
	agl_rvassert(presp != NULL, -1);

	agl_rvassert((pv->vtype & AGL_STYPEALL) == AGL_STYPENONE, 0);

	pel = pdef;

	idx = 0;

	found = 0;

	while (pel != NULL)
	{
		if (((pel->itype & AGL_STYPEALL) == AGL_STYPENONE) && ((pv->itype & AGL_STYPEALL) == AGL_STYPEINT))
		{
			if (pv->iu.num == idx)
				found = 1;
		}
		else if ((pel->itype & AGL_STYPEALL) == (pv->itype & AGL_STYPEALL))
		{
			switch (pv->itype & AGL_STYPEALL)
			{
			case AGL_STYPESTR:
				if (strcmp(pel->iu.str, pv->iu.str) == 0)
					found = 1;
				break;
			case AGL_STYPEOBJREF:
				if (strcmp(pel->vu.ref.name, pv->vu.ref.name) != 0)
					return 0;
				if (strcmp(pel->vu.ref.mod, pv->vu.ref.mod) != 0)
					return 0;
				break;
			case AGL_STYPEDBL:
				if (pel->iu.dbl == pv->iu.dbl)
					found = 1;
				break;
			case AGL_STYPEINT:
				if (pel->iu.num == pv->iu.num)
					found = 1;
				break;
			case AGL_STYPELIST:
				found = agl_defnicmp(pel->iu.list, pv->iu.list);
				if (found < 0)
					return -1;
				break;
			case AGL_STYPEUSER:
				if (pel->iu.user == pv->iu.user)
					found = 1;
				break;
			default:
				agl_error2("unknown type %d", pv->itype & AGL_STYPEALL);
				return -1;
			}

		}
		if (found)
		{
			*presp = pel;
			return 1;
		}
		pel = pel->pnext;
	}
	return 0;
}

int
agl_defnaddrecur(agl_elemtype ** pdef, agl_elemtype * pv, int list)
{
	agl_elemtype *pi, *pf, *pp;
	int idx, ret;
	unsigned short type;

	agl_rvassert(pdef != NULL, -1);
	agl_rvassert(pv != NULL, -1);

	if (pv == NULL)
		return 1;

	if (*pdef == NULL)
	{
		*pdef = (agl_elemtype *) agl_malloc(sizeof(agl_elemtype));
		if (*pdef == NULL)
		{
			agl_memerr();
			return -1;
		}
		(*pdef)->vtype = AGL_STYPENONE;
		(*pdef)->itype = AGL_STYPENONE;
		(*pdef)->pnext = NULL;
	}

	switch (pv->vtype & AGL_STYPEALL)
	{
	case AGL_STYPENONE:
		if (agl_defnsubrecur(*pdef, 0) < 0)
			return -1;
		break;
	case AGL_STYPESTR:
		if (agl_defnsubrecur(*pdef, 0) < 0)
			return -1;
		(*pdef)->vu.str = (char *) agl_malloc(strlen(pv->vu.str) + 1);
		if ((*pdef)->vu.str == NULL)
		{
			agl_memerr();
			return -1;
		}
		strcpy((*pdef)->vu.str, pv->vu.str);
		break;
	case AGL_STYPEOBJREF:
		if (agl_defnsubrecur(*pdef, 0) < 0)
			return -1;
		(*pdef)->vu.ref.name = (char *) agl_malloc(strlen(pv->vu.ref.name) + 1);
		if ((*pdef)->vu.ref.name == NULL)
		{
			agl_memerr();
			return -1;
		}
		strcpy((*pdef)->vu.ref.name, pv->vu.ref.name);

		(*pdef)->vu.ref.mod = (char *) agl_malloc(strlen(pv->vu.ref.mod) + 1);
		if ((*pdef)->vu.ref.mod == NULL)
		{
			agl_memerr();
			agl_free((*pdef)->vu.ref.name);
			return -1;
		}
		strcpy((*pdef)->vu.ref.mod, pv->vu.ref.mod);

	case AGL_STYPEDBL:
		if (agl_defnsubrecur(*pdef, 0) < 0)
			return -1;
		(*pdef)->vu.dbl = pv->vu.dbl;
		break;
	case AGL_STYPEINT:
		if (agl_defnsubrecur(*pdef, 0) < 0)
			return -1;
		(*pdef)->vu.num = pv->vu.num;
		break;
	case AGL_STYPELIST:
		if (((*pdef)->vtype & AGL_STYPEALL) == AGL_STYPELIST)
		{
			pf = pv->vu.list;
			idx = 0;
			pp = NULL;
			pi = NULL;
			while (pf != NULL)
			{
				ret = agl_defnindex((*pdef)->vu.list, pf, &pp);
				if (ret < 0)
					return -1;
				if (ret == 1)
				{
					if (agl_defnaddrecur(&pp, pf, idx) <= 0)
					{
						return -1;
					}
				}
				else
				{
					pi = NULL;
					if (agl_defnaddrecur(&pi, pf, idx) <= 0)
					{
						return -1;
					}
					pp = (*pdef)->vu.list;
					(*pdef)->vu.list = pi;
					pi->pnext = pp;
				}
				idx++;
				pf = pf->pnext;
			}
		}
		else
		{
			if (agl_defnsubrecur(*pdef, 0) < 0)
				return -1;
			pf = pv->vu.list;
			idx = 0;
			pp = NULL;
			pi = NULL;
			(*pdef)->vu.list = NULL;
			while (pf != NULL)
			{
				if (agl_defnaddrecur(&pi, pf, idx) <= 0)
				{
					return -1;
				}
				if ((*pdef)->vu.list == NULL)
					(*pdef)->vu.list = pi;
				else
					pp->pnext = pi;
				pp = pi;
				pi = NULL;
				idx++;
				pf = pf->pnext;
			}
		}
		break;
	case AGL_STYPEUSER:
		if (agl_defnsubrecur(*pdef, 0) < 0)
			return -1;
		(*pdef)->vu.user = pv->vu.user;
		break;
	default:
		agl_error2("unknown type %d", pv->vtype & AGL_STYPEALL);
		return -1;
	}
	(*pdef)->vtype &= ~AGL_STYPEALL;
	(*pdef)->vtype |= pv->vtype & AGL_STYPEALL;

	if (list)
	{
		type = pv->itype;
		switch (pv->itype & AGL_STYPEALL)
		{
		case AGL_STYPENONE:
			if (agl_defnsubrecur(*pdef, 0) < 0)
				return -1;
			if ((pv->vtype & AGL_STYPEALL) != AGL_STYPENONE)
			{
				(*pdef)->iu.num = list;
				type = AGL_STYPEINT;
			}
			break;
		case AGL_STYPESTR:
			if (agl_defnsubrecur(*pdef, 0) < 0)
				return -1;
			(*pdef)->iu.str = (char *) agl_malloc(strlen(pv->iu.str) + 1);
			if ((*pdef)->iu.str == NULL)
			{
				agl_memerr();
				return -1;
			}
			strcpy((*pdef)->iu.str, pv->iu.str);
			break;
		case AGL_STYPEOBJREF:
			if (agl_defnsubrecur(*pdef, 0) < 0)
				return -1;
			(*pdef)->iu.ref.name = (char *) agl_malloc(strlen(pv->iu.ref.name) + 1);
			if ((*pdef)->iu.ref.name == NULL)
			{
				agl_memerr();
				return -1;
			}
			strcpy((*pdef)->iu.ref.name, pv->iu.ref.name);

			(*pdef)->iu.ref.mod = (char *) agl_malloc(strlen(pv->iu.ref.mod) + 1);
			if ((*pdef)->iu.ref.mod == NULL)
			{
				agl_memerr();
				agl_free((*pdef)->iu.ref.name);
				return -1;
			}
			strcpy((*pdef)->iu.ref.mod, pv->iu.ref.mod);

		case AGL_STYPEDBL:
			if (agl_defnsubrecur(*pdef, 0) < 0)
				return -1;
			(*pdef)->iu.dbl = pv->iu.dbl;
			break;
		case AGL_STYPEINT:
			if (agl_defnsubrecur(*pdef, 0) < 0)
				return -1;
			(*pdef)->iu.num = pv->iu.num;
			break;
		case AGL_STYPELIST:
			if (((*pdef)->itype & AGL_STYPEALL) == AGL_STYPELIST)
			{
				pf = pv->iu.list;
				idx = 0;
				pp = NULL;
				pi = NULL;
				while (pf != NULL)
				{
					ret = agl_defnindex((*pdef)->iu.list, pf, &pp);
					if (ret < 0)
						return -1;
					if (ret == 1)
					{
						if (agl_defnaddrecur(&pp, pf, idx) <= 0)
						{
							return -1;
						}
					}
					else
					{
						pi = NULL;
						if (agl_defnaddrecur(&pi, pf, idx) <= 0)
						{
							return -1;
						}
						pp = (*pdef)->iu.list;
						(*pdef)->iu.list = pi;
						pi->pnext = pp;
					}
					idx++;
					pf = pf->pnext;
				}
			}
			else
			{
				if (agl_defnsubrecur(*pdef, 0) < 0)
					return -1;
				pf = pv->iu.list;
				idx = 0;
				pp = NULL;
				pi = NULL;
				(*pdef)->iu.list = NULL;
				while (pf != NULL)
				{
					if (agl_defnaddrecur(&pi, pf, idx) <= 0)
					{
						return -1;
					}
					if ((*pdef)->iu.list == NULL)
						(*pdef)->iu.list = pi;
					else
						pp->pnext = pi;
					pp = pi;
					pi = NULL;
					idx++;
					pf = pf->pnext;
				}
			}
			break;
		case AGL_STYPEUSER:
			if (agl_defnsubrecur(*pdef, 0) < 0)
				return -1;
			(*pdef)->iu.user = pv->iu.user;
			break;
		default:
			agl_error2("unknown type %d", pv->itype & AGL_STYPEALL);
			return -1;
		}
		(*pdef)->itype &= ~AGL_STYPEALL;
		(*pdef)->itype |= type & AGL_STYPEALL;
	}
	else
	{
		(*pdef)->itype &= ~AGL_STYPEALL;
		(*pdef)->itype |= AGL_STYPENONE & AGL_STYPEALL;
	}
	return 1;
}

int
agl_elemexp2u(agl_elemexpand * pexp, agl_elemtype * pval)
{
	agl_rvassert(pexp != NULL, -1);

	agl_rvassert(pval != NULL, -1);

	switch (pexp->vtype & AGL_STYPEALL)
	{
	case AGL_STYPENONE:
	case AGL_STYPEDBL:
		pval->vu.dbl = pexp->ve.dbl;
		break;
	case AGL_STYPEINT:
		pval->vu.num = pexp->ve.num;
		break;
	case AGL_STYPESTR:
		pval->vu.str = pexp->ve.str;
		break;
	case AGL_STYPEOBJREF:
		pval->vu.ref.name = pexp->ve.ref.name;
		pval->vu.ref.mod = pexp->ve.ref.mod;
		break;
	case AGL_STYPELIST:
		pval->vu.list = pexp->ve.list;
		break;
	case AGL_STYPEUSER:
		pval->vu.user = pexp->ve.user;
		break;
	}
	pval->vtype = (pexp->vtype & AGL_STYPEALL) | AGL_STYPESTATIC;

	switch (pexp->itype & AGL_STYPEALL)
	{
	case AGL_STYPENONE:
	case AGL_STYPEDBL:
		pval->iu.dbl = pexp->ie.dbl;
		break;
	case AGL_STYPEINT:
		pval->iu.num = pexp->ie.num;
		break;
	case AGL_STYPESTR:
		pval->iu.str = pexp->ie.str;
		break;
	case AGL_STYPEOBJREF:
		pval->iu.ref.name = pexp->ie.ref.name;
		pval->iu.ref.mod = pexp->ie.ref.mod;
		break;
	case AGL_STYPELIST:
		pval->iu.list = pexp->ie.list;
		break;
	case AGL_STYPEUSER:
		pval->iu.user = pexp->ie.user;
		break;
	}
	pval->itype = pexp->itype;
	return 1;
}

int
agl_freeval(agl_elemtype * pval, int destroy)
{
	agl_elemtype *plist, *pnext;
	int ret;

	ret = 0;
	if (!(pval->vtype & AGL_STYPESTATIC))
		switch (pval->vtype & AGL_STYPEALL)
		{
		case AGL_STYPENONE:
		case AGL_STYPEDBL:
		case AGL_STYPEINT:
			break;
		case AGL_STYPESTR:
			agl_free(pval->vu.str);
			break;
		case AGL_STYPEOBJREF:
			agl_free(pval->vu.ref.name);
			agl_free(pval->vu.ref.mod);
			break;
		case AGL_STYPELIST:
			for (plist = pval->vu.list; plist; plist = pnext)
			{
				pnext = plist->pnext;
				if (agl_freeval(plist, destroy) < 0)
					ret = -1;
			}
			break;
		case AGL_STYPEUSER:
			if (destroy)
			{
				if (agl_objdelete(pval->vu.user) < 0)
					ret = -1;
			}
			break;
		}
	if (!(pval->itype & AGL_STYPESTATIC))
		switch (pval->itype & AGL_STYPEALL)
		{
		case AGL_STYPENONE:
		case AGL_STYPEDBL:
		case AGL_STYPEINT:
			break;
		case AGL_STYPESTR:
			agl_free(pval->iu.str);
			break;
		case AGL_STYPEOBJREF:
			break;
		case AGL_STYPELIST:
			for (plist = pval->iu.list; plist; plist = pnext)
			{
				pnext = plist->pnext;
				if (agl_freeval(plist, destroy) < 0)
					ret = -1;
			}
			break;
		case AGL_STYPEUSER:
			if (destroy)
			{
				if (agl_objdelete(pval->iu.user) < 0)
					ret = -1;
			}
			break;
		}

	pval->itype = AGL_STYPENONE;
	if (pval->vtype & AGL_STYPEALLOC)
		agl_free(pval);
	else
		pval->vtype = AGL_STYPENONE;
	return ret;
}
