
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* got from SDL. Thanks to Sam Lantinga */

#include <stdio.h>
#if defined(USE_DLOPEN)
# include <dlfcn.h>
#else
#if defined(WIN32)
# include <windows.h>
#else
#if defined(__BEOS__)
# include <be/kernel/image.h>
#else
#if defined(macintosh)
# include <string.h>
# include <Strings.h>
# include <CodeFragments.h>
# include <Errors.h>
#endif
#endif
#endif
#endif

#include "agl_private.h"

void *
agl_loadobj(char *libfile)
{
	void *handle = NULL;
	char *loaderror = "agl_loadobj not implemented";

	agl_rvassert(libfile != NULL, NULL);
#if defined(USE_DLOPEN)
	handle = dlopen(file, RTLD_NOW);
	loaderror = (char *) dlerror();
#else
#if defined(WIN32)
	char errbuf[512];

	handle = (void *) LoadLibrary(libfile);

	/* Generate an error message if all loads failed */
	if (handle == NULL)
	{
		FormatMessage((FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM), NULL, GetLastError(),
			      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), errbuf, SDL_TABLESIZE(errbuf), NULL);
		loaderror = errbuf;
	}
#else
#if defined(__BEOS__)
	image_id library_id;

	library_id = load_add_on(libfile);
	if (library_id == B_ERROR)
	{
		loaderror = "BeOS error";
	}
	else
	{
		handle = (void *) (library_id);
	}
#else
#if defined(macintosh)
	CFragConnectionID library_id;
	Ptr mainAddr;
	Str255 errName;
	OSErr error;
	char plibfile[512];

	strncpy(plibfile, libfile, SDL_TABLESIZE(plibfile));
	plibfile[SDL_TABLESIZE(plibfile) - 1] = '\0';
	error = GetSharedLibrary(C2PStr(plibfile), kCompiledCFragArch, kLoadCFrag, &library_id, &mainAddr, errName);
	switch (error)
	{
	case noErr:
		break;
	case cfragNoLibraryErr:
		loaderror = "Library not found";
		break;
	case cfragUnresolvedErr:
		loaderror = "Unabled to resolve symbols";
		break;
	case cfragNoPrivateMemErr:
	case cfragNoClientMemErr:
		loaderror = "Out of memory";
		break;
	default:
		loaderror = "Unknown Code Fragment Manager error";
		break;
	}
	if (loaderror == NULL)
	{
		handle = (void *) (library_id);
	}
#endif
#endif
#endif
#endif

	if (handle == NULL)
	{
		agl_error3("Failed loading %s: %s", libfile, loaderror);
	}
	return (handle);
}

void *
agl_loadfunc(void *handle, const char *name)
{
	void *symbol = NULL;
	const char *loaderror = "agl_loadfunc not implemented";

	agl_rvassert(handle != NULL, NULL);
#if defined(USE_DLOPEN)
	symbol = dlsym(handle, name);
	if (symbol == NULL)
	{
		loaderror = (char *) dlerror();
	}
#else
#if defined(WIN32)
	char errbuf[512];

	symbol = (void *) GetProcAddress((HMODULE) handle, name);
	if (symbol == NULL)
	{
		FormatMessage((FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM), NULL, GetLastError(),
			      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), errbuf, SDL_TABLESIZE(errbuf), NULL);
		loaderror = errbuf;
	}
#else
#if defined(__BEOS__)
	image_id library_id = (image_id) handle;

	if (get_image_symbol(library_id, name, B_SYMBOL_TYPE_TEXT, &symbol) != B_NO_ERROR)
	{
		loaderror = "Symbol not found";
	}
#else
#if defined(macintosh)
	CFragSymbolClass class;
	CFragConnectionID library_id = (CFragConnectionID) handle;
	char pname[512];

	strncpy(pname, name, SDL_TABLESIZE(pname));
	pname[SDL_TABLESIZE(pname) - 1] = '\0';
	if (FindSymbol(library_id, C2PStr(pname), (char **) &symbol, &class) != noErr)
	{
		loaderror = "Symbol not found";
	}
#endif
#endif
#endif
#endif

	if (symbol == NULL)
	{
		SDL_SetError("Failed loading %s: %s", name, loaderror);
	}
	return (symbol);
}

void
agl_unloadobj(void *handle)
{

	agl_rassert(handle != NULL);

#if defined(USE_DLOPEN)
	dlclose(handle);
#else
#if defined(WIN32)
	FreeLibrary((HMODULE) handle);
#else
#if defined(__BEOS__)
	image_id library_id;

	library_id = (image_id) handle;
	unload_add_on(library_id);
#else
#if defined(macintosh)
	CFragConnectionID library_id = (CFragConnectionID) handle;

	CloseConnection(library_id);
#endif
#endif
#endif
#endif
}
