
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* implements the intersection or union of a source zone and a destination zone
 */

#include "config.h"
#include "agl_private.h"

agl_rect *
agl_rectcreate(int x, int y, int width, int height)
{
	agl_rect *new_rect;

	new_rect = agl_malloc(sizeof(agl_rect));
	if (new_rect == NULL)
	{
		agl_memerr();
		return NULL;
	}
	new_rect->x = x;
	new_rect->y = y;
	new_rect->width = width;
	new_rect->height = height;
	return new_rect;
}

void
agl_rectfree(agl_rect * rect)
{

	agl_rassert(rect != NULL);

	agl_free(rect);
}

agl_rect *
agl_rectcopy(agl_rect * rect)
{
	agl_rvassert(rect != NULL, NULL);

	return agl_rectcreate(rect->x, rect->y, rect->width, rect->height);
}

int
agl_rectinter(agl_rect * dest, agl_rect * src)
{
	agl_rect *pa, *pb;

	agl_rvassert(dest != NULL, -1);
	agl_rvassert(src != NULL, -1);

	if (dest->x < src->x)
	{
		pa = dest;
		pb = src;
	}
	else
	{
		pa = src;
		pb = dest;
	}
	dest->width = min(pa->x + pa->width - pb->x, pb->width);
	dest->x = pb->x;
	if (dest->width <= 0)
	{
		dest->width = 0;
		return 0;
	}
	if (dest->y < src->y)
	{
		pa = dest;
		pb = src;
	}
	else
	{
		pa = src;
		pb = dest;
	}
	dest->height = min(pa->y + pa->height - pb->y, pb->height);
	dest->y = pb->y;
	if (dest->height <= 0)
	{
		dest->height = 0;
		return 0;
	}
	return 1;
}

agl_rectlist *
agl_rectlistcreate(int x, int y, int width, int height)
{
	agl_rectlist *new_rect;

	new_rect = agl_malloc(sizeof(agl_rectlist));
	if (new_rect == NULL)
	{
		agl_memerr();
		return NULL;
	}
	new_rect->rect.x = x;
	new_rect->rect.y = y;
	new_rect->rect.width = width;
	new_rect->rect.height = height;
	new_rect->pnext = NULL;
	return new_rect;
}

void
agl_rectlistfree(agl_rectlist * rect)
{
	agl_rectlist *prect;

	agl_rassert(rect != NULL);

	while (rect != NULL)
	{
		prect = rect->pnext;
		agl_free(rect);
		rect = prect;
	}
}

agl_rectlist *
agl_rectlistcopy(agl_rectlist * rect)
{
	agl_rectlist *prect, *phead, *pnext;

	agl_rvassert(rect != NULL, NULL);

	phead = NULL;
	prect = NULL;

	while (rect != NULL)
	{
		pnext = agl_rectlistcreate(rect->rect.x, rect->rect.y, rect->rect.width, rect->rect.height);

		if (pnext == NULL)
		{
			agl_rectlistfree(phead);
			return NULL;
		}
		if (phead == NULL)
			phead = pnext;
		else
			prect->pnext = pnext;
		rect = rect->pnext;
		prect = pnext;
	}

	return phead;
}

int
agl_rectlistsepar(agl_rectlist ** plist)
{
}

int
agl_rectlistvunion(agl_rectlist ** plist)
{
}

int
agl_rectlisthunion(agl_rectlist ** plist)
{
}

int
_agl_rectlistinter(agl_rectlist ** pdest, agl_rect * src)
{
	agl_rect *pa, *pb;
	agl_rectlist *dest, *next;
	agl_rectlist *result;

	agl_rvassert(src != NULL, -1);
	agl_rvassert(pdest != NULL, -1);

	dest = *pdest;
	result = NULL;

	for (; dest != NULL; dest = next)
	{
		next = dest->pnext;
		if (dest->rect.x < src->x)
		{
			pa = &dest->rect;
			pb = src;
		}
		else
		{
			pa = src;
			pb = &dest->rect;
		}
		dest->rect.width = min(pa->x + pa->width - pb->x, pb->width);
		dest->rect.x = pb->x;
		if (dest->rect.width <= 0)
		{
			agl_free(dest);
			continue;
		}
		if (dest->rect.y < src->y)
		{
			pa = &dest->rect;
			pb = src;
		}
		else
		{
			pa = src;
			pb = &dest->rect;
		}
		dest->rect.height = min(pa->y + pa->height - pb->y, pb->height);
		dest->rect.y = pb->y;
		if (dest->rect.height <= 0)
		{
			agl_free(dest);
			continue;
		}
	}
	*pdest = result;
	if (result == NULL)
		return 0;
	else
		return 1;
}

int
agl_rectlistinter(agl_rectlist ** pdest, agl_rect * src)
{
	int ret;

	agl_rvassert(src != NULL, -1);
	agl_rvassert(pdest != NULL, -1);

	if ((ret = _agl_rectlistinter(pdest, src)) <= 0)
		return ret;
	return agl_rectlistvunion(pdest);
}

int
agl_rectlistinterl(agl_rectlist ** pdest, agl_rectlist * src)
{
	agl_rectlist *pcur, *pres, *pend;
	int ret;

	agl_rvassert(pdest != NULL, -1);

	if (*pdest == NULL)
		return 0;

	if (src == NULL)
	{
		agl_rectlistfree(*pdest);
		*pdest = NULL;
		return 0;
	}

	pres = pend = NULL;

	for (; src != NULL; src = src->pnext)
	{
		pcur = agl_rectlistcopy(*pdest);
		agl_rvassert(pcur != NULL, -1);

		if (_agl_rectlistinter(&pcur, &src->rect) == 0)
			return 0;
		if (pres == NULL)
			pres = pcur;
		else
		{
			if (pend == NULL)
				pend = pres;
			while (pend->pnext != NULL)
				pend = pend->pnext;
			pend->pnext = pcur;
		}
	}
	agl_rectlistfree(*pdest);
	*pdest = NULL;

	if ((ret = agl_rectlisthunion(&pres)) <= 0)
		return ret;

	if ((ret = agl_rectlistvunion(&pres)) <= 0)
		return ret;

	*pdest = pres;

	return 1;
}

int
agl_rectlistunion(agl_rectlist ** pdest, agl_rect * src)
{
	int ret;
	agl_rectlist *dest, *next;
	agl_rectlist *result;

	agl_rvassert(src != NULL, -1);
	agl_rvassert(pdest != NULL, -1);

	dest = *pdest;
	result = NULL;
	next = agl_rectlistcreate(src->x, src->y, src->width, src->height);
	agl_rvassert(next != NULL, -1);

	result = next;
	next->pnext = *pdest;
	*pdest = NULL;

	if ((ret = agl_rectlistsepar(&result)) <= 0)
		return ret;

	if ((ret = agl_rectlisthunion(&result)) <= 0)
		return ret;

	if ((ret = agl_rectlistvunion(&result)) <= 0)
		return ret;

	*pdest = result;

	return 1;
}
