
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* the underlying zone object */

#include "config.h"
#include "agl_private.h"

int
agl_zonefill(agl_zone * pzone, int width, int height, int flags)
{
	unsigned long *pixels;

	pixels = (unsigned long *) agl_malloc(width * height * 4);
	if (pixels == NULL)
	{
		agl_memerr();
		return -1;
	}
	pzone->pix.pixels = pixels;
	pzone->pix.row_size = width;
	pzone->pix.size.width = width;
	pzone->pix.size.height = height;
	pzone->pix.alpha = 255;
	pzone->visible = 0;
	pzone->upper = NULL;
	pzone->uplast = NULL;
	pzone->lower = NULL;
	pzone->zl_nb = 0;
	pzone->flags = flags;
	return 1;
}

void
agl_zoneempty(agl_zone * pzone)
{
	agl_zone_list *pupper, *opupper;

	agl_rassert(pzone != NULL);

	agl_rassert(pzone->pix.pixels != NULL);

	agl_free(pzone->pix.pixels);

	for (pupper = pzone->upper; pupper != NULL;)
	{
		opupper = pupper;
		pupper = pupper->next;
		agl_free(opupper);
	}
}

agl_zone *
agl_zonecreate(int width, int height, int flags)
{
	agl_zone *new_zone;

	new_zone = agl_malloc(sizeof(agl_zone));
	if (new_zone == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if (agl_zonefill(new_zone, width, height, flags) != 0)
	{
		agl_memerr();
		agl_free(new_zone);
		return NULL;
	}
	return new_zone;
}

/* cascade is unused. placed here for compatibility */

void
agl_zonefree(agl_zone * zone, int cascade)
{

	agl_rassert(zone != NULL);

	cascade = 1;

	agl_zoneempty(zone);

	agl_free(zone);
}

int
agl_zoneattachunder(agl_zone * parent, agl_zone * child)
{
	agl_zone_list *proot, *pnew;

	agl_rvassert(child != NULL, -1);
	agl_rvassert(parent != NULL, -1);

	pnew = (agl_zone_list *) agl_malloc(sizeof(agl_zone_list));
	if (pnew == NULL)
	{
		agl_memerr();
		return -1;
	}
	pnew->zone = child;
	pnew->nb = parent->zl_nb++;
	proot = parent->upper;
	pnew->next = proot;
	if (proot == NULL)
		parent->uplast = pnew;
	parent->upper = pnew;
	pnew = (agl_zone_list *) agl_malloc(sizeof(agl_zone_list));
	if (pnew == NULL)
	{
		agl_memerr();
		return -1;
	}
	pnew->zone = parent;
	proot = child->lower;
	pnew->next = proot;
	child->lower = pnew;
	return 0;
}

int
agl_zoneattachupper(agl_zone * parent, agl_zone * child)
{
	agl_zone_list *proot, *pnew;

	agl_rvassert(child != NULL, -1);
	agl_rvassert(parent != NULL, -1);

	pnew = (agl_zone_list *) agl_malloc(sizeof(agl_zone_list));
	if (pnew == NULL)
	{
		agl_memerr();
		return -1;
	}
	pnew->zone = child;
	pnew->next = NULL;
	if (parent->upper == NULL)
		parent->upper = pnew;
	else
		parent->uplast->next = pnew;
	parent->uplast = pnew;
	pnew = (agl_zone_list *) agl_malloc(sizeof(agl_zone_list));
	if (pnew == NULL)
	{
		agl_memerr();
		return -1;
	}
	pnew->zone = parent;
	proot = child->lower;
	pnew->next = proot;
	child->lower = pnew;
	return 0;
}

int
agl_zonedetachchilds(agl_zone * pzone)
{
	agl_zone_list *pupper, *opupper;

	agl_rvassert(pzone != NULL, -1);

	for (pupper = pzone->upper; pupper != NULL;)
	{
		//agl_dereferencezone(pupper->zone, pzone);
		opupper = pupper;
		pupper = pupper->next;
		agl_free(opupper);
	}
	agl_zonerefresh(pzone);
	return 0;
}

int
agl_zonedetachparent(agl_zone * pzone)
{
	agl_zone_list *pupper, *opupper;

	agl_rvassert(pzone != NULL, -1);

	for (pupper = pzone->upper; pupper != NULL;)
	{
		//agl_dereferencezone(pupper->zone, pzone);
		opupper = pupper;
		pupper = pupper->next;
		agl_free(opupper);
	}
	return 0;
}

int
agl_zonerefresh(agl_zone * pzone)
{
	pzone = NULL;
	return 0;
}
