
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/*!\file agl_error.h
 * \ingroup
 * \short internal header file used to manager AGL errors
 * \author Stephane APIOU
 *
 */

#ifndef _agl_error_h
#define _agl_error_h


// ! this is the size of one error message string
#define AGL_ERROR_SIZE	256
extern void agl_flusherror(void);

extern void _agl_fatal(char *message, char *file, int line, char *function);
extern void _agl_error(char *message, char *file, int line, char *function);
extern int _agl_assert(char *expression, char *file, int line, char *function);
extern void agl_msg(char *fmt, ...);

extern void *_agl_malloc(unsigned long size, char *file, int line, char *function);
extern void *_agl_calloc(unsigned long nmem, unsigned long size, char *file, int line, char *function);
extern void *_agl_realloc(void *ptr, unsigned long size, char *file, int line, char *function);
extern void _agl_free(void *ptr, char *file, int line, char *function);

#ifdef __GNUC__
#define agl_fatal(s)	_agl_fatal(s, __FILE__, __LINE__, __PRETTY_FUNCTION__)
#define agl_error(s)	_agl_error(s, __FILE__, __LINE__, __PRETTY_FUNCTION__)
#else
#define agl_fatal(s)	_agl_fatal(s, __FILE__, __LINE__, (char *)0)
#define agl_error(s)	_agl_error(s, __FILE__, __LINE__, (char *)0)
#endif

#define agl_error2(s, v1) do {char err2buf[AGL_ERROR_SIZE]; snprintf(err2buf, AGL_ERROR_SIZE, s, v1); agl_error(err2buf);} while(0)

#define agl_error3(s, v1, v2) do {char err2buf[AGL_ERROR_SIZE]; snprintf(err2buf, AGL_ERROR_SIZE, s, v1, v2); agl_error(err2buf);} while(0)

#define agl_error4(s, v1, v2, v3) do {char err2buf[AGL_ERROR_SIZE]; snprintf(err2buf, AGL_ERROR_SIZE, s, v1, v2, v3); agl_error(err2buf);} while(0)

#define agl_error5(s, v1, v2, v3, v4) do {char err2buf[AGL_ERROR_SIZE]; snprintf(err2buf, AGL_ERROR_SIZE, s, v1, v2, v3, v4); agl_error(err2buf);} while(0)

#ifdef DEBUG
#ifdef __GNUC__
#define agl_assert(expr)	((expr) ? 0 : _agl_assert(#expr, __FILE__, __LINE__, __PRETTY_FUNCTION__))
#define agl_rassert(expr)	do if (!(expr)) { _agl_assert(#expr, __FILE__, __LINE__, __PRETTY_FUNCTION__); \
				return; } while(0)
#define agl_rvassert(expr, val)	do if (!(expr)) { _agl_assert(#expr, __FILE__, __LINE__, __PRETTY_FUNCTION__); \
				return val; } while(0)
#else
#define agl_assert(expr)	((expr) ? 0 : _agl_assert(#expr, __FILE__, __LINE__, (char *)0))
#define agl_rassert(expr)	do if (!(expr)) { _agl_assert(#expr, __FILE__, __LINE__, (char *)0); \
				return; } while(0)
#define agl_rvassert(expr, val)	do if (!(expr)) { _agl_assert(#expr, __FILE__, __LINE__, (char *)0); \
				return val; } while(0)
#endif
#else
#define agl_assert(expr)	0
#define agl_rassert(expr)
#define agl_rvassert(expr, val)
#endif

#ifdef DEBUGMEM
#ifdef __GNUC__
#define agl_malloc(size)	_agl_malloc(size, __FILE__, __LINE__, __PRETTY_FUNCTION__)
#define agl_calloc(nmem, size)	_agl_calloc(nmem, size, __FILE__, __LINE__, __PRETTY_FUNCTION__)
#define agl_realloc(ptr, size)	_agl_realloc(ptr, size, __FILE__, __LINE__, __PRETTY_FUNCTION__)
#define agl_free(ptr)	_agl_free(ptr, __FILE__, __LINE__, __PRETTY_FUNCTION__)
#else
#define agl_malloc(size)	_agl_malloc(size, __FILE__, __LINE__, (char *)0)
#define agl_calloc(nmem, size)	_agl_calloc(nmem, size, __FILE__, __LINE__, (char *)0)
#define agl_realloc(ptr, size)	_agl_realloc(ptr, size, __FILE__, __LINE__, (char *)0)
#define agl_free(ptr)	_agl_free(ptr, __FILE__, __LINE__, (char *)0)
#endif
#else
#include <stdlib.h>
#define agl_malloc(size)	malloc(size)
#define agl_calloc(nmem, size)	calloc(nmem, size)
#define agl_realloc(ptr, size)	realloc(ptr, size)
#define agl_free(ptr)	free(ptr)
#endif


//! prints "out of memory" error message

/*! prints "out of memory" error message
 * \param nothing
 *
 * \return nothing 
 */

#define agl_memerr()	agl_error("out of memory")

//! prints "input/output error" error message

/*! prints "input/output error" error message
 * \param nothing
 *
 * \return nothing 
 */

#define agl_ioserr()	agl_error("input/output error")

#endif /* agl_error_h */
