
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#ifndef _agl_hash_h
#define _agl_hash_h

#define AGL_HASH	1

/* REALY USEFULL and UNIVERSAL hash table implementation */

typedef struct agl_hashitem
{
	struct agl_hashitem *next;
	unsigned long value;
	void *data;		/* optional data structure related to this
				 * hashable value */
} agl_hashitem;


typedef struct agl_hashtab
{
	unsigned short size;
	agl_hashitem **buckets;
	unsigned short buckets_used;
	unsigned long items_num;

	agl_hashitem *most_recent;

	unsigned short (*hash_func) (unsigned long value, unsigned short hash_size);
	long (*compare_func) (unsigned long value1, unsigned long value2);
	void (*item_destroy_func) (unsigned long value, void *data);
} agl_hashtab;

typedef struct agl_hashiter
{
	unsigned short curr_bucket;
	agl_hashitem **curr_item;
	agl_hashtab *hash;
} agl_hashiter;

/* Note that all parameters are optional here.
   If it is set to NULL - defaults will be used */

#define DEFAULT_HASH_SIZE 51	/* random value - not too big - not too small */

/* default hash_func is long_val%hash_size */

/* default compare_func is long_val1-long_val2 */

agl_hashtab *agl_hashcreate(unsigned short size, unsigned short (*hash_func) (unsigned long, unsigned short),
			    long (*compare_func) (unsigned long, unsigned long), void (*item_destroy_func) (unsigned long, void *));
void agl_hashdestroy(agl_hashtab ** hash);

int agl_hashadditem(agl_hashtab * hash, unsigned long value, void *data);

/* Note that trg parameter is optional here */
int agl_hashgetitem(agl_hashtab * hash, unsigned long value, void **trg);

/* here if trg is NULL then both data and value will be destroyed;
   otherwise only value will be destroyed.
   Note: if you never specifyed destroy_func  - nothing will be destroyed,
         except for HashItem itself.
 */
int agl_hashremoveitem(agl_hashtab * hash, unsigned long value, void **trg, int destroy);

/* need to run this in order to free up cached memory */
void flush_ashash_memory_pool(void);

/* if max_items == 0 then all hash items will be returned */
unsigned long agl_hashsortitems(agl_hashtab * hash, unsigned long *values, void **data, unsigned long max_items);
unsigned long agl_hashlistitems(agl_hashtab * hash, unsigned long *values, void **data, unsigned long max_items);


int agl_hashstartiter(agl_hashtab * hash, agl_hashiter * iterator);
int agl_hashnextiter(agl_hashiter * iterator);
unsigned long agl_hashcurrval(agl_hashiter * iterator);
void *agl_hashcurrdata(agl_hashiter * iterator);
void agl_hashremovecurrval(agl_hashiter * iterator, int destroy);


/**************************************************************************/

/**************************************************************************/

/* here is the set of implemented hash functions : */

/**************************************************************************/

unsigned short agl_hashptrhash(unsigned long value, unsigned short hash_size);

/* case unsensitive strings  - spaces and control chars are alowed */
unsigned short agl_hashcasestrhash(unsigned long value, unsigned short hash_size);
long agl_hashcasestrcmp(unsigned long value1, unsigned long value2);

/* case sensitive strings  - spaces and control chars are alowed */
unsigned short agl_hashstrhash(unsigned long value, unsigned short hash_size);
long agl_hashstrcmp(unsigned long value1, unsigned long value2);
void agl_hashstrdestroy(unsigned long value, void *data);

/* basically any long value, but was written originally for colors hash */
unsigned short color_hash_value(unsigned long value, unsigned short hash_size);

#endif
