
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#ifndef _agl_module_h
#define _agl_module_h

struct _agl_module;
struct _agl_objectlist;

struct _agl_object
{
	char *name;
	unsigned long flags;
	void *priv;
	struct _agl_module *pmod;
	struct _agl_objectlist *usedby;
	int defmax;
	int defcur;
	int *defidx;
	agl_elemtype *defelem;
	struct _agl_object *pnext;
	struct _agl_object *pprev;
};

#define	AGL_MOD_MOD	1
#define	AGL_MOD_FONT	2
#define	AGL_MOD_IMG	3
#define	AGL_MOD_IO	4


typedef struct _agl_moddefn
{
	char *name;
	agl_elemexpand val;
	int global;
} agl_moddefn;

#define AGL_ELEMMOD_SZ	10

typedef struct _agl_modelem
{
	char *name;
	int index;
	int (*gram_valid) (agl_object * obj, int idx, agl_elemtype * value);
	unsigned short type;
} agl_modelem;

#define AGL_OBJEVT_FLUSHED	1	//!< a linked object is flushed from the cache and will no more exist
#define AGL_OBJEVT_UNLOAD	2	//!< ask module to unload the object
#define AGL_OBJEVT_UPDATE	3	//!< ask module to update the object

//! this structure contains the module definition

/*! modules register data types and their relationships
 * modules also define standard methods to create/delete objects
 * to interface with the lexical analyser and to send events to objects
 */

struct _agl_module
{
	char *name;		//!< class name
	char *detailed_name;	//!< detailed name
	int type;		//<! module type
	agl_caller caller[AGL_MODER_SZ];	//!< caller module name
	void *priv_desc;	//!< private descriptors used by module type
	void *class_struct;	//!< private elements used by class
	int (*mod_init) (struct _agl_module *);	//!< initialize module (used internally)
	int (*mod_delete) (agl_object *);	//!< delete the current instance of the module: free only class_struct elems
	int (*mod_validate) (agl_object *);	//!< validate the new instance 1: ready and stored in cache 0: NOK -1: ERROR
	int (*mod_dependevt) (agl_object *, int, char *, agl_object *);	//!< an event that should modify a sub object status appears -> update this object
	int (*mod_caller) (struct _agl_module *, int, struct _agl_module *, void *);	//!< add/sub a called module to the caller
	int (*mod_exit) (struct _agl_module *);	//!< close the module and destroy all instances

	char *validstr;
	agl_modelem *pmodelem;	//!< lexical elements to analyse if NULL pointer, module is not a loadable one
	agl_moddefn *pmoddefn;	//!< defines for the module
	agl_handle handle;	//!< handle pointer
	struct _agl_module *pnext;	//!< pointer to next module
};

typedef agl_module *(*agl_module_init) (agl_container *);

#endif
