
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* internal header file used to manager AGL errors */

#ifndef _agl_private_h
#define _agl_private_h

#define _AGL_PRIVATE	1

/* general purpose functions */

#ifndef __GNUC__

#define min(x,y)	((x)<(y)?x:y)
#define min3(x,y,z)	(min(x,min(y,z)))
#define min4(w,x,y,z)	(min(min(w,x),min(y,z)))

#define max(x,y)	((x)<(y)?y:x)
#define max3(x,y,z)	(max(x,max(y,z)))
#define max4(w,x,y,z)	(max(max(w,x),max(y,z)))

#else

#define min(x,y)	({ typeof(x) _x=(x), _y=(y); (_x<_y)?_x:_y;})
#define min3(x,y,z)	({ typeof(x) _r=min(x,y), _z=(z); min(_r,_z);})
#define min4(w,x,y,z)	({ typeof(x) _r1=min3(x,y,z), _w=(w); min(_r1,_w);})

#define max(x,y)	({ typeof(x) _x=(x), _y=(y); (_x)<(_y)?_y:_x;})
#define max3(x,y,z)	({ typeof(x) _r=max(x,y), _z=(z); max(_r,_z);})
#define max4(w,x,y,z)	({ typeof(x) _r1=max3(x,y,z), _w=(w); max(_r1,_w);})

#endif

struct _agl_object;
struct _agl_module;

typedef struct _agl_object agl_object;
typedef struct _agl_module agl_module;

typedef struct _agl_container
{
	void *vidstruct;
	int *modes_h;
	int *modes_w;
	unsigned long flags;
	int vid_w;
	int vid_h;
	struct _agl_object *cacheroot;
	struct _agl_module *moduleroot;
} agl_container;

typedef agl_container *agl_handle;

#include "agl_error.h"

#include "agl.h"

#include "agl_module.h"
#include "agl_hash.h"




/* agl_rect associated defines */

#define agl_rectdup(dst, src) \
	(dst).x = (src).x; \
	(dst).y = (src).y; \
	(dst).width = (src).width; \
	(dst).height = (src).height

/* agl_pix associated defines */

#define agl_pixddup(dst, src) \
	(dst).pixels = (src).pixels, \
	agl_rectdup((pdst).rect, (psrc).rect)

// errors and memory functions

extern int agl_memgarbage(agl_container *);
extern int agl_initparachute(void);
extern void agl_exitparachute(void);

// grammar analyser functions

extern unsigned char *agl_getnum(unsigned char *ch, long *value);

// module initialisation functions

extern int agl_modinit(agl_container *);
extern int agl_modexit(agl_container *);

/* module initialisation functions */
extern void *agl_loadobj(char *libfile);
extern void *agl_loadfunc(void *handle, const char *name);
extern void agl_unloadobj(void *handle);

// cache functions

extern int agl_cacheexit(agl_container * phandle);
extern int agl_cacheinit(agl_container * phandle);
extern int agl_cacheadd(agl_object * pobj);
extern int agl_cachesearch(char *name, agl_handle handle, agl_object ** pobj);
extern int agl_cachemodwalk(agl_container * phandle, char *name, agl_object ** pobj);
extern int agl_cacheobjevent(agl_object * pobj, int event, agl_object * pref);
extern int agl_cacheobjget(char *name, agl_handle handle, agl_object ** pobj);

// value object functions

extern int agl_defnsubrecur(agl_elemtype * pdef, int list);
extern int agl_objaddelem(agl_object * pobj, int idx, agl_elemtype * pelem);
extern int agl_defnindex(agl_elemtype * pdef, agl_elemtype * pv, agl_elemtype ** presp);
extern int agl_defnvcmp(agl_elemtype * pdef, agl_elemtype * pv);
extern int agl_defnicmp(agl_elemtype * pdef, agl_elemtype * pv);
extern int agl_freeval(agl_elemtype * pval, int destroy);
extern int agl_elemexp2u(agl_elemexpand * pexp, agl_elemtype * pval);
extern int agl_defnaddrecur(agl_elemtype ** pdef, agl_elemtype * pv, int list);


// sublayer routines

extern int agl_sublayerinit(agl_handle);
extern int agl_sublayerexit(agl_handle);

// input routines

extern int agl_ioinit(agl_handle);
extern int agl_ioexit(agl_handle);

// event routines
extern int agl_event(agl_handle handle);

// module routines
extern agl_module *agl_modsearch(agl_container * handle, char *name);
extern int agl_modregister(agl_container * handle, agl_module * modelem);

// zone routines
extern int agl_zonerefresh(agl_zone * pzone);
extern int agl_zonefill(agl_zone * pzone, int width, int height, int flags);
extern void agl_zoneempty(agl_zone * pzone);
extern agl_zone *agl_zonecreate(int width, int height, int flags);
extern void agl_zonefree(agl_zone * zone, int cascade);
extern int agl_zoneattachunder(agl_zone * parent, agl_zone * child);
extern int agl_zoneattachupper(agl_zone * parent, agl_zone * child);
extern int agl_zonedetachchilds(agl_zone * pzone);
extern int agl_zonedetachparent(agl_zone * pzone);

// imageio objects

typedef struct _agl_imgio
{
	int (*is) (agl_ios *);
	agl_image *(*load) (agl_ios *);
	int (*save) (agl_ios *, agl_image *);
} agl_imgio;

// fontio modules functions

typedef struct _agl_fontio
{
	int (*is) (agl_module *, agl_ios *);
	agl_font *(*open) (agl_module *, agl_ios *, int);
	int (*load) (agl_font *, unsigned char);
	int (*blit) (agl_pix *, agl_font *, unsigned char, agl_pos *);
	int (*close) (agl_font *);
} agl_fontio;

extern int agl_cacheobjexists(agl_object * pobj);

typedef struct _agl_xpmcolors
{
	char *name;
	unsigned long rgba;
} agl_xpmcolors;

#endif /* agl_private_h */
