
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for bmp file format

/*!\file agl_bmpio.c
 * \ingroup video
 * internals.
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"


//! this is a palette structure used to load images
struct palette
{
	unsigned char r;	//!< red channel
	unsigned char g;	//!< green channel
	unsigned char b;	//!< blue channel
	unsigned char a;	//!< alpha channel
};

#ifdef HAVE_BMP

/* See if an image of this type is contained in a data source */
static int
agl_isbmp(agl_ios * ops)
{
	char buf[2];

	if (agl_ios_read(ops, buf, 1, 2, NULL) == 2)
		if (!strncmp(buf, "BM", 2))
			return 1;
	return (0);
}

/* Load the image  */

#define BMP_BIT		0
#define BMP_NIBB	1
#define BMP_BYTE	2
#define BMP_RLE		3
#define BMP_MASK	4
#define BMP_NONE	5

/* See if an image of this type is contained in a data source */
static agl_image *
agl_loadbmp(agl_ios * ops)
{
	struct palette *ppal, *pcol;
	unsigned char buf[52];
	int i, j;
	int palsz;
	int w, h, bpp, compress, code;
	int x, y;
	unsigned long l;
	int remainder;
	int size, offset, hsize, colormap_size;
	int cnt, nb;
	unsigned long bmsk_bmp, amsk_bmp, rmsk_bmp, gmsk_bmp;
	unsigned long bsft_bmp, asft_bmp, rsft_bmp, gsft_bmp;
	unsigned long shft;
	agl_image *new_image;
	unsigned char *ptr;
	unsigned char c;
	int norm;


	ppal = NULL;
	if (agl_ios_read(ops, buf, 1, 2, NULL) < 2)
	{
		agl_error("Error reading the BMP file");
		return NULL;
	}
	buf[2] = 0;
	if (strncmp((char *) buf, "BM", 2))
	{
		agl_error("Error reading the BMP file");
		return NULL;
	}

	/* read header */

	if (agl_ios_read(ops, buf, 1, 0X10, NULL) < 0X10)
	{
		agl_error("Error reading the BMP file");
		return NULL;
	}

	palsz = 4;
	size = agl_readle32(&buf[0X00]);
	offset = agl_readle32(&buf[0X08]);
	hsize = agl_readle32(&buf[0X0C]);
	switch (hsize)
	{
	case 0X28:
		if (agl_ios_read(ops, buf + 0X10, 1, 0X24, NULL) < 0X24)
		{
			agl_error("Error reading the BMP file");
			return NULL;
		}
		w = agl_readle32(&buf[0x10]);
		h = agl_readle32(&buf[0x14]);
		bpp = agl_readle16(&buf[0x1A]);
		compress = agl_readle32(&buf[0x1C]);
		break;
	case 0X0C:
		palsz = 3;
		if (agl_ios_read(ops, buf + 0X10, 1, 0X08, NULL) < 0X08)
		{
			agl_error("Error reading the BMP file");
			return NULL;
		}
		w = agl_readle16(&buf[0x10]);
		h = agl_readle16(&buf[0x12]);
		bpp = agl_readle16(&buf[0x16]);
		compress = 0;
		break;
	default:
		agl_error("Error reading the BMP file");
		return NULL;

	}
	code = BMP_NONE;
	switch (bpp)
	{
	case 1:
		if (compress == 0)
			code = BMP_BIT;
		break;
	case 4:
		if (compress == 0)
			code = BMP_NIBB;
		else if (compress == 2)
			code = BMP_RLE;
		break;
	case 8:
		if (compress == 0)
			code = BMP_BYTE;
		else if (compress == 1)
			code = BMP_RLE;
		break;
	case 16:
		if ((compress == 0) || (compress == 3))
			code = BMP_MASK;
		break;
	case 24:
		if ((compress == 0) || (compress == 3))
			code = BMP_MASK;
		break;
	case 32:
		if ((compress == 0) || (compress == 3))
			code = BMP_MASK;
		break;
	}
	if (code == BMP_NONE)
	{
		agl_error("Error reading the BMP file");
		return NULL;
	}

	/* read color palette */

	colormap_size = (offset - hsize - 14) / palsz;
	if ((colormap_size != 0) && (bpp >= 16))
	{
		agl_error("Error reading the BMP file");
		return NULL;
	}
	ppal = agl_malloc(sizeof(struct palette) * colormap_size);
	if (ppal == NULL)
	{
		agl_memerr();
		return NULL;
	}

	norm = 1;
	amsk_bmp = rmsk_bmp = gmsk_bmp = bmsk_bmp = 0;
	asft_bmp = rsft_bmp = gsft_bmp = bsft_bmp = 0;
	if (bpp <= 16 && colormap_size != 0)
	{
		for (i = 0; i < colormap_size; i++)
		{
			if (agl_ios_read(ops, buf, 1, palsz, NULL) < palsz)
			{
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				return NULL;
			}
			ppal[i].r = buf[2];	/* red */
			ppal[i].g = buf[1];	/* green */
			ppal[i].b = buf[0];	/* blue */
			if (palsz == 4)
				ppal[i].a = buf[3];	/* alpha */
			else
				ppal[i].a = AGL_OPAQUE;	/* alpha */
		}
	}
	else if (bpp >= 16)
	{
		if (compress == 3)
		{

			if (agl_ios_read(ops, buf, 4, 4, NULL) < 4)
			{
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				return NULL;
			}
			bmsk_bmp = agl_readle32(&buf[0]);
			for (i = 0, shft = bmsk_bmp; (i < bpp) && !(shft & 1); shft >>= 1) ;
			if (i == bpp)
			{
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				return NULL;
			}
			bsft_bmp = i;
			gmsk_bmp = agl_readle32(&buf[1]);
			for (i = 0, shft = gmsk_bmp; (i < bpp) && !(shft & 1); shft >>= 1) ;
			if (i == bpp)
			{
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				return NULL;
			}
			gsft_bmp = i;
			rmsk_bmp = agl_readle32(&buf[2]);
			for (i = 0, shft = rmsk_bmp; (i < bpp) && !(shft & 1); shft >>= 1) ;
			if (i == bpp)
			{
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				return NULL;
			}
			rsft_bmp = i;
			amsk_bmp = 0;
			asft_bmp = 0;
			if (bpp == 16)
				norm = 8;
		}
		else if (bpp == 16)
		{
			amsk_bmp = 0;
			rmsk_bmp = agl_swaple32(0x7C00);
			gmsk_bmp = agl_swaple32(0x03E0);
			bmsk_bmp = agl_swaple32(0x001F);
			asft_bmp = 0;
			rsft_bmp = 10;
			gsft_bmp = 5;
			bsft_bmp = 0;
			norm = 8;
		}
		else if (bpp == 24)
		{
			amsk_bmp = 0;
			rmsk_bmp = agl_swaple32(0xFF0000);
			gmsk_bmp = agl_swaple32(0x00FF00);
			bmsk_bmp = agl_swaple32(0x0000FF);
			asft_bmp = 0;
			rsft_bmp = 16;
			gsft_bmp = 8;
			bsft_bmp = 0;
		}
		else if (bpp == 32)
		{
			amsk_bmp = agl_swaple32(0xFF000000);
			rmsk_bmp = agl_swaple32(0x00FF0000);
			gmsk_bmp = agl_swaple32(0x0000FF00);
			bmsk_bmp = agl_swaple32(0x000000FF);
			asft_bmp = 24;
			rsft_bmp = 16;
			gsft_bmp = 8;
			bsft_bmp = 0;
		}
	}
	new_image = agl_imgnew(w, h);
	if (new_image == NULL)
	{
		agl_memerr();
		if (ppal != NULL)
			agl_free(ppal);
		return NULL;
	}
	if (code == BMP_RLE)
	{
		for (pcol = ppal, ptr = (unsigned char *) new_image->pix->pixels, i = w * h; i > 0; i--)
		{
			*ptr++ = pcol->r;
			*ptr++ = pcol->g;
			*ptr++ = pcol->b;
			*ptr++ = AGL_OPAQUE;	/* no alpha */
		}
		cnt = 2;
		remainder = 0;
	}
	else
	{
		cnt = (bpp + 7) / 8;	/* bytes to read rounded */
		remainder = ((32 - (w * bpp) % 32) % 32) / 8;	/* lines are rounded to 32 bits */
	}
	for (y = h - 1; y >= 0; y--)
	{
		ptr = (unsigned char *) new_image->pix->pixels + y * w * 4;
		for (x = 0; x < w;)
		{
			if (agl_ios_read(ops, buf, 1, cnt, NULL) < cnt)
			{
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				agl_imgfree(new_image);
				return NULL;
			}
			switch (code)
			{
			case BMP_BIT:
				c = buf[0];
				for (i = 0; (i < 8) && (x < w); i++, c <<= 1, x++)
				{
					pcol = &ppal[(c & 0X80) ? 1 : 0];
					*ptr++ = pcol->r;
					*ptr++ = pcol->g;
					*ptr++ = pcol->b;
					*ptr++ = pcol->a;
				}
				break;
			case BMP_NIBB:
				c = buf[0];
				for (i = 0; (i < 2) && (x < w); i++, c <<= 4, x++)
				{
					pcol = &ppal[(c & 0XF0) >> 4];
					*ptr++ = pcol->r;
					*ptr++ = pcol->g;
					*ptr++ = pcol->b;
					*ptr++ = pcol->a;
				}
				break;
			case BMP_BYTE:
				pcol = &ppal[buf[0]];
				*ptr++ = pcol->r;
				*ptr++ = pcol->g;
				*ptr++ = pcol->b;
				*ptr++ = pcol->a;
				x++;
				break;
			case BMP_RLE:
				if (!buf[0])
				{	/* command */
					switch (buf[1])
					{
					case 0:	/* end of row */
						break;
						continue;
					case 1:	/* end of bitmap */
						goto bmpend;
					case 2:	/* offset */
						if (agl_ios_read(ops, buf, 1, 2, NULL) < 2)
						{
							agl_error("Error reading the BMP file");
							if (ppal != NULL)
								agl_free(ppal);
							agl_imgfree(new_image);
							return NULL;
						}
						x += buf[0];
						y += buf[1];
					default:	/* uncompressed */
						nb = buf[1];
						for (j = 0; (j < nb) && (x < w);)
						{
							if (agl_ios_read(ops, buf, 1, 1, NULL) < 1)
							{
								agl_error("Error reading the BMP file");
								if (ppal != NULL)
									agl_free(ppal);
								agl_imgfree(new_image);
								return NULL;
							}
							c = buf[0];
							if (bpp == 4)
							{
								for (i = 0; (i < 2) && (x < w) && (j < nb); i++, c <<= 4)
								{
									pcol = &ppal[(c & 0XF0) >> 4];
									*ptr++ = pcol->r;
									*ptr++ = pcol->g;
									*ptr++ = pcol->b;
									*ptr++ = pcol->a;
									x++;;
									j++;
								}
							}
							else
							{
								if (x < w)
								{
									pcol = &ppal[c];
									*ptr++ = pcol->r;
									*ptr++ = pcol->g;
									*ptr++ = pcol->b;
									*ptr++ = pcol->a;
									x++;
									j++;
								}
							}
						}
						if (bpp == 4)
							nb = (nb + 1) / 2;
						if (nb % 2)
							if (agl_ios_read(ops, buf, 1, 1, NULL) < 1)
							{
								agl_error("Error reading the BMP file");
								if (ppal != NULL)
									agl_free(ppal);
								agl_imgfree(new_image);
								return NULL;
							}
					}
				}
				else
				{	/* compressed */
					nb = buf[0];
					for (j = 0; (j < nb) && (x < w);)
					{
						c = buf[1];
						if (bpp == 4)
						{
							for (i = 0; (i < 2) && (x < w) && (j < nb); i++, c <<= 4)
							{
								pcol = &ppal[(c & 0XF0) >> 4];
								*ptr++ = pcol->r;
								*ptr++ = pcol->g;
								*ptr++ = pcol->b;
								*ptr++ = pcol->a;
								x++;;
								j++;
							}
						}
						else
						{
							if (x < w)
							{
								pcol = &ppal[c];
								*ptr++ = pcol->r;
								*ptr++ = pcol->g;
								*ptr++ = pcol->b;
								*ptr++ = pcol->a;
								x++;
								j++;
							}
						}
					}
				}
				break;
			case BMP_MASK:
				l = agl_readle32(buf);
				*ptr++ = ((l & rmsk_bmp) >> rsft_bmp) * norm;
				*ptr++ = ((l & gmsk_bmp) >> gsft_bmp) * norm;
				*ptr++ = ((l & bmsk_bmp) >> bsft_bmp) * norm;
				if (amsk_bmp)
					*ptr++ = ((l & amsk_bmp) >> asft_bmp) * norm;
				else
					*ptr++ = AGL_OPAQUE;
				x++;
				break;
			default:
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				agl_imgfree(new_image);
				return NULL;
			}
		}
		if (remainder)
			if (agl_ios_read(ops, buf, 1, remainder, NULL) < remainder)
			{
				agl_error("Error reading the BMP file");
				if (ppal != NULL)
					agl_free(ppal);
				agl_imgfree(new_image);
				return NULL;
			}
	}
      bmpend:
	if (ppal != NULL)
		agl_free(ppal);
	return (new_image);
}


static agl_imgio agl_bmp_imgio = { agl_isbmp, agl_loadbmp, NULL };

agl_module agl_modtab_imgbmp = {
	"BMPLOAD",
	"loader for BMP format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_bmp_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
#else

agl_module agl_modtab_imgbmp = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_BMP */
