
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for the gif file format

/*!\file agl_gifio.c
 * \ingroup video
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"


//! this is a palette structure used to load images
struct palette
{
	unsigned char r;	//!< red channel
	unsigned char g;	//!< green channel
	unsigned char b;	//!< blue channel
	unsigned char a;	//!< alpha channel
};


#ifdef HAVE_GIF

/* See if an image of this type is contained in a data source */

static int
agl_isgif(agl_ios * ops)
{
	char buf[4];

	if (agl_ios_read(ops, buf, 1, 4, NULL) == 4)
		if (!strncmp(buf, "GIF8", 4))
			return 1;
	return (0);
}

/* Load the image  */

/* adapted from giftopnm written by David Koblas */

static agl_image *
agl_loadgif(agl_ios * ops)
{
	unsigned char buf[255], *ptr, *pb;
	unsigned char c, count, firstchar;
	struct palette *pcol, tpal[256];
	int bpp, w, h;
	int code, incode, max_code, old_code;
	int clear_code, end_code;
	int tcode[0X1000];
	unsigned char tchar[0X1000], stack[0X1000], *sp;
	int datasize, codesize, codemask;
	int i, j, x, y;
	int value, bshift;
	int inlace, oread;
	agl_image *new_image;

	static struct
	{
		char y0;
		char s;
	}
	itab[5] =
	{
		{
		0, 8}
		,
		{
		4, 8}
		,
		{
		2, 4}
		,
		{
		1, 2}
		,
		{
		0, 1}
	};

	new_image = NULL;
	if (agl_ios_read(ops, buf, 1, 6, NULL) < 6)
	{
		agl_error("Error reading the GIF file");
		return NULL;
	}
	buf[6] = 0;
	if ((strcmp((char *) buf, "GIF87a") != 0) && (strcmp((char *) buf, "GIF89a") != 0))
	{
		agl_error("Error reading the GIF file");
		return NULL;
	}
	if (agl_ios_read(ops, buf, 1, 7, NULL) < 7)
	{
		agl_error("Error reading the GIF file");
		return NULL;
	}
	w = agl_readle16(&buf[0]);
	h = agl_readle16(&buf[2]);
	bpp = 2 << (buf[4] & 0x07);	/* between 2 and 256 */

	if (buf[4] & 0X80)
	{			/* a palette follow ?  */
		for (i = 0, pcol = tpal; i < bpp; i++, pcol++)
		{
			if (agl_ios_read(ops, buf, 1, 3, NULL) < 3)
			{
				agl_error("Error reading the GIF file");
				return NULL;
			}
			pcol->r = buf[0];
			pcol->g = buf[1];
			pcol->b = buf[2];
			pcol->a = AGL_OPAQUE;
		}
	}
	while (1)
	{
		if (agl_ios_read(ops, &c, 1, 1, NULL) < 1)
		{
			agl_error("Error reading the GIF file");
			return NULL;
		}
		switch (c)
		{
		case '!':
			if (agl_ios_read(ops, buf, 1, 2, NULL) < 2)
			{
				agl_error("Error reading the GIF file");
				return NULL;
			}
			if (buf[0] == 0XFE)
			{
				c = buf[1];
				while (c)
				{	/* read fields leading with count until count == 0 */
					for (i = 0; i <= c; i++)
						if (agl_ios_read(ops, buf, 1, 1, NULL) < 1)
						{
							agl_error("Error reading the GIF file");
							return NULL;
						}
					c = buf[0];
				}
			}
			break;
		case ',':
			if (agl_ios_read(ops, buf, 1, 9, NULL) < 9)
			{
				agl_error("Error reading the GIF file");
				return NULL;
			}
			inlace = !(buf[8] & 0X40);
			if (buf[8] & 0X80)
			{	/* read local palette */
				bpp = 2 << (buf[8] & 0x07);
				for (i = 0, pcol = tpal; i < bpp; i++, pcol++)
				{
					if (agl_ios_read(ops, buf, 1, 3, NULL) < 3)
					{
						agl_error("Error reading the GIF file");
						return NULL;
					}
					pcol->r = buf[0];
					pcol->g = buf[1];
					pcol->b = buf[2];
					pcol->a = AGL_OPAQUE;
				}
				w = agl_readle16(&buf[4]);
				h = agl_readle16(&buf[6]);
			}

			if (agl_ios_read(ops, &c, 1, 1, NULL) < 1)
			{
				agl_error("Error reading the GIF file");
				return NULL;
			}

			new_image = agl_imgnew(w, h);
			if (new_image == NULL)
			{
				agl_memerr();
				return NULL;
			}

			datasize = c;
			clear_code = 1 << datasize;
			end_code = clear_code + 1;
			max_code = clear_code + 2;
			old_code = -1;
			codesize = datasize + 1;
			codemask = (1 << codesize) - 1;
			for (j = 0; j < clear_code; j++)
			{
				tcode[j] = 0;
				tchar[j] = j;
			}
			sp = stack;
			value = 0;
			bshift = 0;

			i = inlace * 4;
			y = itab[i].y0;
			x = 0;
			ptr = (unsigned char *) new_image->pix->pixels + y * w * 4;
			count = 0;
			pb = buf;
			firstchar = 0;
			oread = 0;
			for (;; count--)
			{
				if (count <= 0)
				{
					if ((agl_ios_read(ops, &count, 1, 1, NULL) < 1)
					    || (agl_ios_read(ops, buf, 1, count, NULL) < count))
					{
						agl_error("Error reading the GIF file");
						agl_imgfree(new_image);
						return NULL;
					}
					pb = buf;
				}
				if (count == 0)
					break;
				value += ((unsigned long) *pb++ << bshift);
				bshift += 8;
				while (bshift >= codesize)
				{
					code = value & codemask;
					value >>= codesize;
					bshift -= codesize;
					if (code == end_code)
					{
						if (oread != (w * h))
						{
							agl_error("Error reading the GIF file");
							agl_imgfree(new_image);
							return NULL;
						}
						else
							return (new_image);
					}
					if (code == clear_code)
					{
						codesize = datasize + 1;
						codemask = (1 << codesize) - 1;
						max_code = clear_code + 2;
						old_code = -1;
					}
					else if (old_code == -1)
					{
						*sp++ = tchar[code];
						firstchar = old_code = code;
					}
					else if (code > max_code)
					{
						agl_error("Error reading the GIF file");
						agl_imgfree(new_image);
						return NULL;
					}
					else
					{
						incode = code;
						if (code == max_code)
						{
							*sp++ = firstchar;
							code = old_code;
						}
						while (code > clear_code)
						{
							*sp++ = tchar[code];
							code = tcode[code];
						}

						*sp++ = firstchar = tchar[code];
						tcode[max_code] = old_code;
						tchar[max_code] = firstchar;
						max_code++;
						if (((max_code & codemask) == 0) && (max_code < 4096))
						{
							codesize++;
							codemask += max_code;
						}
						old_code = incode;
					}
					while (sp > stack)
					{
						pcol = tpal + *--sp;
						*ptr++ = pcol->r;
						*ptr++ = pcol->g;
						*ptr++ = pcol->b;
						*ptr++ = pcol->a;
						x++;
						oread++;
						if (x >= w)
						{
							x = 0;
							y += itab[i].s;
							if (y >= h)
							{
								i++;
								if (i >= inlace + 4)
									return (new_image);
								y = itab[i].y0;
							}
							ptr = (unsigned char *) new_image->pix->pixels + y * w * 4;
						}
					}
				}
			}
			break;
		default:
			agl_error("Error reading the GIF file");
			agl_imgfree(new_image);
			return NULL;
		}
	}
	return (new_image);

}

static agl_imgio agl_gif_imgio = { agl_isgif, agl_loadgif, NULL };

agl_module agl_modtab_imggif = {
	"GIFLOAD",
	"loader for GIF format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_gif_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#else

agl_module agl_modtab_imggif = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_GIF */
