
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for pcx file format

/*!\file agl_pcxio.c
 * \ingroup video
 * internals.
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"


//! this is a palette structure used to load images
struct palette
{
	unsigned char r;	//!< red channel
	unsigned char g;	//!< green channel
	unsigned char b;	//!< blue channel
	unsigned char a;	//!< alpha channel
};

#ifdef HAVE_PCX

/* See if an image of this type is contained in a data source */

static int
agl_ispcx(agl_ios * ops)
{
	char buf[4];

	if (agl_ios_read(ops, buf, 1, 3, NULL) != 3)
		return 0;
	if ((buf[0] == 10) && (buf[2] == 1))
		return 1;
	return (0);
}

/* Load the image  */


static agl_image *
agl_loadpcx(agl_ios * ops)
{
	unsigned char buf[128];
	struct palette ppal[16];
	struct palette *pcol, *tcol;
	unsigned char *ptr;
	unsigned char c;
	int i, j;
	int x, y;
	int first;
	int rle_cnt, rows;
	int bpp, w, h;
	int bpplanes, planes;
	int bpl;
	unsigned char mask;
	agl_image *new_image;
	int oread;
	int sft;

	if (agl_ios_read(ops, buf, 1, 128, NULL) < 128)
	{
		agl_error("Error reading the PCX file");
		return NULL;
	}
	if ((buf[0] != 10) || (buf[2] != 1))
	{
		agl_error("Error reading the PCX file");
		return NULL;
	}
	bpplanes = buf[3];
	w = agl_readle16(&buf[8]) - agl_readle16(&buf[4]) + 1;
	h = agl_readle16(&buf[10]) - agl_readle16(&buf[6]) + 1;
	for (i = 0; i < 16; i++)
	{
		ppal[i].r = buf[16 + i * 3];
		ppal[i].g = buf[17 + i * 3];
		ppal[i].b = buf[18 + i * 3];
		ppal[i].a = AGL_OPAQUE;
	}
	planes = buf[65];
	bpl = agl_readle16(&buf[66]);
	bpp = bpplanes * planes;
	if ((bpp != 1) && (bpp != 2) && (bpp != 4) && (bpp != 8) && (bpp != 24) && (bpp != 32))
	{
		agl_error("Error reading the PCX file");
		return NULL;
	}

	new_image = agl_imgnew(w, h);
	if (new_image == NULL)
	{
		agl_memerr();
		return NULL;
	}
	rle_cnt = 0;
	first = 0;
	rows = 1;
	sft = 0;
	if (bpp <= 4)
		rows = planes;
	if (bpp == 32)
		rows = 4;
	if (bpp == 24)
		rows = 3;
	for (y = 0; y < h; y++)
	{
		for (i = 0; i < rows; i++)
		{
			if (bpp > 16)
				ptr = ((unsigned char *) new_image->pix->pixels) + y * w * 4 + i;
			else
				ptr = ((unsigned char *) new_image->pix->pixels) + y * w * 4;
			for (x = 0, oread = 0, mask = 0; (x < w) || (oread < bpl); x++)
			{
				/* performs rle decodings */
				if (!mask)
				{
					if (!rle_cnt)
					{
						if (agl_ios_read(ops, &c, 1, 1, NULL) < 1)
						{
							agl_error("Error reading the PCX file");
							agl_imgfree(new_image);
							return NULL;
						}
						if ((c & 0XC0) == 0XC0)
						{
							rle_cnt = (c & 0X3F);
							if (agl_ios_read(ops, &c, 1, 1, NULL) < 1)
							{
								agl_error("Error reading the PCX file");
								agl_imgfree(new_image);
								return NULL;
							}
						}
						else
							rle_cnt = 1;
					}
					rle_cnt--;
					oread++;
					if ((bpp < 8) && (x < w))
					{
						j = 0;
						mask = -(1 << (8 - bpplanes));
						sft = 8 - (i + 1) * bpplanes;
					}
				}
				if (x < w)
				{
					switch (bpp)
					{
					case 32:
						*ptr = c;
						ptr += 4;
						break;
					case 24:
						*ptr = c;
						if (i == 2)
							ptr[1] = AGL_OPAQUE;	/* no alpha */

						ptr += 4;
						break;
					case 8:
						*ptr = c;	/* it's only color map */
						ptr += 4;
						break;
					default:	/* well, it's complicated */
						if (i == 0)
							*ptr = (c & mask) >> sft;
						else if (sft >= 0)
							*ptr |= (c & mask) >> sft;
						else
							*ptr |= (c & mask) << (-sft);
						ptr += 4;
						j++;
						sft -= bpplanes;
						mask >>= bpplanes;
					}
				}
				else
					mask = 0;
			}
		}
	}
	/* ok. for 4 bits or less , convert into colors */

	if (bpp <= 4)
	{
		for (ptr = (unsigned char *) new_image->pix->pixels, y = 0; y < h; y++)
			for (x = 0; x < w; x++)
			{
				pcol = &ppal[*ptr];
				*ptr++ = pcol->r;
				*ptr++ = pcol->g;
				*ptr++ = pcol->b;
				*ptr++ = pcol->a;
			}
	}

	/* ok. for 8 bits,  read color map and convert it */

	if (bpp == 8)
	{
		if (agl_ios_read(ops, &c, 1, 1, NULL) < 1)
		{
			agl_error("Error reading the PCX file");
			agl_imgfree(new_image);
			return NULL;
		}
		if (c != 0X0C)
		{		/* is magic here ? */
			agl_error("Error reading the PCX file");
			agl_imgfree(new_image);
			return NULL;
		}
		tcol = agl_malloc(sizeof(struct palette) * 256);
		if (tcol == NULL)
		{
			agl_memerr();
			agl_imgfree(new_image);
			return NULL;
		}
		for (pcol = tcol, i = 0; i < 256; i++, pcol++)
		{
			if (agl_ios_read(ops, buf, 1, 3, NULL) < 3)
			{
				agl_error("Error reading the PCX file");
				agl_imgfree(new_image);
				agl_free(tcol);
				return NULL;
			}
			pcol->r = buf[0];
			pcol->g = buf[1];
			pcol->b = buf[2];
			pcol->a = AGL_OPAQUE;

		}
		for (ptr = (unsigned char *) new_image->pix->pixels, y = 0; y < h; y++)
			for (x = 0; x < w; x++)
			{
				pcol = &tcol[(int) *ptr];
				*ptr++ = pcol->r;
				*ptr++ = pcol->g;
				*ptr++ = pcol->b;
				*ptr++ = pcol->a;
			}
		agl_free(tcol);
	}
	return (new_image);
}

static agl_imgio agl_pcx_imgio = { agl_ispcx, agl_loadpcx, NULL };

agl_module agl_modtab_imgpcx = {
	"PCXLOAD",
	"loader for PCX format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_pcx_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#else

agl_module agl_modtab_imgpcx = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_PCX */
