
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for png file format

/*!\file agl_pngio.c
 * \ingroup video
 * internals.
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"


#ifdef HAVE_PNG

#include <png.h>

/* this code is a modified version of  PNGLIB. thanks ! */


/* See if an image of this type is contained in a data source */
static int
agl_ispng(agl_ios * ops)
{
	unsigned char buf[8];

	if (agl_ios_read(ops, buf, 1, 8, NULL) == 8)
		return png_check_sig(buf, 8);
	return 0;
}

static void
agl_pngread(png_structp png_ptr, png_bytep data, png_size_t byte)
{
	agl_ios_read((agl_ios *) png_get_io_ptr(png_ptr), data, byte, 1, NULL);
}

/* Load the image  */

static agl_image *
agl_loadpng(agl_ios * ops)
{
	agl_image *new_image;
	png_structp png_ptr;
	png_infop info_ptr;
	int bit_depth, color_type, interlace_type;
	png_uint_32 w, h;
	png_bytep *row_tab, *rowp;
	unsigned char *pixelp;

	/* initialize png struct for reading. use it with default error and
	 * warning fn */
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL)
	{
		agl_error("png_create_read_struct error");
		return NULL;
	}

	/* initialize info struct for image size and plans reading  */
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
	{
		agl_error("png_create_info error");
		png_destroy_read_struct(&png_ptr, info_ptr ? &info_ptr : NULL, NULL);
	}

	/* standard png error handler */
	if (setjmp(png_ptr->jmpbuf))
	{
		agl_error("Error reading the PNG file");
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return NULL;
	}

	/* Set up the input control */
	png_set_read_fn(png_ptr, ops, agl_pngread);

	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &w, &h, &bit_depth, &color_type, &interlace_type, NULL, NULL);

	/* reduces encoding to 8 bits/channel if necessary */
	if (bit_depth == 16)
		png_set_strip_16(png_ptr);


	if (color_type == PNG_COLOR_TYPE_PALETTE && bit_depth <= 8)
		png_set_palette_to_rgb(png_ptr);

	png_set_expand(png_ptr);
	/* see pnglib for these calls */

	if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
		png_set_gray_1_2_4_to_8(png_ptr);

	if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_tRNS_to_alpha(png_ptr);

	if ((color_type == PNG_COLOR_TYPE_GRAY) || (color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
		png_set_gray_to_rgb(png_ptr);
	png_set_filler(png_ptr, AGL_OPAQUE, PNG_FILLER_AFTER);

	/* alpha in agl and png files are inverted */
#if AGL_OPAQUE == 0
	png_set_invert_alpha(png_ptr);
#endif

	/* update the info struct with all the settings */
	png_read_update_info(png_ptr, info_ptr);

	/* here pnglib would provide an RGBA in 4 bytes */

	/* update informatons */
	png_get_IHDR(png_ptr, info_ptr, &w, &h, &bit_depth, &color_type, &interlace_type, NULL, NULL);
	new_image = agl_imgnew(w, h);
	if (new_image == NULL)
	{
		agl_memerr();
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return NULL;
	}
	row_tab = (png_bytep *) agl_malloc(sizeof(png_bytep) * h);
	if (row_tab == NULL)
	{
		agl_memerr();
		agl_imgfree(new_image);
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return NULL;
	}
	for (rowp = row_tab, pixelp = (unsigned char *) new_image->pix->pixels; rowp < &row_tab[h]; pixelp += w * 4, rowp++)
		*rowp = (png_bytep) pixelp;
	/* standard png error handler redeclared to free the agl image if an
	 * error occurs */
	if (setjmp(png_ptr->jmpbuf))
	{
		agl_error("Error reading the PNG file");
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		agl_imgfree(new_image);
		agl_free(row_tab);
		return NULL;
	}

	/* read image */
	png_read_image(png_ptr, row_tab);
	png_read_end(png_ptr, info_ptr);

	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
	agl_free(row_tab);
	return (new_image);
}

static agl_imgio agl_png_imgio = { agl_ispng, agl_loadpng, NULL };

agl_module agl_modtab_imgpng = {
	"PNGLOAD",
	"loader for PNG format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_png_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#else

agl_module agl_modtab_imgpng = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}
	 }
	,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_PNG */
