
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for ppm file format

/*!\file agl_ppmio.c
 * \ingroup video
 * internals.
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"

#ifdef HAVE_PPM

/* See if an image of this type is contained in a data source */
static int
agl_isppm(agl_ios * ops)
{
	char buf[4];

	if (agl_ios_read(ops, buf, 1, 3, NULL) != 3)
		return 0;
	buf[3] = 0;

	if (!strcmp("P6\n", buf))
		return 1;
	if (!strcmp("P5\n", buf))
		return 1;
	if (!strcmp("P4\n", buf))
		return 1;
	if (!strcmp("P3\n", buf))
		return 1;
	if (!strcmp("P2\n", buf))
		return 1;
	if (!strcmp("P1\n", buf))
		return 1;

	return (0);
}

/* Load the image  */

static int
read_ppmnumber(agl_ios * ops, unsigned long *val)
{
	unsigned long digits;
	unsigned char c;

	digits = 0;
	while (1)
	{
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
			return 1;
		if (c == '#')
			while (c != '\n')
				if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
					return 1;
		if (isdigit(c))
			break;
	}
	while (1)
	{
		if (isdigit(c))
			digits = digits * 10 + c - '0';
		else
			break;
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
			break;
	}
	*val = digits;
	return 0;

}

static agl_image *
agl_loadppm(agl_ios * ops)
{
	unsigned char buf[3], *ptr;
	int type;
	unsigned long w, h, c;
	unsigned char val;
	int i, ascii;
	agl_image *new_image;
	unsigned long x, y;

	if (agl_ios_read(ops, buf, 1, 3, NULL) < 3)
	{
		agl_error("Error reading the PPM file");
		return NULL;
	}
	if (buf[0] != 'P')
	{
		agl_error("Error reading the PPM file");
		return NULL;
	}
	if ((buf[1] < '1') || (buf[1] > '6'))
	{
		agl_error("Error reading the PPM file");
		return NULL;
	}
	else
		type = buf[1] - '0';
	if (buf[2] != '\n')
	{
		agl_error("Error reading the PPM file");
		return NULL;
	}
	/* read header format: <space><number><space><number> [ <space><number> ] \n
	 *                                                     not for P1 and P4
	 * with comments in spaces bginning with '#' character and ending whith \n
	 */

	if (read_ppmnumber(ops, &w))
	{
		agl_error("Error reading the PPM file");
		return NULL;
	}
	if (read_ppmnumber(ops, &h))
	{
		agl_error("Error reading the PPM file");
		return NULL;
	}
	if ((type != 1) && (type != 4))
		if (read_ppmnumber(ops, &c))
		{
			agl_error("Error reading the PPM file");
			return NULL;
		}
	new_image = agl_imgnew(w, h);
	if (new_image == NULL)
	{
		agl_memerr();
		return NULL;
	}
	x = y = 0;
	ascii = (type < 4);
	switch ((type - 1) % 3)
	{
	case 0:
		ptr = (unsigned char *) new_image->pix->pixels;
		y = x = 0;
		while (y < h)
		{

			if (ascii)
			{
				if (read_ppmnumber(ops, &c))
				{
					agl_error("Error reading the PPM file");
					return NULL;
				}
				val = (c == 1) ? 0 : 255;
				*ptr++ = val;
				*ptr++ = val;
				*ptr++ = val;
				*ptr++ = AGL_OPAQUE;	/* no alpha */
				x++;
			}
			else
			{
				if (agl_ios_read(ops, &val, 1, 1, NULL) < 1)
				{
					agl_error("Error reading the PPM file");
					return NULL;
				}

				/* each bit=1 is a black pixel decode until end of line */

				for (i = 0; (i < 8) && (x < w); i++, x++, val <<= 1)
				{
					*ptr++ = (val & 0X80) ? 0 : 255;
					*ptr++ = (val & 0X80) ? 0 : 255;
					*ptr++ = (val & 0X80) ? 0 : 255;
					*ptr++ = AGL_OPAQUE;	/* no alpha */
				}
			}
			if (x >= w)
			{
				y++;
				x = 0;
			}
		}
		break;
	case 1:		/* gray scales  8 bits per pixels */
		for (ptr = (unsigned char *) new_image->pix->pixels, y = 0; y < h; y++)
			for (x = 0; x < w; x++)
			{
				if (ascii)
				{
					if (read_ppmnumber(ops, &c))
					{
						agl_error("Error reading the PPM file");
						return NULL;
					}
					val = c;
				}
				else if (agl_ios_read(ops, &val, 1, 1, NULL) < 1)
				{
					agl_error("Error reading the PPM file");
					return NULL;
				}
				*ptr++ = val;	/* red */
				*ptr++ = val;	/* green */
				*ptr++ = val;	/* blue */
				*ptr++ = AGL_OPAQUE;	/* no alpha */
			}

		break;
	case 2:		/* color 24 bits per pixels */
		for (ptr = (unsigned char *) new_image->pix->pixels, y = 0; y < h; y++)
			for (x = 0; x < w; x++)
				if (ascii)
				{
					if (read_ppmnumber(ops, &c))
					{
						agl_error("Error reading the PPM file");
						return NULL;
					}
					*ptr++ = c;	/* red */
					if (read_ppmnumber(ops, &c))
					{
						agl_error("Error reading the PPM file");
						return NULL;
					}
					*ptr++ = c;	/* green */
					if (read_ppmnumber(ops, &c))
					{
						agl_error("Error reading the PPM file");
						return NULL;
					}
					*ptr++ = c;	/* blue */
					*ptr++ = AGL_OPAQUE;	/* no alpha */
				}
				else
				{
					if (agl_ios_read(ops, ptr, 1, 3, NULL) < 3)
					{
						agl_error("Error reading the PPM file");
						return NULL;
					}
					ptr += 3;	/* red, green, blue read */
					*ptr++ = AGL_OPAQUE;	/* no alpha */
				}
		break;
	default:
		agl_error("Error reading the PPM file");
		agl_imgfree(new_image);
		return NULL;
		break;
	}
	return (new_image);
}

static agl_imgio agl_ppm_imgio = { agl_isppm, agl_loadppm, NULL };

agl_module agl_modtab_imgppm = {
	"PPMLOAD",
	"loader for PPM format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_ppm_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#else

agl_module agl_modtab_imgppm = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}
	 }
	,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_PPM */
