
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for tga file format

/*!\file agl_tgaio.c
 * \ingroup video
 * internals.
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"


//! this is a palette structure used to load images
struct palette
{
	unsigned char r;	//!< red channel
	unsigned char g;	//!< green channel
	unsigned char b;	//!< blue channel
	unsigned char a;	//!< alpha channel
};

#ifdef HAVE_TGA

/* based on tga2ppm from Jef Poskanzer. Thanks ! */

/* See if an image of this type is contained in a data source */

static int
agl_istga(agl_ios * ops)
{
	char buf[17];

	if (agl_ios_read(ops, buf, 1, 17, NULL) != 17)
		return 0;
	switch (buf[16])
	{
	case 8:
	case 16:
	case 24:
	case 32:
		break;
	default:
		return 0;
	}
	switch (buf[2])
	{
	case 1:
	case 2:
	case 3:
	case 9:
	case 10:
	case 11:
	case 32:
	case 33:
		return 1;
	}
	return (0);
}

/* Load the image  */


static agl_image *
agl_loadtga(agl_ios * ops)
{
	unsigned char buf[18];
	unsigned char buf1[256];
	unsigned char *ptr;
	struct palette *ppal, *pcol;
	int i;
	int x, y;
	char r, g, b, a;
	int mapsz;
	unsigned char c;
	unsigned short w, h, nc;
	unsigned long l;
	unsigned char rle_cnt, rle, rle_flag;
	int imgtype, col_strt, col_len;
	int bpp, reverse, interlace;
	unsigned long rmsk_bmp, gmsk_bmp, bmsk_bmp, amsk_bmp;
	unsigned long rsft_bmp, gsft_bmp, bsft_bmp, asft_bmp;
	int norm;
	agl_image *new_image;

	/* Read header of the file */

	if (agl_ios_read(ops, buf, 1, 18, NULL) < 18)
	{
		agl_error("Error reading the TGA file");
		return NULL;
	}
	i = buf[0];

	/* read the file comment */

	if (agl_ios_read(ops, buf1, 1, i, NULL) < i)
	{
		agl_error("Error reading the TGA file");
		return NULL;
	}

	imgtype = buf[2];
	col_strt = agl_readle16(&buf[3]);
	col_len = agl_readle16(&buf[5]);
	mapsz = buf[7];
	w = agl_readle16(&buf[12]);
	h = agl_readle16(&buf[14]);
	bpp = buf[16];
	reverse = (buf[17] & 0X20) == 0;
	interlace = 1 << ((buf[17] & 0XC0) >> 6);

	/* read palette */

	ppal = NULL;
	if ((imgtype == 1) || (imgtype == 9) || (imgtype == 32) || (imgtype == 33))
	{
		if (buf[1] == 1)
		{
			ppal = agl_malloc(sizeof(struct palette) * col_len);
			if (ppal == NULL)
			{
				agl_memerr();
				return NULL;
			}

			for (i = 0, pcol = ppal; i < col_len; i++, pcol++)
			{
				switch (mapsz)
				{
				case 8:
					if (agl_ios_read(ops, &c, 1, 1, NULL) < 1)
					{
						agl_error("Error reading the TGA file");
						if (ppal != NULL)
							agl_free(ppal);
						return NULL;
					}
					pcol->r = pcol->g = pcol->b = c;
					pcol->a = AGL_OPAQUE;
					break;
				case 16:
					if (agl_ios_read(ops, &w, 1, 2, NULL) < 2)
					{
						agl_error("Error reading the TGA file");
						if (ppal != NULL)
							agl_free(ppal);
						return NULL;
					}
					w = agl_swaple16(w);
					pcol->r = (w & agl_swaple16(0X7C00)) * 8;
					pcol->g = (w & agl_swaple16(0X03E0)) * 8;
					pcol->b = (w & agl_swaple16(0X001F)) * 8;
					pcol->a = AGL_OPAQUE;
					break;
				case 24:
					if (agl_ios_read(ops, buf, 1, 3, NULL) < 3)
					{
						agl_error("Error reading the TGA file");
						if (ppal != NULL)
							agl_free(ppal);
						return NULL;
					}
					pcol->r = buf[2];
					pcol->g = buf[1];
					pcol->b = buf[0];
					pcol->a = AGL_OPAQUE;
					break;
				case 32:
					if (agl_ios_read(ops, buf, 1, 4, NULL) < 4)
					{
						agl_error("Error reading the TGA file");
						if (ppal != NULL)
							agl_free(ppal);
						return NULL;
					}
					pcol->a = buf[3];
					pcol->r = buf[2];
					pcol->g = buf[1];
					pcol->b = buf[0];
					break;
				default:
					agl_error("Error reading the TGA file");
					if (ppal != NULL)
						agl_free(ppal);
					return NULL;
				}
			}
		}
		else
		{
			agl_error("Error reading the TGA file");
			if (ppal != NULL)
				agl_free(ppal);
			return NULL;
		}
	}
	rle = (imgtype & 0X08) != 0;
	/* calculate mask for each encoding */

	norm = 1;
	if (bpp == 8)
	{
		amsk_bmp = 0;
		rmsk_bmp = agl_swaple32(0xFF);
		gmsk_bmp = agl_swaple32(0xFF);
		bmsk_bmp = agl_swaple32(0xFF);
		asft_bmp = 0;
		rsft_bmp = 0;
		gsft_bmp = 0;
		bsft_bmp = 0;
	}
	else if (bpp == 16)
	{
		amsk_bmp = 0;
		rmsk_bmp = agl_swaple32(0x7C00);
		gmsk_bmp = agl_swaple32(0x03E0);
		bmsk_bmp = agl_swaple32(0x001F);
		asft_bmp = 0;
		rsft_bmp = 10;
		gsft_bmp = 5;
		bsft_bmp = 0;
		norm = 8;
	}
	else if (bpp == 24)
	{
		amsk_bmp = 0;
		rmsk_bmp = agl_swaple32(0xFF0000);
		gmsk_bmp = agl_swaple32(0x00FF00);
		bmsk_bmp = agl_swaple32(0x0000FF);
		asft_bmp = 0;
		rsft_bmp = 16;
		gsft_bmp = 8;
		bsft_bmp = 0;
	}
	else if (bpp == 32)
	{
		amsk_bmp = agl_swaple32(0xFF000000);
		rmsk_bmp = agl_swaple32(0x00FF0000);
		gmsk_bmp = agl_swaple32(0x0000FF00);
		bmsk_bmp = agl_swaple32(0x000000FF);
		asft_bmp = 24;
		rsft_bmp = 16;
		gsft_bmp = 8;
		bsft_bmp = 0;
	}
	else
	{
		agl_error("Error reading the TGA file");
		if (ppal != NULL)
			agl_free(ppal);
		return NULL;
	}

	new_image = agl_imgnew(w, h);
	if (new_image == NULL)
	{
		agl_memerr();
		if (ppal != NULL)
			agl_free(ppal);
		return NULL;
	}

	r = g = b = a = 0;
	rle_cnt = 0;
	rle_flag = 0;
	for (i = 0; i < interlace; i++)
		for (y = i; y < h; y += interlace)
		{
			if (reverse)
				ptr = ((unsigned char *) new_image->pix->pixels) + (h - 1 - y) * w * 4;
			else
				ptr = ((unsigned char *) new_image->pix->pixels) + y * w * 4;
			for (x = 0; x < w; x++)
			{
				if (rle_flag)
				{
					rle_cnt--;
					if (rle_cnt == 0)
						rle_flag = 0;
				}
				else
				{
					if ((rle) && (rle_cnt == 0))
					{
						if (agl_ios_read(ops, &rle_cnt, 1, 1, NULL) < 1)
						{
							agl_error("Error reading the TGA file");
							if (ppal != NULL)
								agl_free(ppal);
							agl_imgfree(new_image);
							return NULL;
						}
						rle_flag = rle_cnt & 0X80;
						rle_cnt = (rle_cnt & 0X7F) + 1;
					}
					if (rle_cnt)
					{
						rle_cnt--;
						if (rle_cnt == 0)
							rle_flag = 0;
					}
					if (ppal != NULL)
					{
						switch (bpp)
						{
						case 8:
							if (agl_ios_read(ops, &c, 1, 1, NULL) < 1)
								goto tga_error;
							if (c < col_strt)
								goto tga_error;
							nc = c - col_strt;
							break;
						case 16:
							if (agl_ios_read(ops, &nc, 1, 2, NULL) < 2)
								goto tga_error;
							if (nc < col_strt)
								goto tga_error;
							nc -= col_strt;
							break;
						default:
						      tga_error:
							agl_error("Error reading the TGA file");
							if (ppal != NULL)
								agl_free(ppal);
							agl_imgfree(new_image);
							return NULL;
						}
						if (nc >= col_len)
						{
							agl_error("Error reading the TGA file");
							if (ppal != NULL)
								agl_free(ppal);
							agl_imgfree(new_image);
							return NULL;
						}
						pcol = &ppal[nc];
						r = pcol->r;
						g = pcol->g;
						b = pcol->b;
						a = pcol->a;
					}
					else
					{
						if (agl_ios_read(ops, buf, 1, bpp / 8, NULL) < (bpp / 8))
						{
							agl_error("Error reading the TGA file");
							if (ppal != NULL)
								agl_free(ppal);
							agl_imgfree(new_image);
							return NULL;
						}
						l = agl_readle32(buf);
						r = ((l & rmsk_bmp) >> rsft_bmp) * norm;
						g = ((l & gmsk_bmp) >> gsft_bmp) * norm;
						b = ((l & bmsk_bmp) >> bsft_bmp) * norm;
						if (amsk_bmp)
							a = ((l & amsk_bmp) >> asft_bmp) * norm;
						else
							a = AGL_OPAQUE;
					}
				}
				*ptr++ = r;
				*ptr++ = g;
				*ptr++ = b;
				*ptr++ = a;
			}
		}
	if (ppal != NULL)
		agl_free(ppal);
	return (new_image);
}

static agl_imgio agl_tga_imgio = { agl_istga, agl_loadtga, NULL };

agl_module agl_modtab_imgtga = {
	"TGALOAD",
	"loader for TGA format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_tga_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#else

agl_module agl_modtab_imgtga = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_TGA */
