
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for the tif file format

/*!\file agl_tifio.c
 * \ingroup video
 * internals.
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"


#ifdef HAVE_TIF

#include <tiffio.h>

/* See if an image of this type is contained in a data source */

static int
agl_istif(agl_ios * ops)
{
	char buf[4];

	if (agl_ios_read(ops, buf, 1, 4, NULL) != 4)
		return 0;
	if ((buf[0] == 'M') && (buf[1] == 'M') && (buf[2] == 0x00) && (buf[3] == 0x2a))
		return 1;
	if ((buf[0] == 'I') && (buf[1] == 'I') && (buf[2] == 0x2a) && (buf[3] == 0x00))
		return 1;
	return 0;
}

/* Load the image  */

static tsize_t
tif_readproc(thandle_t ptr, tdata_t buf, tsize_t size)
{

	return agl_ios_read((agl_ios *) ptr, (char *) buf, 1, size, NULL);
}

static tsize_t
tif_writeproc(thandle_t ptr, tdata_t buf, tsize_t size)
{
	/* nothing to do : we're just reading */
	ptr = buf = NULL;
	size = 0;
	return 0;
}

static int
tif_closeproc(thandle_t ptr)
{
	/* nothing to do here */
	ptr = NULL;
	return 0;
}

static toff_t
tif_seekproc(thandle_t ptr, toff_t offset, int whence)
{
	return agl_ios_seek((agl_ios *) ptr, offset, whence);

}

static toff_t
tif_sizeproc(thandle_t ptr)
{
	/* not used yet by tif library */
	long pos;
	long size;

	pos = agl_ios_tell((agl_ios *) ptr);
	size = agl_ios_seek((agl_ios *) ptr, 0, SEEK_END);
	agl_ios_seek((agl_ios *) ptr, pos, SEEK_SET);
	return size;
}

static int
tif_mapproc(thandle_t ptr, tdata_t * buf, toff_t * psize)
{
	/* not implemented */
	psize = ptr = buf = NULL;
	return 0;
}

static void
tif_unmapproc(thandle_t ptr, tdata_t buf, toff_t size)
{
	/* nothing to do */
	ptr = buf = NULL;
	size = 0;
}

static agl_image *
agl_loadtif(agl_ios * ops)
{
	TIFF *tif;
	int w, h;
	agl_image *new_image;

	tif = TIFFClientOpen("", "r", (thandle_t) ops, tif_readproc, tif_writeproc, tif_seekproc, tif_closeproc, tif_sizeproc,
			     tif_mapproc, tif_unmapproc);
	if (tif == NULL)
	{
		agl_error("Error reading the TIF file");
		TIFFClose(tif);
		return NULL;
	}
	TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_BOTLEFT);
	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
	new_image = agl_imgnew(w, h);
	if (new_image == NULL)
	{
		agl_memerr();
		TIFFClose(tif);
		return NULL;
	}
	if (TIFFReadRGBAImage(tif, w, h, new_image->pix->pixels, 0) == 0)
	{
		agl_error("Error reading the TIF file");
		TIFFClose(tif);
		agl_imgfree(new_image);
		return NULL;
	}
	TIFFClose(tif);
	return (new_image);

}

static agl_imgio agl_tif_imgio = { agl_istif, agl_loadtif, NULL };

agl_module agl_modtab_imgtif = {
	"TIFLOAD",
	"loader for TIF format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_tif_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#else

agl_module agl_modtab_imgtif = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}
	 }
	,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_TIF */
