
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* thanks to imlib2 creator ( cK) for his adaptation of the gimp sources */

#include "config.h"
#include "agl_private.h"

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#ifdef HAVE_XCF

#define TILE_WIDTH   64
#define TILE_HEIGHT  64

typedef struct _Layer Layer;
typedef struct _Tile Tile;


#define  PROP_END  0
#define  PROP_COLORMAP  1
#define  PROP_ACTIVE_LAYER  2
#define  PROP_ACTIVE_CHANNEL  3
#define  PROP_SELECTION  4
#define  PROP_FLOATING_SELECTION  5
#define  PROP_OPACITY  6
#define  PROP_MODE  7
#define  PROP_VISIBLE  8
#define  PROP_LINKED  9
#define  PROP_PRESERVE_TRANSPARENCY  10
#define  PROP_APPLY_MASK  11
#define  PROP_EDIT_MASK  12
#define  PROP_SHOW_MASK  13
#define  PROP_SHOW_MASKED  14
#define  PROP_OFFSETS  15
#define  PROP_COLOR  16
#define  PROP_COMPRESSION  17
#define  PROP_GUIDES  18
#define  PROP_RESOLUTION  19
#define  PROP_TATTOO  20
#define  PROP_PARASITES  21
#define  PROP_UNIT  22
#define  PROP_PATHS  23
#define  PROP_USER_UNIT  24

/* Defines the compression mode of the tiles supported */

#define  COMPRESS_NONE	0
#define  COMPRESS_RLE 1

/* Layer modes */

#define  NORMAL_MODE	0
#define  DISSOLVE_MODE	1
#define  BEHIND_MODE	2	/* Unused */
#define  MULTIPLY_MODE	3
#define  SCREEN_MODE	4
#define  OVERLAY_MODE	5
#define  DIFFERENCE_MODE	6
#define  ADDITION_MODE	7
#define  SUBTRACT_MODE	8
#define  DARKEN_ONLY_MODE	9
#define  LIGHTEN_ONLY_MODE	10
#define  HUE_MODE	11
#define  SATURATION_MODE	12
#define  COLOR_MODE	13
#define  VALUE_MODE	14
#define  DIVIDE_MODE	15
#define  ERASE_MODE 	16	/* Unused */
#define  REPLACE_MODE 17	/* Unused */
#define  ANTI_ERASE_MODE 18	/* Unused */

/* Base image types */

#define  RGB	0
#define  GRAY	1
#define  INDEXED	2

struct _Layer
{
	int visible;		/* controls visibility            */
	int width, height;	/* size of drawable               */
	int bpp;		/* depth                          */
	int offset_x, offset_y;	/* offset of layer in image       */
	Layer *mask;		/*  possible layer mask           */

	int opacity;		/*  layer opacity                 */
	char mode;		/*  layer combination mode        */
	Tile *tiles;		/* tiles for drawable data        */
	int num_rows;
	int num_cols;
	unsigned char *data;		/* contains tiles serialized */
	struct _Layer *next;
	struct _Layer *prev;
};

/* The tile structure: */

struct _Tile
{
	unsigned char bpp;	/* the bytes per pixel (1, 2, 3 or 4) */
	unsigned short ewidth;	/* the effective width of the tile */
	unsigned short eheight;	/* the effective height of the tile */
	unsigned char *data;
};

#define RS AGL_REDVAL(src + s_idx)
#define GS AGL_GREENVAL(src + s_idx)
#define BS AGL_BLUEVAL(src + s_idx)
#define AS AGL_ALPHAVAL(src + s_idx)
#define RD AGL_REDVAL(dest + d_idx)
#define GD AGL_GREENVAL(dest + d_idx)
#define BD AGL_BLUEVAL(dest + d_idx)
#define AD AGL_ALPHAVAL(dest + d_idx)

#define EPS 0.00001
#define PI  3.141592654
#define INT_MULT(a,b,t)  ((t) = (a) * (b) + 0x80, ((((t) >> 8) + (t)) >> 8))
#define LINEAR(x,y,w) ((w*y + x)*4)

static Tile *
allocate_tiles(int width, int height, int bpp, int *num_rows, int *num_cols)
{
	Tile *tiles;
	int i, j, k, right_tile, bottom_tile;
	int tile_width, tile_height;

	(*num_rows) = (height + TILE_HEIGHT - 1) / TILE_HEIGHT;
	(*num_cols) = (width + TILE_WIDTH - 1) / TILE_WIDTH;

	tiles = (Tile *) agl_malloc(sizeof(Tile) * (*num_rows) * (*num_cols));
	if (!tiles)
	{
		agl_memerr();
		return NULL;
	}

	right_tile = width - (((*num_cols) - 1) * TILE_WIDTH);
	bottom_tile = height - (((*num_rows) - 1) * TILE_HEIGHT);

	for (i = 0, k = 0; i < (*num_rows); i++)
	{
		for (j = 0; j < (*num_cols); j++, k++)
		{
			tile_width = ((j == (*num_cols) - 1) ? right_tile : TILE_WIDTH);
			tile_height = ((i == (*num_rows) - 1) ? bottom_tile : TILE_HEIGHT);
			tiles[k].bpp = bpp;
			tiles[k].ewidth = tile_width;
			tiles[k].eheight = tile_height;
			tiles[k].data = (char *) agl_malloc(sizeof(char) * tile_width * tile_height * bpp);
			if (!tiles[k].data)
			{
				agl_memerr();
				agl_free(tiles);
				return NULL;
			}

		}
	}
	return tiles;
}

static void
free_tiles(Tile * tiles, int num_tiles)
{
	int i;

	for (i = 0; i < num_tiles; i++)
	{
		if (tiles[i].data)
			agl_free(tiles[i].data);
	}
	agl_free(tiles);
}

static Layer *
new_layer(int width, int height, char mode)
{
	Layer *layer;

	layer = (Layer *) agl_malloc(sizeof(Layer));
	if (!layer)
	{
		agl_error("Couldn't allocate layer.\n");
		return NULL;
	}

	bzero(layer, sizeof(Layer));

	layer->width = width;
	layer->height = height;
	layer->opacity = 255;
	layer->mode = mode;

	layer->tiles = NULL;
	layer->next = NULL;
	layer->mask = NULL;

	return layer;
}


static void
free_layer(Layer * layer)
{
	if (layer)
	{
		if (layer->tiles)
			free_tiles(layer->tiles, layer->num_rows * layer->num_cols);

		if (layer->mask)
			free_layer(layer->mask);

		if (layer->data)
			agl_free(layer->data);

		agl_free(layer);
	}
}

static void
rgb_to_hls(unsigned char *red, unsigned char *green, unsigned char *blue)
{
	int r, g, b;
	double h, l, s;
	int mini, maxi;
	int delta;

	r = *red;
	g = *green;
	b = *blue;

	if (r > g)
	{
		maxi = max(r, b);
		mini = min(g, b);
	}
	else
	{
		maxi = max(g, b);
		mini = min(r, b);
	}

	l = (maxi + mini) / 2.0;

	if (maxi == mini)
	{
		s = 0.0;
		h = 0.0;
	}
	else
	{
		delta = (maxi - mini);

		if (l < 128)
			s = 255 * (double) delta / (double) (maxi + mini);
		else
			s = 255 * (double) delta / (double) (511 - maxi - mini);

		if (r == maxi)
			h = (g - b) / (double) delta;
		else if (g == maxi)
			h = 2 + (b - r) / (double) delta;
		else
			h = 4 + (r - g) / (double) delta;

		h = h * 42.5;

		if (h < 0)
			h += 255;
		else if (h > 255)
			h -= 255;
	}

	*red = h;
	*green = l;
	*blue = s;
}


static unsigned char
gimp_hls_value(double n1, double n2, double hue)
{
	double value;

	if (hue > 255)
		hue -= 255;
	else if (hue < 0)
		hue += 255;
	if (hue < 42.5)
		value = n1 + (n2 - n1) * (hue / 42.5);
	else if (hue < 127.5)
		value = n2;
	else if (hue < 170)
		value = n1 + (n2 - n1) * ((170 - hue) / 42.5);
	else
		value = n1;

	return (unsigned char) (value * 255);
}


static void
hls_to_rgb(unsigned char *hue, unsigned char *lightness, unsigned char *saturation)
{
	double h, l, s;
	double m1, m2;

	h = *hue;
	l = *lightness;
	s = *saturation;

	if (s == 0)
	{
		/*  achromatic case  */
		*hue = l;
		*lightness = l;
		*saturation = l;
	}
	else
	{
		if (l < 128)
			m2 = (l * (255 + s)) / 65025.0;
		else
			m2 = (l + s - (l * s) / 255.0) / 255.0;

		m1 = (l / 127.5) - m2;

		/*  chromatic case  */
		*hue = gimp_hls_value(m1, m2, h + 85);
		*lightness = gimp_hls_value(m1, m2, h);
		*saturation = gimp_hls_value(m1, m2, h - 85);
	}
}


static void
rgb_to_hsv(unsigned char *red, unsigned char *green, unsigned char *blue)
{
	int r, g, b;
	double h, s, v;
	int mini, maxi;
	int delta;

	h = 0.0;

	r = *red;
	g = *green;
	b = *blue;

	if (r > g)
	{
		maxi = max(r, b);
		mini = min(g, b);
	}
	else
	{
		maxi = max(g, b);
		mini = min(r, b);
	}

	v = maxi;

	if (maxi != 0)
		s = ((maxi - mini) * 255) / (double) maxi;
	else
		s = 0;

	if (s == 0)
		h = 0;
	else
	{
		delta = maxi - mini;
		if (r == maxi)
			h = (g - b) / (double) delta;
		else if (g == maxi)
			h = 2 + (b - r) / (double) delta;
		else if (b == maxi)
			h = 4 + (r - g) / (double) delta;
		h *= 42.5;

		if (h < 0)
			h += 255;
		if (h > 255)
			h -= 255;
	}

	*red = h;
	*green = s;
	*blue = v;
}

static void
hsv_to_rgb(unsigned char *hue, unsigned char *saturation, unsigned char *value)
{
	double h, s, v;
	double f, p, q, t;

	if (*saturation == 0)
	{
		*hue = *value;
		*saturation = *value;
		*value = *value;
	}
	else
	{
		h = *hue * 6.0 / 255.0;
		s = *saturation / 255.0;
		v = *value / 255.0;

		f = h - (int) h;
		p = v * (1.0 - s);
		q = v * (1.0 - (s * f));
		t = v * (1.0 - (s * (1.0 - f)));

		switch ((int) h)
		{
		case 0:
			*hue = v * 255;
			*saturation = t * 255;
			*value = p * 255;
			break;

		case 1:
			*hue = q * 255;
			*saturation = v * 255;
			*value = p * 255;
			break;

		case 2:
			*hue = p * 255;
			*saturation = v * 255;
			*value = t * 255;
			break;

		case 3:
			*hue = p * 255;
			*saturation = q * 255;
			*value = v * 255;
			break;

		case 4:
			*hue = t * 255;
			*saturation = p * 255;
			*value = v * 255;
			break;

		case 5:
			*hue = v * 255;
			*saturation = p * 255;
			*value = q * 255;
			break;
		}
	}
}

/* translate negative destinations */
static void
_clip(int *src_tl_x, int *src_tl_y, int *src_br_x, int *src_br_y, int *dest_x, int *dest_y, int dest_w, int dest_h)
{
	if (*dest_x + *src_br_x >= dest_w)
	{
		*src_br_x -= (*dest_x + *src_br_x) - dest_w;
	}

	if (*dest_y + *src_br_y >= dest_h)
	{
		*src_br_y -= (*dest_y + *src_br_y) - dest_h;
	}

	if (*dest_x < 0)
	{
		*src_tl_x = -(*dest_x);
		*dest_x = 0;
	}
	if (*dest_y < 0)
	{
		*src_tl_y = -(*dest_y);
		*dest_y = 0;
	}
}

static void
combine_pixels_normal(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;
	long tmp;
	unsigned char *dest = (unsigned char *) ppix->pixels;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					RD = (unsigned char) (RS * ratio + RD * compl_ratio + EPS);
					GD = (unsigned char) (GS * ratio + GD * compl_ratio + EPS);
					BD = (unsigned char) (BS * ratio + BD * compl_ratio + EPS);
				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_add(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	int tmp;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp = RD + RS;
					RD = (unsigned char) (RD * compl_ratio + (tmp > 255 ? 255 : tmp) *ratio + EPS);

					tmp = GD + GS;
					GD = (unsigned char) (GD * compl_ratio + (tmp > 255 ? 255 : tmp) *ratio + EPS);

					tmp = BD + BS;
					BD = (unsigned char) (BD * compl_ratio + (tmp > 255 ? 255 : tmp) *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_sub(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	int tmp;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp = RD - RS;
					RD = (unsigned char) (RD * compl_ratio + (tmp < 0 ? 0 : tmp) *ratio + EPS);

					tmp = GD - GS;
					GD = (unsigned char) (GD * compl_ratio + (tmp < 0 ? 0 : tmp) *ratio + EPS);

					tmp = BD - BS;
					BD = (unsigned char) (BD * compl_ratio + (tmp < 0 ? 0 : tmp) *ratio + EPS);
				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_diff(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	int tmp;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp = RD - RS;
					RD = (unsigned char) (RD * compl_ratio + (tmp < 0 ? -tmp : tmp) *ratio + EPS);

					tmp = GD - GS;
					GD = (unsigned char) (GD * compl_ratio + (tmp < 0 ? -tmp : tmp) *ratio + EPS);

					tmp = BD - BS;
					BD = (unsigned char) (BD * compl_ratio + (tmp < 0 ? -tmp : tmp) *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_darken(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;
	int tmp;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp = min(RD , RS);
					RD = (unsigned char) (RD * compl_ratio + tmp *ratio + EPS);

					tmp = min(GD , GS);
					GD = (unsigned char) (GD * compl_ratio + tmp *ratio + EPS);

					tmp = min(BD , BS);
					BD = (unsigned char) (BD * compl_ratio + tmp *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_lighten(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;
	int tmp;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp = max(RD , RS);
					RD = (unsigned char) (RD * compl_ratio + tmp *ratio + EPS);

					tmp = max(GD , GS);
					GD = (unsigned char) (GD * compl_ratio + tmp *ratio + EPS);

					tmp = max(BD , BS);
					BD = (unsigned char) (BD * compl_ratio + tmp *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_mult(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	int tmp, tmp1;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp1 = INT_MULT(RS, RD, tmp);
					RD = (unsigned char) (RD * compl_ratio + tmp1 *ratio + EPS);

					tmp1 = INT_MULT(GS, GD, tmp);
					GD = (unsigned char) (GD * compl_ratio + tmp1 *ratio + EPS);

					tmp1 = INT_MULT(BS, BD, tmp);
					BD = (unsigned char) (BD * compl_ratio + tmp1 *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_div(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	int tmp, tmp1;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp1 = min(255, ((float) RD / (RS + 1)) * 256);
					RD = (unsigned char) (RD * compl_ratio + tmp1 *ratio + EPS);

					tmp1 = min(255, ((float) GD / (GS + 1)) * 256);
					GD = (unsigned char) (GD * compl_ratio + tmp1 *ratio + EPS);

					tmp1 = min(255, ((float) BD / (BS + 1)) * 256);
					BD = (unsigned char) (BD * compl_ratio + tmp1 *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_screen(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	int tmp;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp = 255 - (((255 - RD) * (255 - RS)) >> 8);
					RD = (unsigned char) (RD * compl_ratio + tmp *ratio + EPS);

					tmp = 255 - (((255 - GD) * (255 - GS)) >> 8);
					GD = (unsigned char) (GD * compl_ratio + tmp *ratio + EPS);

					tmp = 255 - (((255 - BD) * (255 - BS)) >> 8);
					BD = (unsigned char) (BD * compl_ratio + tmp *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_overlay(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	int tmp_screen, tmp_mult;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	int tmp;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;

					tmp_screen = 255 - (((255 - RD) * (255 - RS)) >> 8);
					tmp_mult = (RD * RS) >> 8;
					tmp = (RD * tmp_screen + (255 - RD) * tmp_mult) >> 8;
					RD = (unsigned char) (RD * compl_ratio + tmp *ratio + EPS);

					tmp_screen = 255 - (((255 - GD) * (255 - GS)) >> 8);
					tmp_mult = (GD * GS) >> 8;
					tmp = (GD * tmp_screen + (255 - GD) * tmp_mult) >> 8;
					GD = (unsigned char) (GD * compl_ratio + tmp *ratio + EPS);

					tmp_screen = 255 - (((255 - BD) * (255 - BS)) >> 8);
					tmp_mult = (BD * BS) >> 8;
					tmp = (BD * tmp_screen + (255 - BD) * tmp_mult) >> 8;
					BD = (unsigned char) (BD * compl_ratio + tmp *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_hsv(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y, int mode)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char tr, tg, tb;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;
	int tmp;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;
					rgb_to_hsv(&RS, &GS, &BS);
					tr = RD;
					tg = GD;
					tb = BD;
					rgb_to_hsv(&tr, &tg, &tb);

					switch (mode)
					{
						case 0:	/* hue mode */
							tr = RS;
							break;
						case 1:	/* saturation mode */
							tg = GS;
							break;
						case 2:	/* value mode */
							tb = BS;
							break;
						default:
							break;
					}

					hsv_to_rgb(&tr, &tg, &tb);
					RD = (unsigned char) (RD * compl_ratio + tr *ratio + EPS);

					GD = (unsigned char) (GD * compl_ratio + tg *ratio + EPS);

					BD = (unsigned char) (BD * compl_ratio + tb *ratio + EPS);

				}

				AD = new_alpha;
			}
		}
}


static void
combine_pixels_hue(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	combine_pixels_hsv(src, src_w, src_h, ppix, dest_x, dest_y, 0);
}


static void
combine_pixels_sat(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	combine_pixels_hsv(src, src_w, src_h, ppix, dest_x, dest_y, 1);
}


static void
combine_pixels_val(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	combine_pixels_hsv(src, src_w, src_h, ppix, dest_x, dest_y, 2);
}


static void
combine_pixels_col(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char tr, tg, tb;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;
	int tmp;

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			src_alpha = AS;

			if (src_alpha != 0)
			{
				if (src_alpha == 255)
					new_alpha = src_alpha;
				else
					new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

				if (new_alpha != 0)
				{
					ratio = (float) src_alpha / new_alpha;
					compl_ratio = 1.0 - ratio;
					rgb_to_hls(&RS, &GS, &BS);
					tr = RD;
					tg = GD;
					tb = BD;
					rgb_to_hls(&tr, &tg, &tb);
					tr = RS;
					tb = BS;
					hls_to_rgb(&tr, &tg, &tb);
					RD = (unsigned char) (RD * compl_ratio + tr *ratio + EPS);

					GD = (unsigned char) (GD * compl_ratio + tg *ratio + EPS);

					BD = (unsigned char) (BD * compl_ratio + tb *ratio + EPS);

				}

				AD = new_alpha;
			}

		}
}


static void
combine_pixels_diss(unsigned char *src, int src_w, int src_h, agl_pix * ppix, int dest_x, int dest_y)
{
	int x, y, s_idx, d_idx;
	int src_tl_x = 0, src_tl_y = 0;
	int src_br_x = src_w, src_br_y = src_h;
	unsigned char *dest = (unsigned char *) ppix->pixels;
	unsigned char src_alpha;
	unsigned char new_alpha;
	float ratio, compl_ratio;
	long tmp;

	srand(12345);

	_clip(&src_tl_x, &src_tl_y, &src_br_x, &src_br_y, &dest_x, &dest_y, ppix->size.width, ppix->size.height);

	for (y = src_tl_y; y < src_br_y; y++)
		for (x = src_tl_x; x < src_br_x; x++)
		{
			d_idx = LINEAR((dest_x + x - src_tl_x), (dest_y + y - src_tl_y), ppix->row_size);
			s_idx = LINEAR(x, y, src_w);

			if ((rand() % 255) < AS)
			{
				src_alpha = AS;

				if (src_alpha != 0)
				{
					if (src_alpha == 255)
						new_alpha = src_alpha;
					else
						new_alpha = AD + INT_MULT((255 - AD), src_alpha, tmp);

					if (new_alpha != 0)
					{
						ratio = (float) src_alpha / new_alpha;
						compl_ratio = 1.0 - ratio;
						RD = (unsigned char) (RS * ratio + RD * compl_ratio + EPS);
						GD = (unsigned char) (GS * ratio + GD * compl_ratio + EPS);
						BD = (unsigned char) (BS * ratio + BD * compl_ratio + EPS);
					}

					AD = new_alpha;
				}
			}
		}
}

static char
xcf_load_image_props(agl_ios * ops, int file_version, int *num_cols, char **cmap, int *compression)
{
	char prop_type;
	unsigned long prop_size;
	unsigned char compress;
	char buf[8];
	int i;

	while (1)
	{
		if (agl_ios_read(ops, buf, 1, 8, NULL) < 8)
		{
			agl_error("Error reading the xcf file");
			return 0;
		}
		prop_type = agl_readbe32(&buf[0X00]);
		prop_size = agl_readbe32(&buf[0X04]);

		switch (prop_type)
		{
		case PROP_END:
			return 1;
		case PROP_COLORMAP:
			if (agl_ios_read(ops, buf, 1, 4, NULL) < 4)
			{
				agl_error("Error reading the xcf file");
				return 0;
			}
			*num_cols = agl_readbe32(&buf[0X00]);
			*cmap = (char *) agl_malloc(sizeof(char) * *num_cols * 3);
			if (file_version == 0)
			{
				agl_ios_seek(ops, *num_cols, SEEK_CUR);
				for (i = 0; i < *num_cols; i++)
				{
					*cmap[i * 3 + 0] = i;
					*cmap[i * 3 + 1] = i;
					*cmap[i * 3 + 2] = i;
				}
			}
			else
			{
				if (agl_ios_read(ops, *cmap, *num_cols * 3, 1, NULL) < 1)
				{
					agl_error("Error reading the xcf file");
					return 0;
				}
			}
			break;

		case PROP_COMPRESSION:
			if (agl_ios_read(ops, &compress, 1, 1, NULL) < 1)
			{
				agl_error("Error reading the xcf file");
				return 0;
			}
			*compression = compress;
			break;

			/* don't care */
		case PROP_TATTOO:
		case PROP_PARASITES:
		case PROP_UNIT:
		case PROP_PATHS:
		case PROP_USER_UNIT:
		case PROP_GUIDES:
		case PROP_RESOLUTION:
		default:
			agl_ios_seek(ops, prop_size, SEEK_CUR);
			break;
		}
	}

	return 0;
}

static char
xcf_load_layer_props(agl_ios * ops, Layer * layer)
{
	char prop_type;
	unsigned long prop_size;
	char buf[16];
	int amount;

	while (1)
	{
		if (agl_ios_read(ops, buf, 1, 8, NULL) < 8)
		{
			agl_error("Error reading the xcf file");
			return 0;
		}
		prop_type = agl_readbe32(&buf[0X00]);
		prop_size = agl_readbe32(&buf[0X04]);

		/* for useful type, data are read in one pass */
		while (prop_size > 0)
		{
			amount = (16 < prop_size ? 16 : prop_size);
			if (agl_ios_read(ops, buf, 1, amount, NULL) < amount)
			{
				agl_error("Error reading the xcf file");
				return 0;
			}
			prop_size -= (16 < amount ? 16 : amount);
		}

		switch (prop_type)
		{
		case PROP_END:
			return 1;
		case PROP_OPACITY:
			layer->opacity = agl_readbe32(buf);
			break;
		case PROP_VISIBLE:
			layer->visible = agl_readbe32(buf);
			break;
		case PROP_OFFSETS:
			layer->offset_x = agl_readbe32(&buf[0X00]);
			layer->offset_y = agl_readbe32(&buf[0X04]);
			break;
		case PROP_MODE:
			layer->mode = agl_readbe32(buf);
			break;

			/* don't care */
		case PROP_FLOATING_SELECTION:
		case PROP_PRESERVE_TRANSPARENCY:
		case PROP_LINKED:
		case PROP_ACTIVE_LAYER:
		case PROP_TATTOO:
		case PROP_APPLY_MASK:
		case PROP_EDIT_MASK:
		case PROP_SHOW_MASK:
		case PROP_PARASITES:
		default:
			break;
		}
	}
	return 0;
}

static char
xcf_load_tile_rle(agl_ios * ops, Tile * tile, int data_length)
{
	unsigned char *data;
	unsigned char val;
	int size;
	int count;
	int length;
	int bpp;
	int i, j;
	int byte_read;
	unsigned char *xcfdata, *xcfodata, *xcfdatalimit;

	data = tile->data;
	bpp = tile->bpp;


	xcfdata = xcfodata = agl_malloc(sizeof(char) * data_length);

	 byte_read = agl_ios_read(ops, xcfdata, 1, data_length, NULL);
	if (byte_read < 0)
	{
		agl_error("Error reading the xcf file");
		return 0;
	}

	xcfdatalimit = &xcfodata[byte_read - 1];

	for (i = 0; i < bpp; i++)
	{
		data = (tile->data) + i;
		size = tile->ewidth * tile->eheight;
		count = 0;

		while (size > 0)
		{
			if (xcfdata > xcfdatalimit)
			{
				goto bogus_rle;
			}

			val = *xcfdata++;

			length = val;
			if (length >= 128)
			{
				length = 255 - (length - 1);
				if (length == 128)
				{
					if (xcfdata >= xcfdatalimit)
					{
						goto bogus_rle;
					}

					length = (*xcfdata << 8) + xcfdata[1];
					xcfdata += 2;
				}

				count += length;
				size -= length;

				if (size < 0)
				{
					goto bogus_rle;
				}

				if (&xcfdata[length - 1] > xcfdatalimit)
				{
					goto bogus_rle;
				}

				while (length-- > 0)
				{
					*data = *xcfdata++;
					data += bpp;
				}
			}
			else
			{
				length += 1;
				if (length == 128)
				{
					if (xcfdata >= xcfdatalimit)
					{
						goto bogus_rle;
					}

					length = (*xcfdata << 8) + xcfdata[1];
					xcfdata += 2;
				}

				count += length;
				size -= length;

				if (size < 0)
				{
					goto bogus_rle;
				}

				if (xcfdata > xcfdatalimit)
				{
					goto bogus_rle;
				}

				val = *xcfdata++;

				for (j = 0; j < length; j++)
				{
					*data = val;
					data += bpp;
				}
			}
		}
	}
	agl_free(xcfodata);
	return 1;

      bogus_rle:
	agl_error("bogus rle?");
	if (xcfodata)
		agl_free(xcfodata);
	return 0;
}

static char
xcf_load_level(agl_ios * ops, Tile ** tiles_p, int hierarchy_width, int hierarchy_height, int bpp, int *num_rows, int *num_cols,
	       int compression)
{
	unsigned long saved_pos;
	unsigned long offset, offset2;
	int ntiles;
	int width;
	int height;
	int i;
	int fail;
	char buf[12];

	Tile *tiles;
	Tile *current_tile;

	*tiles_p = NULL;
	if (agl_ios_read(ops, buf, 1, 12, NULL) < 12)
	{
		agl_error("Error reading the xcf file");
		return 0;
	}
	width = agl_readbe32(&buf[0X00]);
	height = agl_readbe32(&buf[0X04]);
	offset = agl_readbe32(&buf[0X08]);

	if ((width != hierarchy_width) || (height != hierarchy_height))
		return 0;
	if (offset == 0)
		return 1;

	tiles = allocate_tiles(width, height, bpp, num_rows, num_cols);
	if (tiles == NULL)
		return 0;


	ntiles = (*num_rows) * (*num_cols);
	for (i = 0; i < ntiles; i++)
	{
		current_tile = &(tiles[i]);
		fail = 0;

		if (offset == 0)
		{
			agl_error("Not enough tiles found in level\n");
			free_tiles(tiles, (*num_rows) * (*num_cols));
			return 0;
		}
		saved_pos = agl_ios_tell(ops);

		if (agl_ios_read(ops, buf, 1, 4, NULL) < 4)
		{
			agl_error("Error reading the xcf file");
			free_tiles(tiles, (*num_rows) * (*num_cols));
			return 0;
		}
		offset2 = agl_readbe32(buf);

		/* if the offset is 0 then we need to read in the maximum possible
		 * allowing for negative compression */
		if (offset2 == 0)
			offset2 = offset + TILE_WIDTH * TILE_WIDTH * 4 * 1.5;	/* 1.5 is probably more than we need to allow */

		agl_ios_seek(ops, offset, SEEK_SET);
		switch (compression)
		{
		case COMPRESS_NONE:
			if (agl_ios_read
			    (ops, current_tile->data, current_tile->ewidth * current_tile->eheight * current_tile->bpp, 1, NULL) < 1)
			{
				agl_error("Error reading the xcf file");
				free_tiles(tiles, (*num_rows) * (*num_cols));
				return 0;
			}
			break;
		case COMPRESS_RLE:
			if (!xcf_load_tile_rle(ops, current_tile, offset2 - offset))
			{
				free_tiles(tiles, (*num_rows) * (*num_cols));
				return 0;
			}
			break;
		default:
			agl_error("unimplemented compression algorithm");
			free_tiles(tiles, (*num_rows) * (*num_cols));
			return 0;
		}
		agl_ios_seek(ops, saved_pos, SEEK_SET);
		if (agl_ios_read(ops, buf, 4, 1, NULL) < 1)
		{
			agl_error("Error reading the xcf file");
			free_tiles(tiles, (*num_rows) * (*num_cols));
			return 0;
		}
		offset = agl_readbe32(buf);
	}

	if (offset != 0)
	{
		agl_error2("encountered garbage after reading level: %lu\n", offset);
		free_tiles(tiles, (*num_rows) * (*num_cols));
		return 0;
	}

	*tiles_p = tiles;
	return 1;
}

static char
xcf_load_hierarchy(agl_ios * ops, Tile ** tiles, int *num_rows, int *num_cols, int *bpp, int compression)
{
	unsigned long saved_pos;
	unsigned long offset;
	unsigned long junk;
	int width;
	int height;
	char buf[16];

	if (agl_ios_read(ops, buf, 1, 16, NULL) < 16)
	{
		agl_error("Error reading the xcf file");
		return 0;
	}
	width = agl_readbe32(&buf[0X00]);
	height = agl_readbe32(&buf[0X04]);
	*bpp = agl_readbe32(&buf[0X08]);
	offset = agl_readbe32(&buf[0X0C]);

	/* discard offsets for layers below first, if any.  */
	do
	{
		if (agl_ios_read(ops, &junk, 1, 4, NULL) < 4)
		{
			agl_error("Error reading the xcf file");
			return 0;
		}
		junk = agl_readbe32(&junk);
	}
	while (junk != 0);

	/* save the current position as it is where the
	 *  next level offset is stored.
	 */
	saved_pos = agl_ios_tell(ops);

	/* seek to the level offset */
	agl_ios_seek(ops, offset, SEEK_SET);

	/* read in the level */
	if (!xcf_load_level(ops, tiles, width, height, *bpp, num_rows, num_cols, compression))
		return 0;

	/* restore the saved position so we'll be ready to
	 *  read the next offset.
	 */
	agl_ios_seek(ops, saved_pos, SEEK_SET);


	return 1;
}

static char
xcf_load_channel_props(agl_ios * ops, Layer * layer)
{
	char prop_type;
	unsigned long prop_size;
	char buf[16];
	int amount;

	while (1)
	{
		if (agl_ios_read(ops, buf, 1, 8, NULL) < 8)
		{
			agl_error("Error reading the xcf file");
			return 0;
		}
		prop_type = agl_readbe32(&buf[0X00]);
		prop_size = agl_readbe32(&buf[0X04]);
		while (prop_size > 0)
		{
			amount = (16 < prop_size ? 16 : prop_size);
			if (agl_ios_read(ops, buf, 1, amount, NULL) < amount)
			{
				agl_error("Error reading the xcf file");
				return 0;
			}
			prop_size -= (16 < amount ? 16 : amount);
		}

		switch (prop_type)
		{
		case PROP_END:
			return 1;
		case PROP_OPACITY:
			layer->opacity = agl_readbe32(buf);
			break;
		case PROP_VISIBLE:
			layer->visible = agl_readbe32(buf);
			break;
		case PROP_ACTIVE_CHANNEL:
		case PROP_SHOW_MASKED:
		case PROP_SELECTION:
		case PROP_COLOR:
		case PROP_TATTOO:
		case PROP_PARASITES:
		default:
			break;
		}
	}
	return 0;
}

static void
read_tiles_into_data(Tile * tiles, int num_cols, int width, int height, int bpp, unsigned char **data_p, int use_cmap, char *cmap)
{
	int tile_x, tile_y, x, y, offset_x, offset_y;
	unsigned char *data;
	unsigned char *ptr;
	unsigned char *ptr2;
	Tile *t;

	if (tiles)
	{
		if (*data_p)
			agl_free(*data_p);

		/* Always allocate the data as 4 bytes per pixel */
		data = (*data_p) = (unsigned char *) agl_malloc(sizeof(unsigned long) * width * height);

		ptr = data;

		for (y = 0; y < height; y++)
		{
			for (x = 0; x < width; x++)
			{
				tile_x = x / TILE_WIDTH;
				tile_y = y / TILE_HEIGHT;
				offset_x = x % TILE_WIDTH;
				offset_y = y % TILE_HEIGHT;

				t = &tiles[tile_y * num_cols + tile_x];
				ptr2 = &(t->data[offset_y * t->ewidth * bpp + offset_x * bpp]);

				switch (bpp)
				{
				case 1:
					/* use colormap if the image has one */
					if (cmap && use_cmap)
					{
						AGL_REDVAL(ptr) = cmap[*(ptr2) * 3];
						AGL_GREENVAL(ptr) = cmap[*(ptr2) * 3 + 1];
						AGL_BLUEVAL(ptr) = cmap[*(ptr2) * 3 + 2];
						AGL_ALPHAVAL(ptr) = 255;
					}
					/* else use colors themselves */
					else
					{
						AGL_REDVAL(ptr) = *(ptr2);
						AGL_GREENVAL(ptr) = *(ptr2);
						AGL_BLUEVAL(ptr) = *(ptr2);
						AGL_ALPHAVAL(ptr) = 255;
					}
					break;
				case 2:
					/* use colormap if the image has one */
					if (cmap && use_cmap)
					{
						AGL_REDVAL(ptr) = cmap[*(ptr2) * 3];
						AGL_GREENVAL(ptr) = cmap[*(ptr2) * 3 + 1];
						AGL_BLUEVAL(ptr) = cmap[*(ptr2) * 3 + 2];
						AGL_ALPHAVAL(ptr) = *(ptr2 + 1);
					}
					else
						agl_error("There's nothing to see here. 2 bpp without colormap not implemented yet.\n");
					break;
				case 3:
					AGL_REDVAL(ptr) = *(ptr2);
					AGL_GREENVAL(ptr) = *(ptr2 + 1);
					AGL_BLUEVAL(ptr) = *(ptr2 + 2);
					AGL_ALPHAVAL(ptr) = 255;
					break;
				default:
					AGL_REDVAL(ptr) = *(ptr2);
					AGL_GREENVAL(ptr) = *(ptr2 + 1);
					AGL_BLUEVAL(ptr) = *(ptr2 + 2);
					AGL_ALPHAVAL(ptr) = *(ptr2 + 3);
					break;
				}
				ptr += 4;
			}
		}
	}
}

static Layer *
xcf_load_channel(agl_ios * ops, char *cmap, int compression)
{
	Layer *layer;
	unsigned long hierarchy_offset;
	int width;
	int height;
	char buf[12];
	unsigned long tmp;

	/* read in the layer width, height and name */
	if (agl_ios_read(ops, buf, 1, 12, NULL) < 12)
	{
		agl_error("Error reading the xcf file");
		return NULL;
	}
	width = agl_readbe32(&buf[0X00]);
	height = agl_readbe32(&buf[0X04]);
	tmp = agl_readbe32(&buf[0X08]);
	agl_ios_seek(ops, tmp, SEEK_CUR);	/* skip layer name */

	/* create a new channel */
	layer = new_layer(width, height, NORMAL_MODE);
	if (!layer)
		return NULL;

	/* read in the channel properties */
	if (!xcf_load_channel_props(ops, layer))
		goto error;

	/* read the hierarchy and layer mask offsets */
	if (agl_ios_read(ops, buf, 1, 4, NULL) < 4)
	{
		agl_error("Error reading the xcf file");
		return NULL;
	}
	hierarchy_offset = agl_readbe32(buf);

	/* read in the hierarchy */
	agl_ios_seek(ops, hierarchy_offset, SEEK_SET);
	if (!xcf_load_hierarchy(ops, &(layer->tiles), &(layer->num_rows), &(layer->num_cols), &(layer->bpp), compression))
		goto error;

	read_tiles_into_data(layer->tiles, layer->num_cols, layer->width, layer->height, layer->bpp, &(layer->data), 0, cmap);
	free_tiles(layer->tiles, layer->num_rows * layer->num_cols);
	layer->tiles = NULL;


	return layer;

      error:
	free_layer(layer);
	return NULL;
}

static void
set_layer_opacity(Layer * layer)
{
	int i;
	unsigned char *ptr;

	if (layer)
	{
		if (layer->opacity != 255)
		{
			for (i = 0, ptr = layer->data; i < layer->width * layer->height; i++, ptr += 4)
			{
				*(ptr + 3) = (*(ptr + 3) * layer->opacity) >> 8;
			}
		}
	}
}

static void
apply_layer_mask(Layer * layer)
{
	unsigned char *ptr1;
	unsigned char *ptr2;
	int i, tmp;


	if (layer)
	{
		if (layer->mask)
		{
			ptr1 = layer->data;
			ptr2 = layer->mask->data;

			for (i = 0; i < layer->width * layer->height; i++)
			{
				tmp = (*(ptr1 + 3) * *(ptr2)) / 256;
				if (tmp > 255)
					tmp = 255;

				*(ptr1 + 3) = (unsigned char) tmp;
				ptr1 += 4;
				ptr2 += 4;
			}
		}
	}
}

static int
xcf_load_layer(agl_ios * ops, Layer ** player, char *cmap, int compression)
{
	Layer *layer;
	Layer *layer_mask;
	unsigned long hierarchy_offset;
	unsigned long layer_mask_offset;
	int width;
	int height;
	int tmp;
	char buf[16];


	/* read in the layer width, height and type */
	if (agl_ios_read(ops, buf, 1, 16, NULL) < 16)
	{
		agl_error("Error reading the xcf file");
		return -1;
	}
	width = agl_readbe32(&buf[0X00]);
	height = agl_readbe32(&buf[0X04]);
	tmp = agl_readbe32(&buf[0X0C]);
	agl_ios_seek(ops, tmp, SEEK_CUR);	/* skip layer name */

	/* create a new layer */
	layer = new_layer(width, height, NORMAL_MODE);
	if (!layer)
		return -1;

	/* read in the layer properties */
	if (!xcf_load_layer_props(ops, layer))
		goto error;


	if (!layer->visible) {
		free_layer(layer);
		*player = NULL;
		return 1;
	}

	/* read the hierarchy and layer mask offsets */
	if (agl_ios_read(ops, buf, 1, 8, NULL) < 8)
	{
		agl_error("Error reading the xcf file");
		goto error;
	}
	hierarchy_offset = agl_readbe32(&buf[0X00]);
	layer_mask_offset = agl_readbe32(&buf[0X04]);

	/* read in the hierarchy */
	agl_ios_seek(ops, hierarchy_offset, SEEK_SET);
	if (!xcf_load_hierarchy(ops, &(layer->tiles), &(layer->num_rows), &(layer->num_cols), &(layer->bpp), compression))
		goto error;

	/* read in the layer mask */
	if (layer_mask_offset != 0)
	{
		agl_ios_seek(ops, layer_mask_offset, SEEK_SET);

		layer_mask = xcf_load_channel(ops, cmap, compression);
		if (!layer_mask)
			goto error;

		/* set the offsets of the layer_mask */
		layer_mask->offset_x = layer->offset_x;
		layer_mask->offset_y = layer->offset_y;
		layer->mask = layer_mask;
	}

	read_tiles_into_data(layer->tiles, layer->num_cols, layer->width, layer->height, layer->bpp, &(layer->data), 1, cmap);
	free_tiles(layer->tiles, layer->num_rows * layer->num_cols);
	layer->tiles = NULL;

	set_layer_opacity(layer);

	if (layer->mask)
		apply_layer_mask(layer);

	*player = layer;
	return 1;

      error:
	free_layer(layer);
	*player = NULL;
	return -1;
}

static void
flatten_image(agl_pix * ppix, Layer * player)
{
	Layer *l = player;
	Layer *lp;

	while (l)
	{
		switch (l->mode)
		{
		case MULTIPLY_MODE:
			combine_pixels_mult(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case DIVIDE_MODE:
			combine_pixels_div(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case SCREEN_MODE:
			combine_pixels_screen(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case OVERLAY_MODE:
			combine_pixels_overlay(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case DIFFERENCE_MODE:
			combine_pixels_diff(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case ADDITION_MODE:
			combine_pixels_add(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case SUBTRACT_MODE:
			combine_pixels_sub(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case DARKEN_ONLY_MODE:
			combine_pixels_darken(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case LIGHTEN_ONLY_MODE:
			combine_pixels_lighten(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;

		case HUE_MODE:
			combine_pixels_hue(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case SATURATION_MODE:
			combine_pixels_sat(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case VALUE_MODE:
			combine_pixels_val(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case COLOR_MODE:
			combine_pixels_col(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case DISSOLVE_MODE:
			combine_pixels_diss(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;
		case NORMAL_MODE:
			combine_pixels_normal(l->data, l->width, l->height, ppix, l->offset_x, l->offset_y);
			break;

		default:
			agl_error2("Unknown layer mode: %i. Skipping.\n", l->mode);
		}

		lp = l->prev;
		free_layer(l);

		l = lp;
	}
}

static agl_image *
agl_loadxcf(agl_ios * ops)
{
	char id[14];
	Layer *lp;
	agl_image *new_image;
	Layer *layer, *layers, *last_layer;
	unsigned long saved_pos;
	unsigned long offset;
	char buf[12];
	char *cmap;
	int width, height;
	int compression, file_version;
	int num_cols;

	agl_rvassert(ops != NULL, NULL);

	layers = NULL;
	last_layer = NULL;
	cmap = NULL;
	num_cols = 0;
	compression = 0;

	if (agl_ios_read(ops, id, 1, 14, NULL) == 14)
	{
		if (strncmp(id, "gimp xcf ", 9))
			return NULL;
		else if (strcmp(id + 9, "file") == 0)
			file_version = 0;
		else if (id[9] == 'v')
			file_version = atoi(id + 10);
		else
			return NULL;
	}
	else
		return NULL;
	/* read in the image width, height and type */
	if (agl_ios_read(ops, buf, 1, 12, NULL) < 12)
	{
		agl_error("Error reading the xcf file");
		return 0;
	}
	width = agl_readbe32(&buf[0X00]);
	height = agl_readbe32(&buf[0X04]);

	/* read the image properties */
	if (!xcf_load_image_props(ops, file_version, &num_cols, &cmap, &compression))
	{
		agl_error("Error reading the xcf file");
		goto error;
	}

	new_image = agl_imgnew(width, height);
	if (new_image == NULL)
	{
		agl_memerr();
		return NULL;
	}

	while (1)
	{
		/* read in the offset of the next layer */
		if (agl_ios_read(ops, buf, 1, 4, NULL) < 4)
		{
			agl_error("Error reading the xcf file");
			agl_imgfree(new_image);
			new_image = NULL;
			goto error;
		}
		offset = agl_readbe32(buf);
		if (offset == 0)
			break;

		/* read in the layer */
		saved_pos = agl_ios_tell(ops);
		agl_ios_seek(ops, offset, SEEK_SET);
		if (xcf_load_layer(ops, &layer, cmap, compression) < 0)
		{
			agl_error("invalid xcf file");
			agl_imgfree(new_image);
			new_image = NULL;
			goto error;
		}
		if (layer)
		{
			if (last_layer)
			{
				last_layer->next = layer;
				layer->prev = last_layer;
			}
			else
			{
				layers = layer;
				layer->prev = NULL;
			}
			layer->next = NULL;
			last_layer = layer;
		}
		agl_ios_seek(ops, saved_pos, SEEK_SET);

	}
	flatten_image(new_image->pix, last_layer);
	layers = NULL;
	last_layer = NULL;

      error:
	layer = last_layer;
	while (layer)
	{
		lp = layer->prev;
		free_layer(layer);
		layer = lp;
	}

	if (cmap)
		agl_free(cmap);

	return new_image;
}

static int
agl_isxcf(agl_ios * ops)
{
	char buf[9];

	if (agl_ios_read(ops, buf, 1, 9, NULL) == 9)
		if (!strncmp(buf, "gimp xcf ", 9))
			return 1;
	return 0;
}

static agl_imgio agl_xcf_imgio = { agl_isxcf, agl_loadxcf, NULL };

agl_module agl_modtab_imgxcf = {
	"XCFLOAD",
	"loader for XCF format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_xcf_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
#else

agl_module agl_modtab_imgxcf = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}
	 }
	,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_XCF */
