
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* This file contains code for unified image loading from XPM file  *
********************************************************************
* Copyright (c) 2001 Sasha Vasko <sasha at aftercode.net>           *
********************************************************************
*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "config.h"
#include "agl_private.h"

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#ifdef HAVE_XPM

extern agl_xpmcolors agl_xpmcol[];

/* See if an image of this type is contained in a data source */

static int
agl_isxpm(agl_ios * ops)
{
	char buf[9];

	if (agl_ios_read(ops, buf, 1, 9, NULL) == 9)
		if (!strncmp(buf, "/* XPM */", 9))
			return 1;
	return 0;
}

static int
strcmpcase(char *src, char *dst)
{
	char cs, cd;

	while (*src)
		if ((cs = toupper(*src++)) != (cd = toupper(*dst++)))
		{
			if (cs > cd)
				return 1;
			else
				return -1;
		}
	return 0;
}

static unsigned long
xpm_parse_color(char *color)
{
	agl_xpmcolors *pxpm;
	int i, j;
	unsigned long c[3];
	int len;
	char str[8], *ptr;

	c[0] = c[1] = c[2] = 0;

	if (color[0] == '#')
	{
		len = strlen(color) - 1;
		if (len <= 24)
		{
			ptr = color + 1;
			len /= 3;
			str[len] = 0;
			for (ptr = color + 1, j = 0; j < 3; j++)
			{
				for (i = 0; i < len; i++)
					str[i] = *ptr++;
				c[j] = strtol(str, NULL, 16);
				if (len == 1)
					c[j] = (c[j] << 4) | c[j];
				else if (len > 2)
					c[j] >>= (len - 2) << 2;
			}
		}
		return AGL_VALRGBA2LONG(c[0], c[1], c[2], 255);
	}

	if (!strcmpcase(color, "none"))
		return AGL_SETALPHA(AGL_TRANSPARENT);

	/* look in rgb txt database */

	for (pxpm = agl_xpmcol; pxpm->name; pxpm++)
		if (!strcmpcase(pxpm->name, color))
			return pxpm->rgba;

	return AGL_SETALPHA(AGL_TRANSPARENT);
}

static agl_image *
agl_loadxpm(agl_ios * ops)
{
	agl_image *new_image;
	int pc, c, i, j, k, w, h, ncolors, cpp, comment, quote;
	long lineidx, lidx, colidx;
	int context, len, done, lsz;
	char *nptr, *line, s[256], col[256];
	unsigned long *cmap, cmapsz, *ptr, *ptr1;
	int iscolor, isset;

	agl_rvassert(ops != NULL, NULL);

	done = 0;

	if (agl_ios_read(ops, s, 1, 9, NULL) == 9)
		if (strncmp(s, "/* XPM */", 9))
			return NULL;

	i = 0;
	lineidx = 0;
	lidx = 0;
	lsz = 256;
	cmap = NULL;
	w = 0;
	h = 0;
	ptr = NULL;
	c = ' ';
	comment = 0;
	quote = 0;
	context = 0;

	line = agl_malloc(lsz);
	while (!done)
	{
		pc = c;
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
		{
			agl_error("xpm read error");
			goto end;
		}
		/* Scan in line from XPM file */
		if ((!comment) && (quote) && (c != '"'))
		{
			if (c < 32)
				c = 32;
			else if (c > 127)
				c = 127;
			line[lidx++] = c;
		}
		if (lidx >= lsz)
		{
			agl_error("xpm line malformed. unable to load file");
		}
		if (!quote)
		{
			if ((pc == '/') && (c == '*'))
				comment = 1;
			else if ((pc == '*') && (c == '/') && (comment))
				comment = 0;
		}
		if ((!comment) && (!quote) && (c == '"'))
		{
			quote = 1;
			lidx = 0;
			continue;
		}
		if ((!comment) && (quote) && (c == '"'))
		{
			line[lidx] = 0;
			quote = 0;
			switch (context)
			{
			case 0:	/* Header */
				sscanf(line, "%i %i %i %i", &w, &h, &ncolors, &cpp);
				if ((w < 0) || (h < 0) || (ncolors < 0) || (cpp < 0) || (cpp > 3))
				{
					agl_error("invalid or unsupported xpm format");
					goto end;
				}
				for (i = 0, cmapsz = 1; i < cpp; i++)
					cmapsz *= 96;
				cmap = agl_malloc(sizeof(unsigned long) * cmapsz);
				if (!cmap)
				{
					agl_memerr();
					agl_free(line);
					return 0;
				}
				if (lsz < (w * cpp))
				{
					lsz = w * cpp + 1;
					nptr = agl_realloc(line, lsz + 1);
					if (nptr == NULL)
					{
						agl_memerr();
						goto end;
					}
					line = nptr;
				}
				lineidx = 0;
				context++;
				break;
			case 1:	/* Color Table */
				iscolor = 0;
				isset = 0;
				col[0] = 0;
				len = strlen(line);
				for (colidx = 0, i = 0, j = 1; i < cpp; i++, j *= 96)
					colidx += (line[i] - ' ') * j;
				for (i = cpp; i < len; i++)
				{
					if ((line[i] != ' ') && (line[i] != '\t'))
					{
						for (j = 0; (i < len) && (line[i] != ' ') && (line[i] != '\t'); i++, j++)
							s[j] = line[i];
						s[j] = 0;
						if ((!strcmp(s, "m")) || (!strcmp(s, "s")) || (!strcmp(s, "g4")) || (!strcmp(s, "g"))
						    || (!strcmp(s, "c")) || (i >= len))
						{
							if (i >= len)
							{
								if (col[0])
									strcat(col, " ");
								if (strlen(col) + strlen(s) < sizeof(col))
									strcat(col, s);
							}
							if (col[0])
							{
								if (iscolor > isset)
								{
									isset = iscolor;
									cmap[colidx] = xpm_parse_color(col);
								}
							}
							col[0] = 0;
							if (!strcmp(s, "s"))
								iscolor = 0;
							if (!strcmp(s, "m"))
								iscolor = 1;
							if (!strcmp(s, "g4"))
								iscolor = 2;
							if (!strcmp(s, "g"))
								iscolor = 3;
							if (!strcmp(s, "c"))
								iscolor = 4;
						}
						else
						{
							if (col[0])
								strcat(col, " ");
							strcat(col, s);
						}
					}
				}
				lineidx++;

				if (lineidx >= ncolors)
				{
					new_image = agl_imgnew(w, h);
					if (new_image == NULL)
					{
						agl_memerr();
						return NULL;
					}
					context++;
					lineidx = 0;
					ptr = new_image->pix->pixels;
				}
				break;
			default:	/* Image Data */
				if ((w * cpp) != lidx)
				{
					agl_error("malformed xpm file");
					goto end;
				}
				for (i = 0, ptr1 = ptr; i < lidx;)
				{
					for (colidx = 0, j = 0, k = 1; j < cpp; j++, i++, k *= 96)
						colidx += (line[i] - ' ') * k;
					*ptr1++ = cmap[colidx];
				}
				lineidx++;
				ptr += new_image->pix->row_size;

				if (lineidx >= h)
					done = 1;
			}
		}
	}

      end:
	agl_free(cmap);
	agl_free(line);
	if ((!done) && (new_image != NULL))
	{
		agl_imgfree(new_image);
		return NULL;
	}

	return new_image;
}

static agl_imgio agl_xpm_imgio = { agl_isxpm, agl_loadxpm, NULL };

agl_module agl_modtab_imgxpm = {
	"XPMLOAD",
	"loader for XPM format",
	AGL_MOD_IMG,
	{{NULL, NULL}},
	(void *) &agl_xpm_imgio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
#else

agl_module agl_modtab_imgxpm = {
	NULL,
	NULL,
	AGL_MOD_IMG,
	{{NULL, NULL}
	 }
	,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_XPM */
