
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/


#include <unistd.h>
#include <string.h>

#include "config.h"
#include "agl_private.h"

static int
fhandle_is(agl_module * pmod, char *name, agl_ios * parent)
{
	long handle;
	char *p;

	agl_rvassert(pmod != NULL, -1);
	agl_rvassert(parent == NULL, -1);

	p = agl_getnum(name, &handle);
	if (p == NULL)
	{
		agl_error("HANDLE: unable to get handle desc");
		return -1;
	}
	return 1;
}

static agl_ios *
fhandle_open(agl_module * pmod, char *name, agl_ios * parent)
{
	long handle;
	char *p;
	agl_ios *ios;

	agl_rvassert(pmod != NULL, NULL);
	agl_rvassert(parent == NULL, NULL);

	p = agl_getnum(name, &handle);
	if (p == NULL)
	{
		agl_error("HANDLE: unable to get handle desc");
		return NULL;
	}

	if ((ios = agl_malloc(sizeof(agl_ios))) == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if ((ios->filename = agl_malloc(p - name + 1)) == NULL)
	{
		agl_memerr();
		agl_free(ios);
		return NULL;
	}
	strncpy(ios->filename, name, p - name);
	ios->priv = (void *) handle;
	ios->pio = (agl_iosio *) pmod->priv_desc;
	ios->handle = pmod->handle;

	return ios;

}

static long
fhandle_seek(agl_ios * src, long offset, int whence)
{
	long pos;

	agl_rvassert(src != NULL, -1);

	if ((pos = lseek((long) src->priv, offset, whence)) > 0)
		return pos;
	agl_msg("file error: %s", strerror);
	agl_ioserr();
	return -1;
}

static long
fhandle_read(agl_ios * src, void *buf, long size, long maxobj, void *supp)
{
	long nb;
	long count;
	long rew;

	agl_rvassert(src != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	count = size * maxobj;
	if ((nb = read((long) src->priv, buf, count)) < 0)
	{
		agl_msg("file error: %s", strerror);
		agl_ioserr();
		return -1;
	}
	rew = nb % size;
	if (rew)
	{
		if (lseek((long) src->priv, -rew, SEEK_CUR) < 0)
		{
			agl_msg("file error: %s", strerror);
			agl_ioserr();
			return -1;
		}
	}
	return nb / size;
}

static long
fhandle_write(agl_ios * src, void *buf, long size, long maxobj, void *supp)
{
	long nb;
	long count;
	long rew;

	agl_rvassert(src != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	count = size * maxobj;
	if ((nb = write((long) src->priv, buf, count)) < 0)
	{
		agl_msg("file error: %s", strerror);
		agl_ioserr();
		return -1;
	}
	rew = nb % size;
	if (rew)
	{
		if (lseek((long) src->priv, -rew, SEEK_CUR) < 0)
		{
			agl_msg("file error: %s", strerror);
			agl_ioserr();
			return -1;
		}
	}
	return nb / size;
}

static int
fhandle_close(agl_ios * src)
{
	agl_rvassert(src != NULL, -1);

	agl_free(src->filename);
	agl_free(src);
	return 0;
}

static agl_iosio agl_fhandle_iosio = {
	"HANDLE",
	fhandle_is,
	fhandle_open,
	fhandle_seek,
	fhandle_read,
	fhandle_write,
	fhandle_close,
};


agl_module agl_modtab_fhandle = {
	"HANDLE",
	"loader for Handle",
	AGL_MOD_IO,
	{{NULL, NULL}},
	(void *) &agl_fhandle_iosio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
