
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "config.h"
#include "agl_private.h"

static int
file_is(agl_module * pmod, char *name, agl_ios * parent)
{
	agl_rvassert(pmod != NULL, -1);
	agl_rvassert(name != NULL, -1);
	agl_rvassert(parent == NULL, -1);

	return 1;
}

static agl_ios *
file_open(agl_module * pmod, char *name, agl_ios * parent)
{
	agl_ios *ios;
	FILE *stream;
	char *p, *mode, *pname;

	agl_rvassert(pmod != NULL, NULL);
	agl_rvassert(name != NULL, NULL);
	agl_rvassert(parent == NULL, NULL);

	mode = NULL;
	p = strchr(name, ':');

	if (p == NULL)
		stream = fopen(name, "rb+");
	else
	{
		if ((pname = agl_malloc(p - name + 1)) == NULL)
		{
			agl_memerr();
			return NULL;
		}
		strncpy(pname, name, p - name);
		mode = p + 1;
		stream = fopen(pname, mode);
		agl_free(pname);
	}
	if (stream == NULL)
	{
		agl_error("cannot open file");
		return NULL;
	}
	if ((ios = agl_malloc(sizeof(agl_ios))) == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if ((ios->filename = agl_malloc(strlen(name) + 1)) == NULL)
	{
		agl_memerr();
		agl_free(ios);
		return NULL;
	}
	strcpy(ios->filename, name);

	ios->priv = (void *) stream;
	ios->pio = (agl_iosio *) pmod->priv_desc;
	ios->handle = pmod->handle;

	return ios;

}


static long
file_seek(agl_ios * src, long offset, int whence)
{
	FILE *file;

	agl_rvassert(src != NULL, -1);

	file = (FILE *) src->priv;

	agl_rvassert(file != NULL, -1);

	if (fseek(file, offset, whence) == 0)
		return ftell(file);
	agl_msg("file error: %s", strerror);
	agl_ioserr();
	return -1;
}

static long
file_read(agl_ios * src, void *buf, long size, long maxobj, void *supp)
{
	long nb;
	FILE *file;

	agl_rvassert(src != NULL, -1);

	file = (FILE *) src->priv;

	agl_rvassert(file != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	if ((nb = fread(buf, size, maxobj, file)) >= 0)
		return nb;
	agl_msg("file error: %s", strerror);
	agl_ioserr();
	return -1;
}

static long
file_write(agl_ios * src, void *buf, long size, long maxobj, void *supp)
{
	long nb;
	FILE *file;

	agl_rvassert(src != NULL, -1);

	file = (FILE *) src->priv;

	agl_rvassert(file != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	if ((nb = fwrite(buf, size, maxobj, file)) >= 0)
		return nb;
	agl_msg("file error: %s", strerror);
	agl_ioserr();
	return -1;
}

static int
file_close(agl_ios * src)
{
	agl_rvassert(src != NULL, -1);

	fclose((FILE *) src->priv);
	agl_free(src->filename);
	agl_free(src);
	return 0;
}

static agl_iosio agl_file_iosio = {
	"FILE",
	file_is,
	file_open,
	file_seek,
	file_read,
	file_write,
	file_close,
};

agl_module agl_modtab_file = {
	"FILENAME",
	"loader for filename",
	AGL_MOD_IO,
	{{NULL, NULL}},
	(void *) &agl_file_iosio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
