
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#include <string.h>

#include "config.h"
#include "agl_private.h"


static int
fnull_is(agl_module * pmod, char *name, agl_ios * parent)
{
	agl_rvassert(pmod != NULL, -1);
	agl_rvassert(name != NULL, -1);
	agl_rvassert(parent == NULL, -1);

	return 1;
}

static agl_ios *
fnull_open(agl_module * pmod, char *name, agl_ios * parent)
{
	agl_ios *ios;

	agl_rvassert(pmod != NULL, NULL);
	agl_rvassert(name != NULL, NULL);
	agl_rvassert(parent == NULL, NULL);

	if ((ios = agl_malloc(sizeof(agl_ios))) == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if ((ios->filename = agl_malloc(1)) == NULL)
	{
		agl_memerr();
		agl_free(ios);
		return NULL;
	}
	ios->filename = '\0';
	ios->priv = (void *) NULL;
	ios->pio = (agl_iosio *) pmod->priv_desc;
	ios->handle = pmod->handle;

	return ios;

}


static long
fnull_seek(agl_ios * src, long offset, int whence)
{
	agl_rvassert(src != NULL, -1);

	offset = whence;

	return 0;
}

static long
fnull_rdwr(agl_ios * src, void *buf, long size, long maxobj, void *supp)
{
	agl_rvassert(src != NULL, -1);
	buf = supp = NULL;
	size = maxobj = 0;

	return 0;
}

static int
fnull_close(agl_ios * src)
{
	agl_rvassert(src != NULL, -1);

	agl_free(src->filename);

	agl_free(src);
	return 0;
}

static agl_iosio agl_fnull_iosio = {
	"NULL",
	fnull_is,
	fnull_open,
	fnull_seek,
	fnull_rdwr,
	fnull_rdwr,
	fnull_close,
};

agl_module agl_modtab_fnull = {
	"NULL",
	"loader for nothing",
	AGL_MOD_IO,
	{{NULL, NULL}},
	(void *) &agl_fnull_iosio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
