
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! This file manages inputs from all the medias used by adgali

/*! \file
 * \ingroup ios
 * This file manges the input and output from various medias into adgali.
 * yet, it manages file stream (files), file ios, memory zones
 * The output is not implemented.
 * \todo sockets ios (TCP and UDP) should be implemented in the future
 */


#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "config.h"
#include "agl_private.h"

static int
fstream_is(agl_module * pmod, char *name, agl_ios * parent)
{
	long stream;
	char *p;

	agl_rvassert(pmod != NULL, -1);
	agl_rvassert(name != NULL, -1);
	agl_rvassert(parent == NULL, -1);

	p = agl_getnum(name, &stream);
	if (p == NULL)
	{
		agl_error("STREAM: unable to get stream desc");
		return -1;
	}
	return 1;
}

static agl_ios *
fstream_open(agl_module * pmod, char *name, agl_ios * parent)
{
	long stream;
	char *p;
	agl_ios *ios;

	agl_rvassert(pmod != NULL, NULL);
	agl_rvassert(name != NULL, NULL);
	agl_rvassert(parent == NULL, NULL);

	p = agl_getnum(name, &stream);
	if (p == NULL)
	{
		agl_error("STREAM: unable to get stream desc");
		return NULL;
	}

	if ((ios = agl_malloc(sizeof(agl_ios))) == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if ((ios->filename = agl_malloc(p - name + 1)) == NULL)
	{
		agl_memerr();
		agl_free(ios);
		return NULL;
	}
	strncpy(ios->filename, name, p - name);
	ios->priv = (void *) stream;
	ios->pio = (agl_iosio *) pmod->priv_desc;
	ios->handle = pmod->handle;

	return ios;

}


static long
fstream_seek(agl_ios * src, long offset, int whence)
{
	FILE *stream;

	agl_rvassert(src != NULL, -1);

	stream = (FILE *) src->priv;

	agl_rvassert(stream != NULL, -1);

	if (fseek(stream, offset, whence) == 0)
		return ftell(stream);
	agl_msg("file error: %s", strerror);
	agl_ioserr();
	return -1;
}

static long
fstream_read(agl_ios * src, void *buf, long size, long maxobj, void *supp)
{
	long nb;
	FILE *stream;

	agl_rvassert(src != NULL, -1);

	stream = (FILE *) src->priv;

	agl_rvassert(stream != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	if ((nb = fread(buf, size, maxobj, stream)) >= 0)
		return nb;
	agl_msg("file error: %s", strerror);
	agl_ioserr();
	return -1;
}

static long
fstream_write(agl_ios * src, void *buf, long size, long maxobj, void *supp)
{
	long nb;
	FILE *stream;

	agl_rvassert(src != NULL, -1);

	stream = (FILE *) src->priv;

	agl_rvassert(stream != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	if ((nb = fwrite(buf, size, maxobj, stream)) >= 0)
		return nb;
	agl_msg("file error: %s", strerror);
	agl_ioserr();
	return -1;
}

static int
fstream_close(agl_ios * src)
{
	agl_rvassert(src != NULL, -1);

	agl_free(src->filename);
	agl_free(src);
	return 0;
}

static agl_iosio agl_fstream_iosio = {
	"STREAM",
	fstream_is,
	fstream_open,
	fstream_seek,
	fstream_read,
	fstream_write,
	fstream_close,
};

agl_module agl_modtab_fstream = {
	"STREAM",
	"loader for Streams",
	AGL_MOD_IO,
	{{NULL, NULL}},
	(void *) &agl_fstream_iosio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
