
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/


#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"

//! \warning this code assumes that the resulting string is lesser than 1000 character

static char *
agl_iosvarsubs(agl_handle handle, char *address)
{
	char *str, *pt;
	char *var, *pout, *pcur;
	agl_object *pobj;
	agl_elemtype *pelem;
	int sz;

	pout = pcur = (char *) agl_malloc(1000);
	if (pout == NULL)
	{
		agl_memerr();
		return NULL;
	}
	str = address;
	while ((pt = strchr(str, '$')) != NULL)
	{
		strncpy(pcur, pt, pt - str);
		pcur += pt - str;
		if (*(++pt) == '\0')
		{
			agl_error("$ at end of string");
			agl_free(pout);
			return NULL;
		}
		if ((str = strpbrk(pt, "$/:@?")) == NULL)
			sz = strlen(pt);
		else
			sz = str - pt;
		if (sz == 0)
		{
			agl_error("$ without a variable name");
			agl_free(pout);
			return NULL;
		}
		var = (char *) agl_malloc(sz);
		if (var == NULL)
		{
			agl_memerr();
			agl_free(pout);
			return NULL;
		}
		strncpy(var, pt, sz);
		if (agl_cacheobjget(var, handle, &pobj) != 1)
		{
			agl_error2("variable %s not found or error", var);
			agl_free(var);
			agl_free(pout);
			return NULL;
		}
		agl_free(var);
		if (strcmp(pobj->pmod->name, "ELEM") != 0)
		{
			agl_error2("variable %s is not of type ELEM", var);
			agl_free(pout);
			return NULL;
		}
		pelem = (agl_elemtype *) pobj->priv;
		switch (pelem->vtype)
		{
		case AGL_STYPESTR:
			strcpy(pcur, pelem->vu.str);
			break;
		case AGL_STYPEDBL:
			sprintf(pcur, "%g", pelem->vu.dbl);
			break;
		case AGL_STYPEINT:
			sprintf(pcur, "%ld", pelem->vu.num);
			break;
		default:
			agl_error2("variable %s is from an unknown type", var);
			agl_free(pout);
			return NULL;

		}
		pcur += strlen(pcur);
	}
	strcpy(pcur, str);

	agl_rvassert(strlen(pout) < 1000, NULL);

	return pout;
}

agl_ios *
agl_ios_open(agl_container * handle, char *address)
{
	int j, last;
	char *p, *str, *ptype, *pdefault;
	agl_ios *pio, *npio;
	agl_module *ptr;
	agl_iosio *piofn;

	j = 0;
	last = 0;
	pdefault = "FILE";
	pio = NULL;
	str = agl_iosvarsubs(handle, address);
	if (str == NULL)
	{
		agl_error("unable to analyse address");
		return NULL;
	}
	while (str[j])
	{
		// search for a character not present in a protocol type
		if ((p = strpbrk(str + j, "?:")) == NULL)
			ptype = pdefault;
		else
		{
			// the first found is not a protocol termination 
			// we assume the name is a default loader address
			if (*p != ':')
				ptype = pdefault;
			else
			{
				// else we're controlling it's written like
				// FILE://adress:/path/to/file
				p = strstr(str + j, "://");
				if (p == NULL)
					ptype = pdefault;
				else
				{
					ptype = (char *) agl_malloc(p - str - j + 1);
					if (ptype == NULL)
					{
						agl_memerr();
						agl_free(str);
						if (pio != NULL)
							agl_ios_close(pio);
						return NULL;
					}
					strncpy(ptype, str + j, p - str - j);
					j = p - str + 3;	// add 3 to pass the :// string
				}
			}
		}
		if (ptype == pdefault)
		{
			ptype = (char *) agl_malloc(strlen(pdefault) + 1);
			if (ptype == NULL)
			{
				agl_memerr();
				agl_free(str);
				if (pio != NULL)
					agl_ios_close(pio);
				return NULL;
			}
			strcpy(ptype, pdefault);
		}
		// search loader
		p = ptype;
		while (*p)
		{
			*p = toupper(*p);
			p++;
		}

		piofn = NULL;
		for (ptr = handle->moduleroot; ptr; ptr = ptr->pnext)
		{
			if (ptr->type == AGL_MOD_IO)
			{
				piofn = ptr->priv_desc;
				if (strcmp(piofn->desc, ptype) == 0)
					break;
			}
		}
		if (piofn == NULL)
		{
			agl_error2("loader %s not found", ptype);
			if (ptype != pdefault)
				agl_free(ptype);
			agl_free(str);
			if (pio != NULL)
				agl_ios_close(pio);
			return NULL;
		}
		if (ptype != pdefault)
			agl_free(ptype);
		p = strchr(str + j, '?');
		if (p == NULL)
			ptype = (char *) agl_malloc(strlen(str + j) + 1);
		else
			ptype = (char *) agl_malloc(p - str - j + 1);
		if (ptype == NULL)
		{
			agl_memerr();
			agl_free(str);
			if (pio != NULL)
				agl_ios_close(pio);
			return NULL;
		}
		if (p == NULL)
			strcpy(ptype, str + j);
		else
			strncpy(ptype, str + j, p - str - j);
		// here: the loader is found
		if ((*piofn->is) (ptr, ptype, pio) < 0)
		{
			agl_error2("bad file type: %s", str);
			agl_free(str);
			agl_free(ptype);
			if (pio != NULL)
				agl_ios_close(pio);
			return NULL;
		}
		if ((npio = (*piofn->open) (ptr, ptype, pio)) == NULL)
		{
			agl_error("unable to open file");
			agl_free(str);
			agl_free(ptype);
			if (pio != NULL)
				agl_ios_close(pio);
			return NULL;
		}
		pio = npio;
		if (p == NULL)
			j = strlen(str);
		else
			j = p - str + 1;
		agl_free(ptype);
	}
	agl_free(str);
	return pio;
}
