
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* this file interfaces adgali with the underlying graphical layer */

#include "config.h"
#include "agl_private.h"

#ifdef HAVE_SDL
#include "SDL.h"

struct agl_sdlstruct
{
	SDL_Surface *agl_sdlsurface;
	SDL_Surface *agl_videosurface;
};

static SDL_Surface *
agl_sublayercreate(int width, int height, unsigned long **pixels)
{
	SDL_Surface *surface;

	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, 32, AGL_RED, AGL_GREEN, AGL_BLUE, AGL_ALPHA);
	if (surface == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if (pixels != NULL)
		*pixels = (unsigned long *) surface->pixels;
	return (surface);
}

static int
agl_sublayerfree(void *ptr)
{
	agl_rvassert(ptr != NULL, -1);

	SDL_FreeSurface((SDL_Surface *) ptr);
	return 1;
}

int
agl_sublayerinit(agl_handle handle)
{
	struct agl_sdlstruct *psdl;


	agl_rvassert(handle != NULL, -1);

	if (SDL_Init(SDL_INIT_EVERYTHING | SDL_INIT_NOPARACHUTE) < 0)
	{
		agl_error("can't init sublayer");
		return -1;
	}

	psdl = (struct agl_sdlstruct *) agl_malloc(sizeof(struct agl_sdlstruct));
	if (psdl == NULL)
	{
		agl_memerr();
		return -1;
	}

	psdl->agl_sdlsurface = agl_sublayercreate(0, 0, NULL);
	if (psdl->agl_sdlsurface == NULL)
	{
		agl_error("can't init sublayer");
		agl_free(psdl);
		return -1;
	}
	SDL_ShowCursor(0);

	handle->vidstruct = psdl;

	return 1;
}

int
agl_sublayerexit(agl_handle handle)
{
	struct agl_sdlstruct *psdl;
	int ret;

	agl_rvassert(handle != NULL, -1);

	psdl = (struct agl_sdlstruct *) handle->vidstruct;

	agl_rvassert(psdl->agl_sdlsurface != NULL, -1);

	psdl->agl_sdlsurface->pixels = NULL;
	psdl->agl_sdlsurface->pitch = 0;
	psdl->agl_sdlsurface->w = 0;
	psdl->agl_sdlsurface->h = 0;

	ret = agl_sublayerfree(psdl->agl_sdlsurface);

	if (psdl->agl_videosurface != NULL)
		ret = agl_sublayerfree(psdl->agl_videosurface) < 0 ? -1 : ret;

	SDL_Quit();
	agl_free(psdl);
	if (handle->modes_h)
	{
		agl_free(handle->modes_h);
		handle->modes_h = NULL;
	}
	if (handle->modes_w)
	{
		agl_free(handle->modes_w);
		handle->modes_w = NULL;
	}

	return ret;
}

int
agl_sublayergetvidsz(agl_handle handle, int flags, int **h, int **w)
{
	SDL_Rect **modes;
	int sdflags;
	int i, nb;

	agl_rvassert(handle != NULL, -1);

	if (handle->modes_h != NULL)
		agl_free(handle->modes_h);
	if (handle->modes_w != NULL)
		agl_free(handle->modes_w);

	*h = handle->modes_h = NULL;
	*w = handle->modes_w = NULL;

	sdflags = SDL_HWSURFACE;
	if (flags & AGL_FULLSCREEN)
		sdflags |= SDL_FULLSCREEN;

	modes = SDL_ListModes(NULL, sdflags);

	if (modes == (SDL_Rect **) 0)
		return 0;
	if (modes == (SDL_Rect **) - 1)
		return 1;
	for (nb = 0; modes[nb]; nb++) ;
	*h = handle->modes_h = (int *) agl_malloc(sizeof(int) * (nb + 1));
	if (*h == NULL)
	{
		agl_memerr();
		return -1;
	}
	*w = handle->modes_w = (int *) agl_malloc(sizeof(int) * (nb + 1));
	if (*w == NULL)
	{
		agl_memerr();
		return -1;
	}

	for (i = 0; i < nb; i++)
	{
		handle->modes_h[i] = modes[i]->h;
		handle->modes_w[i] = modes[i]->w;
	}
	handle->modes_h[nb] = 0;
	handle->modes_w[nb] = 0;

	return 1;
}

int
agl_sublayersetvid(agl_handle handle, int width, int height, int flag)
{
	struct agl_sdlstruct *psdl;
	int sdflags;

	agl_rvassert(handle != NULL, -1);

	psdl = (struct agl_sdlstruct *) handle->vidstruct;

	agl_rvassert(psdl->agl_sdlsurface != NULL, -1);


#ifdef SA
	{
		SDL_VideoInfo *info;

		info = SDL_GetVideoInfo();
		printf("hw_available:%d\nwm_availaible:%d\nblit_hw:%d\nblit_hw_CC:%d\nblit_hw_A:%d\n", info->hw_available,
		       info->wm_available, info->blit_hw, info->blit_hw_CC, info->blit_hw_A);
		printf("blit_sw:%d\nblit_sw_CC:%d\nblit_sw_A:%d\nblit_fill:%d\nvideo_mem:%d\nbit_per_pixel:%d\n", info->blit_sw,
		       info->blit_sw_CC, info->blit_sw_A, info->blit_fill, info->video_mem, info->vfmt->BitsPerPixel);
	}
#endif

	sdflags = SDL_HWSURFACE;
	if (flag & AGL_FULLSCREEN)
		sdflags |= SDL_FULLSCREEN;
	psdl->agl_videosurface = SDL_SetVideoMode(width, height, 32, sdflags);
	if (psdl->agl_videosurface == NULL)
	{
		agl_fatal("can't init video surface");
		return -1;
	}

	if (psdl->agl_videosurface->flags & SDL_FULLSCREEN)
		handle->flags |= AGL_FULLSCREEN;

	handle->vid_w = psdl->agl_videosurface->w;
	handle->vid_h = psdl->agl_videosurface->h;

	return 1;
}

int
agl_sublayerblit(agl_handle handle, agl_pix * pix, agl_rect * srcrect, agl_rect * dstrect, agl_pos * pos)
{
	SDL_Rect src, dst;
	short nw, nh, sw, sh;
	short sx, sy, dx, dy, ex, ey, dw, dh, bx, by;
	int ret;
	struct agl_sdlstruct *psdl;


	agl_rvassert(handle != NULL, -1);

	psdl = (struct agl_sdlstruct *) handle->vidstruct;

	agl_rvassert(pix != NULL, -1);
	agl_rvassert(pos != NULL, -1);
	agl_rvassert(psdl->agl_sdlsurface != NULL, -1);
	agl_rvassert(psdl->agl_videosurface != NULL, -1);
	/* calculate rectangles. Anyway, SDL rectangle calculation is too basic */
	if (srcrect == NULL)
	{
		sw = pix->size.width;
		sh = pix->size.height;
		sx = 0;
		sy = 0;
	}
	else
	{
		sw = srcrect->width;
		sh = srcrect->height;
		sx = srcrect->x;
		sy = srcrect->y;
	}
	if (dstrect == NULL)
	{
		dw = psdl->agl_videosurface->w;
		dh = psdl->agl_videosurface->h;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = dstrect->width;
		dh = dstrect->height;
		dx = dstrect->x;
		dy = dstrect->y;
	}
	bx = max4(0, pos->x, pos->x - sx, dx);
	ex = min4(psdl->agl_videosurface->w, dx + dw, pos->x + sw, pos->x + pix->size.width - sx);
	nw = ex - bx;
	by = max4(0, pos->y, pos->y - sy, dy);
	ey = min4(psdl->agl_videosurface->h, dy + dh, pos->y + sh, pos->y + pix->size.height - sy);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return -1;

	src.x = sx + bx - pos->x;
	src.y = sy + by - pos->y;
	src.w = nw;
	src.h = nh;

	dst.x = bx;
	dst.y = by;
	dst.w = nw;
	dst.h = nh;

	psdl->agl_sdlsurface->pixels = pix->pixels;
	psdl->agl_sdlsurface->pitch = pix->size.width * 4;
	psdl->agl_sdlsurface->w = pix->size.width;
	psdl->agl_sdlsurface->h = pix->size.height;
	if ((ret = SDL_LowerBlit(psdl->agl_sdlsurface, &src, psdl->agl_videosurface, &dst)) < 0)
		return ret;
	SDL_UpdateRect(psdl->agl_videosurface, dst.x, dst.y, dst.w, dst.h);

	return 1;
}


#endif
