%{
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! This file is the parser
/*! \file
 * \ingroup module
 * here is the file parser. This parser DO NOT create the objects
 * described in the file to scan but generates a tokenised string
 * that is like some machine code. The machine code in a second pass
 * is interpreted to create objects.
 * It's a better structure for easy futures extensions like tests
 * and loops.
 */


#include <stdlib.h>

#include "config.h"
#include "agl_private.h"
#include "agl_gramint.h"


#define YYDEBUG	1
/*int agldebug = 1 */

// make a local copy of parser

#define yymaxdepth	aglmaxdepth
#define yyparse	aglparse
#define yylex	agllex
#define yyerror	aglerror
#define yylval	agllval
#define yychar	aglchar
#define yydebug	agldebug
#define yypact	aglpact
#define yyr1	aglr1
#define yyr2	aglr2
#define yydef	agldef
#define yychk	aglchk
#define yypgo	aglpgo
#define yyact	aglact
#define yyexca	aglexca
#define yyerrflag	aglerrflag
#define yynerrs	aglnerrs
#define yyps	aglps
#define yypv	aglpv
#define yys	agls
#define yy_yys	aglyys
#define yystate	aglstate
#define yytmp	agltmp
#define yyv	aglv
#define yy_yyv	aglyyv
#define yyval	aglval
#define yylloc	agllloc
#define yyreds	aglreds
#define yytoks	agltoks
#define yylhs	aglyylhs
#define yylen	aglyylen
#define yydefred	aglyydefred
#define yydgoto	aglyydgoto
#define yysindex	aglyysindex
#define yyrindex	aglyyrindex
#define yygindex	aglyygindex
#define yytable	aglyytable
#define yycheck	aglyycheck
#define yyname	aglyyname
#define yyrule	aglyyrule

#define YYPARSE_PARAM	analyse	
#define YYLEX_PARAM	analyse
#define YYLSP_NEEDED	0 // not needed
#define YYMAXDEPTH	0 // defined by yacc
#define aglerror(str) agl_error3("grammar: %s at line %d", str, ((agl_analyse *)analyse)->lineno)

#define encodei(s) { agl_graminstr g; g.type = AGL_GRAMINSTR; g.u.instr = s; agl_gramencode(analyse, &g); }
#define encodei(s) { agl_graminstr g; g.type = AGL_GRAMINSTR; g.u.instr = s; agl_gramencode(analyse, &g); }
#define encodev(s) { agl_graminstr g; g.type = AGL_GRAMVAL; g.u.val = s; agl_gramencode(analyse, &g); }
#define encodek(s) { agl_graminstr g; g.type = AGL_GRAMKEYW; g.u.keyw = s; agl_gramencode(analyse, &g); }
#define encodem(s) { agl_graminstr g; g.type = AGL_GRAMMOD; g.u.mod = s; agl_gramencode(analyse, &g); }
#define encoded(s) { agl_graminstr g; g.type = AGL_GRAMDEFN; g.u.defn = s; agl_gramencode(analyse, &g); }
#define encodes(s) { agl_graminstr g; g.type = AGL_GRAMSYMB; g.u.symb = s; agl_gramencode(analyse, &g); }
#define encodel(s) { agl_graminstr g; g.type = AGL_GRAMLSTSZ; g.u.lstsz = s; agl_gramencode(analyse, &g); }
#define encodevoid() { agl_graminstr g; g.type = AGL_GRAMVAL; g.u.val.itype = g.u.val.vtype = AGL_GRAMNONE; g.u.val.pnext = NULL; agl_gramencode(analyse, &g); }
%}

%union {
	agl_elemtype	val;
	agl_gramsymb	*sym;
	agl_moddefn	*defn;
	agl_modelem	*keyw;
	agl_module	*mod;
	agl_graminstr	*inst;
	int	listsz;
}

%{
int agllex(YYSTYPE *pval, agl_analyse *analyse);
%}

%pure_parser

%start file

%token <sym> IDENT
%token <val> STRING
%token <val> NUMBER
%token <val> FNUMBER
%token <keyw> KEYW
%token <defn> DEFN
%token <mod> MOD

%type <inst> file stmt stmt2 section affect values list stbeg preindices
%type <listsz> indices indices2 indelems listelems decllist listvalue indvalue

%right	'='
%left	"=>"

%%

file: /*empty */ { encodei(GRAM_NONE); }
	| list	{ encodei(GRAM_NONE); };
/*	| error 	{yyerrok;} */

list: list stmt
	| stmt;


stmt:	stmt2 section { ((agl_analyse *)analyse)->curmodule = NULL; encodei(agl_keywend); };

stmt2: stbeg "LIKE" IDENT { encodei (agl_keywlike); encodes($3); }
     | MOD IDENT "ADD" { ((agl_analyse *)analyse)->curmodule = $1; encodei(agl_keywadd); encodem($1); encodes($2); }
     | stbeg;

stbeg:	MOD IDENT { ((agl_analyse *)analyse)->curmodule = $1; encodei(agl_keywstmt2); encodem($1); encodes($2); };

section: '{' decllist '}' { }
	| '{' '}' { };

decllist: decllist affect { $$ = $1 + 1; }
	| affect { $$ = 1; };

affect:	KEYW '=' values { encodei(agl_keywassign); encodek($1); }
	| KEYW preindices '=' values { encodei(agl_keywindex); encodei(agl_keywassign); encodek($1); };

preindices: indices { encodei(agl_indicescrnpush); encodel($1); };

indices: indices indices2 { $$ = $1 + $2; }
       	| indices2 { $$ = $1; };

indices2: '[' indelems ']' { $$ = $2; };

indelems: indelems ',' indvalue { $$ = $1 + $3; }
	| indvalue { $$ = $1; };

indvalue: values { $$ = 1; }
	| { $$ = 1; encodevoid();};

values: DEFN { encodei(agl_valpush); encoded($1); encodei(agl_ieval); }
	| NUMBER { encodei(agl_valpush); encodev($1); }
	| STRING {encodei(agl_valpush); encodev($1); }
	| FNUMBER {encodei(agl_valpush); encodev($1); }
	| IDENT { encodei(agl_valpush); encodes($1); encodei(agl_ieval); }
	|'{' listelems '}' { encodei(agl_listcrnpush); encodel($2); };

listelems: listelems ',' listvalue { $$ = $1 + $3; }
	| listvalue { $$ = $1; };

listvalue: {$$ = 1; encodevoid();}
	| values {$$ = 1; }
	| values "=>" values {$$ = 1; encodei(agl_gramhash); };

