
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! manipulation of channels of a pixmap

/*!\file agl_channel.c
 * \author Stephane APIOU
 * \ingroup video
 * this file manipulates channels and the pixmaps associated
 */
#include <string.h>
#include "config.h"
#include "agl_private.h"

#define CHNL_CHANNEL	0
#define CHNL_IMGCHNL	1

static int
agl_channelload(agl_object * pobj)
{
	agl_image *channel;
	agl_channel *chan;
	agl_ios *ops;
	agl_elemtype elem;
	int i, ch;

	agl_rvassert((pobj != NULL), -1);

	if (pobj->flags & AGL_OBJ_LOADED)
		return 1;

	if ((pobj->flags & AGL_OBJ_VALID) == 0)
		return 0;

	channel = NULL;

	ch = -1;
	for (i = 0; i < pobj->defmax; i++)
	{
		if (agl_getelem(pobj->pmod->handle, pobj->defelem + i, &elem) < 0)
		{
			agl_error("could'nt get elemtype value");
			goto error;
		}
		switch (pobj->defidx[i])
		{
		case CHNL_CHANNEL:
			switch (elem.vtype)
			{
			case AGL_STYPESTR:
				ops = agl_ios_open(pobj->pmod->handle, elem.vu.str);
				if (ops == NULL)
				{
					agl_error2("unable to load image: %s", elem.vu.str);
					goto error;
				}
				channel = agl_imgioload(ops);
				agl_ios_close(ops);
				if (channel == NULL)
				{
					agl_error2("unable to load alpha: %s", elem.vu.str);
					goto error;
				}
				break;
			case AGL_STYPEUSER:
				switch (agl_typeidx(elem.vu.user, 2, "IMAGE", "CHANNEL"))
				{
				case 1:	/* IMAGE */
					channel = agl_imgdup((agl_image *) elem.vu.user);
					if (channel == NULL)
					{
						agl_error("unable to dup image");
						goto error;
					}
					agl_imgfree(channel);
					break;
				case 2:	/* CHANNEL */
					chan = agl_chandup((agl_channel *) elem.vu.user);
					if (chan == NULL)
					{
						agl_error("unable to dup channel");
						goto error;
					}
					break;
				default:
					agl_error("unexpected type");
					goto error;
				}
			default:
				agl_error("invalid object reference");
				goto error;
			}
			break;
		case CHNL_IMGCHNL:
			if (elem.vtype == AGL_STYPEINT)
				ch = elem.vu.num;
			else
			{
				agl_error("IMAGE_CHANNEL is not a number");
				goto error;
			}
			break;
		default:
			agl_error2("parameter %d not known", pobj->defidx[i]);
			goto error;
		}
	}

	if (channel != NULL)
	{
		if (chan != NULL)
		{
			agl_error("image defined where a channel is already loaded");
			goto error;
		}
		switch (ch)
		{
		case 0:
			if (agl_frompixmap(AGL_RGB, NULL, channel->pix, NULL, &chan, NULL, NULL) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 1:
			if (agl_frompixmap(AGL_RGB, NULL, channel->pix, NULL, NULL, &chan, NULL) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 2:
			if (agl_frompixmap(AGL_RGB, NULL, channel->pix, NULL, NULL, NULL, &chan) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 3:
			if (agl_frompixmap(AGL_CMY, NULL, channel->pix, NULL, &chan, NULL, NULL) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 4:
			if (agl_frompixmap(AGL_CMY, NULL, channel->pix, NULL, NULL, &chan, NULL) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 5:
			if (agl_frompixmap(AGL_CMY, NULL, channel->pix, NULL, NULL, NULL, &chan) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 6:
			if (agl_frompixmap(AGL_HSV, NULL, channel->pix, NULL, &chan, NULL, NULL) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 7:
			if (agl_frompixmap(AGL_HSV, NULL, channel->pix, NULL, NULL, &chan, NULL) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		case 8:
			if (agl_frompixmap(AGL_HSV, NULL, channel->pix, NULL, NULL, NULL, &chan) != 1)
			{
				agl_error("unable to load channel");
				goto error;
			}
			break;
		default:
			agl_error("invalid value for IMAGE_CHANNEl field");
			goto error;
		}
		agl_imgfree(channel);
	}
	if (chan == NULL)
		return -1;
	return 1;
      error:
	if (chan != NULL)
		agl_chanfree(chan);
	if (channel != NULL)
		agl_imgfree(channel);
	return -1;

}

//! treats events

/*! an object receives an event. we must treat it
 * \param pcur a pointer to the current object
 * \param event the event that occurs (please consult the defines)
 * \param name the name of the object concerned by 
 * \param pobj a pointer to the object concerned by or 
 * \return
 * - -1 if a fatal error occurs
 * - 0 if it fails but without fatal error
 * - 1 if success
 */


static int
agl_channeldepevt(agl_object * pcur, int event, char *name, agl_object * pobj)
{
	agl_rvassert(pcur != NULL, -1);
	agl_rvassert(pobj != NULL, -1);
	agl_rvassert(name != NULL, -1);

	switch (event)
	{
	case AGL_OBJEVT_FLUSHED:	// nothing to do this is a leaf object
		break;
	case AGL_OBJEVT_UPDATE:
		if (agl_channelload(pcur) != 1)
		{
			agl_error("unable to load channel");
			return -1;
		}
		break;
	case AGL_OBJEVT_UNLOAD:
		agl_chanfree(pcur->priv);
		pcur->priv = NULL;
		pcur->flags &= ~AGL_OBJ_LOADED;
		break;
	default:
		agl_error2("event %d not catched by the module", event);
		return 0;
	}
	return 1;
}

static agl_modelem agl_chnlmodtab[] = {
	{"CHANNEL", CHNL_CHANNEL, NULL, AGL_STYPEOBJREF},
	{"IMAGE_CHANNEL", CHNL_IMGCHNL, NULL, AGL_STYPEINT},
	{NULL, 0, NULL, 0}
};

static agl_moddefn agl_chnldefntab[] = {
	DEFN_DECLINTVAL("RED", 0),
	DEFN_DECLINTVAL("GREEN", 1),
	DEFN_DECLINTVAL("BLUE", 2),
	DEFN_DECLINTVAL("CYAN", 3),
	DEFN_DECLINTVAL("MAGENTA", 4),
	DEFN_DECLINTVAL("YELLOW", 5),
	DEFN_DECLINTVAL("HUE", 6),
	DEFN_DECLINTVAL("SATURATION", 7),
	DEFN_DECLINTVAL("VALUE", 8),
	DEFN_DECLEND
};


agl_module agl_modtab_channelmod = {
	"CHANNEL",
	"Image channel",
	AGL_MOD_MOD,
	{{NULL, NULL}},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	agl_channeldepevt,
	NULL,
	NULL,
	"0|(12)",
	agl_chnlmodtab,
	agl_chnldefntab,
	NULL,
	NULL
};
