
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image loading module for various file types

/*! \file
 * \ingroup video
 * load various images file types into an image struct
 * internals.
 */


#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"

#define IMG_IMAGE	0
#define IMG_CHANNEL	1
#define IMG_ALPHA	2
#define IMG_KEYCOLOR	3


static int
agl_imgload(agl_object * pobj)
{
	agl_image *image, *alpha;
	agl_channel *alchan, *ch[3];
	agl_ios *ops;
	agl_elemtype elem, subelem, *psubelem;
	char *alstr;
	int i, j, chtype;
	int ret;

	agl_rvassert((pobj != NULL), -1);

	if (pobj->flags & AGL_OBJ_LOADED)
		return 1;

	if ((pobj->flags & AGL_OBJ_VALID) == 0)
		return 0;

	agl_rvassert(pobj->defelem != NULL, -1);

	image = alpha = NULL;
	ch[1] = ch[2] = ch[3] = NULL;
	alstr = NULL;
	chtype = 0;
	ret = -1;

	for (i = 0; i < pobj->defmax; i++)
	{
		if (agl_getelem(pobj->pmod->handle, pobj->defelem + i, &elem) < 0)
		{
			agl_error("could'nt get elemtype value");
			goto error;
		}
		switch (pobj->defidx[i])
		{
		case IMG_IMAGE:
			if (image != NULL)
				agl_imgfree(image);
			switch (elem.vtype)
			{
			case AGL_STYPESTR:
				ops = agl_ios_open(pobj->pmod->handle, elem.vu.str);
				if (ops == NULL)
				{
					agl_error2("unable to load image: %s", elem.vu.str);
					goto error;
				}
				image = agl_imgioload(ops);
				agl_ios_close(ops);
				if (image == NULL)
				{
					agl_error2("unable to load image: %s", elem.vu.str);
					goto error;
				}
				break;
			case AGL_STYPEUSER:
				switch (agl_typeidx(elem.vu.user, 1, "IMAGE"))
				{
				case 1:	/* IMAGE */
					image = agl_imgdup((agl_image *) elem.vu.user);
					if (image == NULL)
					{
						agl_error("unable to dup image");
						goto error;
					}
					break;
				default:
					agl_error("unexpected type");
					goto error;
				}
			default:
				agl_error("invalid object reference");
				goto error;
			}
			break;
		case IMG_ALPHA:
			switch (elem.vtype)
			{
			case AGL_STYPESTR:
				ops = agl_ios_open(pobj->pmod->handle, elem.vu.str);
				if (ops == NULL)
				{
					agl_error2("unable to load image: %s", elem.vu.str);
					goto error;
				}
				alpha = agl_imgioload(ops);
				agl_ios_close(ops);
				if (alpha == NULL)
				{
					agl_error2("unable to load alpha: %s", elem.vu.str);
					goto error;
				}
				if (agl_frompixmap(AGL_RGB, NULL, alpha->pix, NULL, &alchan, NULL, NULL) != 1)
				{
					agl_error2("unable to load alpha: %s", elem.vu.str);
					agl_imgfree(alpha);
					goto error;
				}
				agl_imgfree(alpha);
				alstr = elem.vu.str;
				break;
			case AGL_STYPEUSER:
				switch (agl_typeidx(elem.vu.user, 2, "IMAGE", "CHANNEL"))
				{
				case 1:	/* IMAGE */
					alpha = agl_imgdup((agl_image *) elem.vu.user);
					if (alpha == NULL)
					{
						agl_error("unable to dup image");
						goto error;
					}
					if (agl_frompixmap(AGL_RGB, NULL, alpha->pix, NULL, &alchan, NULL, NULL) != 1)
					{
						agl_error("unable to load alpha:");
						agl_imgfree(alpha);
						goto error;
					}
					agl_imgfree(alpha);
					break;
				case 2:	/* CHANNEL */
					alchan = agl_chandup((agl_channel *) elem.vu.user);
					if (alchan == NULL)
					{
						agl_error("unable to dup channel");
						goto error;
					}
					break;
				default:
					agl_error("unexpected type");
					goto error;
				}
			default:
				agl_error("invalid object reference");
				goto error;
			}
			break;
		case IMG_KEYCOLOR:
			if (agl_pixkey(image->pix, NULL, elem.vu.num) != 1)
			{
				agl_error("unable to apply color key");
				goto error;
			}
			image->color_key = elem.vu.num;
			break;
		case IMG_CHANNEL:
			if (elem.vtype != AGL_STYPELIST)
			{
				agl_error("object is not a list");
				goto error;
			}
			for (psubelem = elem.vu.list; psubelem; psubelem = psubelem->pnext)
			{
				if (agl_getelem(pobj->pmod->handle, psubelem, &subelem))
				{
					agl_error("could'nt get elemtype value");
					goto error;
				}
				agl_rvassert(subelem.itype == AGL_STYPEINT, -1);
				j = subelem.iu.num;
				if (chtype == 0)
					chtype = j / 3 + 1;	// get type of color space
				else if ((j < ((chtype - 1) * 3)) || (j >= (chtype * 3)))
				{
					agl_error("not the same color space");
					goto error;
				}
				if (ch[j % 3] != NULL)
				{
					agl_error("this channel is already loaded");
					goto error;
				}
				switch (subelem.vtype)
				{
				case AGL_STYPESTR:
					ops = agl_ios_open(pobj->pmod->handle, elem.vu.str);
					if (ops == NULL)
					{
						agl_error2("unable to load image: %s", elem.vu.str);
						goto error;
					}
					alpha = agl_imgioload(ops);
					agl_ios_close(ops);
					if (alpha == NULL)
					{
						agl_error2("unable to load channel: %s", elem.vu.str);
						goto error;
					}
					if (agl_frompixmap(AGL_RGB, NULL, alpha->pix, NULL, ch + (j % 3), NULL, NULL) != 1)
					{
						agl_error2("unable to load channel: %s", elem.vu.str);
						agl_imgfree(alpha);
						goto error;
					}
					agl_imgfree(alpha);
					break;
				case AGL_STYPEUSER:
					switch (agl_typeidx(elem.vu.user, 2, "IMAGE", "CHANNEL"))
					{
					case 0:	/* IMAGE */
						alpha = agl_imgdup((agl_image *) elem.vu.user);
						if (alpha == NULL)
						{
							agl_error("unable to dup image");
							goto error;
						}
						if (agl_frompixmap(AGL_RGB, NULL, alpha->pix, NULL, ch + (j % 3), NULL, NULL) != 1)
						{
							agl_error("unable to load channel:");
							agl_imgfree(alpha);
							goto error;
						}
						agl_imgfree(alpha);
						break;
					case 1:	/* CHANNEL */
						ch[j % 3] = agl_chandup((agl_channel *) elem.vu.user);
						if (ch[j % 3] == NULL)
						{
							agl_error("unable to dup channel");
							goto error;
						}
						break;
					default:
						agl_error("unexpected type");
						goto error;
					}
				default:
					agl_error("invalid object reference");
					goto error;
				}
			}
		default:
			agl_error2("parameter %d not known", pobj->defidx[i]);
			goto error;
		}
	}

	if (image == NULL)
	{
		image = agl_imgnew(alchan->size.width, alchan->size.height);
		if (image == NULL)
		{
			agl_error("unable to create image");
			goto error;
		}
	}
	switch (chtype)
	{
	case 1:
		chtype = AGL_RGB;
		break;
	case 2:
		chtype = AGL_CMY;
		break;
	case 3:
		chtype = AGL_HSV;
		break;
	default:
		chtype = AGL_AVG;
	}
	if (image->alpha_name)
		agl_free(image->alpha_name);
	if (alstr != NULL)
	{
		image->alpha_name = (char *) agl_malloc(strlen(alstr) + 1);
		if (image->alpha_name)
		{
			agl_memerr();
			goto error;
		}
		strcpy(image->alpha_name, alstr);
	}
	else
		image->alpha_name = NULL;

	if (agl_topixmap(chtype, NULL, image->pix, alchan, ch[0], ch[1], ch[2]) != 1)
	{
		agl_error("unable to put channels");
		goto error;
	}

	pobj->priv = image;
	pobj->flags |= AGL_OBJ_LOADED;
	ret = 1;

      error:
	if (alchan != NULL)
		agl_chanfree(alchan);
	for (i = 0; i < 3; i++)
		if (ch[i] != NULL)
			agl_chanfree(ch[i]);

	return ret;
}

//! treats events

/*! an object receives an event. we must treat it
 * \param pcur a pointer to the current object
 * \param event the event that occurs (please consult the defines)
 * \param name the name of the object concerned by 
 * \param pobj a pointer to the object concerned by or 
 * \return
 * - -1 if a fatal error occurs
 * - 0 if it fails but without fatal error
 * - 1 if success
 */


static int
agl_imgdepevt(agl_object * pcur, int event, char *name, agl_object * pobj)
{
	agl_rvassert(pcur != NULL, -1);
	agl_rvassert(pobj != NULL, -1);
	agl_rvassert(name != NULL, -1);

	switch (event)
	{
	case AGL_OBJEVT_FLUSHED:
		break;
	case AGL_OBJEVT_UPDATE:
		if (agl_imgload(pcur) < 0)
		{
			agl_error("unable to load img");
			return -1;
		}
		break;
	case AGL_OBJEVT_UNLOAD:
		agl_imgfree(pcur->priv);
		pcur->priv = NULL;
		pcur->flags &= ~AGL_OBJ_LOADED;
		break;
	default:
		agl_error2("event %d not catched by the module", event);
		return 0;
	}
	return 1;
}

static agl_modelem agl_imgmodtab[] = {
	{"IMAGE", IMG_IMAGE, NULL, AGL_STYPESTR | AGL_STYPEOBJREF},
	{"CHANNEL", IMG_CHANNEL, NULL, AGL_STYPESTR | AGL_STYPEOBJREF},
	{"ALPHA", IMG_ALPHA, NULL, AGL_STYPESTR | AGL_STYPEOBJREF},
	{"KEYCOLOR", IMG_KEYCOLOR, NULL, AGL_STYPEINT},
	{NULL, 0, NULL, 0}
};

static agl_moddefn agl_imgdefntab[] = {
	DEFN_DECLINTVAL("RED", 0),
	DEFN_DECLINTVAL("GREEN", 1),
	DEFN_DECLINTVAL("BLUE", 2),
	DEFN_DECLINTVAL("CYAN", 3),
	DEFN_DECLINTVAL("MAGENTA", 4),
	DEFN_DECLINTVAL("YELLOW", 5),
	DEFN_DECLINTVAL("HUE", 6),
	DEFN_DECLINTVAL("SATURATION", 7),
	DEFN_DECLINTVAL("VALUE", 8),
	DEFN_DECLEND
};

agl_module agl_modtab_imgmod = {
	"IMAGE",
	"Bitmap images",
	AGL_MOD_MOD,
	{{NULL, NULL}},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	agl_imgdepevt,
	NULL,
	NULL,
	"0|1|2",
	agl_imgmodtab,
	agl_imgdefntab,
	NULL,
	NULL
};
