
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/*!
 * \file agl_basedraw.c
 *
 * \brief base drawing functions
 *
 *
 * $Id: agl_basedraw.c,v 1.9 2003/06/23 23:58:47 stef Exp $
 *
 * \author Stephane APIOU
 * \date   mar jui 30 01:33:09 CEST 2002
*/

#include "config.h"
#include "agl_private.h"

/*!
 * \brief implements the intersection or union of a source zone and a destination zone
 *
 * \param ptr
 * \param rect
 * \param pos
 * \param  color
 *
 * \return int
*/

int
agl_bdrpixcolr(agl_pix * ptr, agl_rect * rect, agl_pos * pos, unsigned long color)
{
	register unsigned long *pixdst;
	register short nw, nh;
	register short dstrow;
	short dx, dy, ex, ey, dw, dh, bx, by;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(pos != NULL, -1);

	if (rect == NULL)
	{
		dw = ptr->size.width;
		dh = ptr->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = rect->width;
		dh = rect->height;
		dx = rect->x;
		dy = rect->y;
	}
	dstrow = ptr->row_size;

	bx = max(0, dx);
	ex = min(ptr->size.width, dx + dw);
	nw = ex - bx;
	by = max(0, dy);
	ey = min(ptr->size.height, dy + dh);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return 0;
	if ((pos->x < bx) || (pos->x >= ex) || (pos->y < by) || (pos->y >= ey))
		return 0;
	pixdst = ptr->pixels + by * dstrow + bx;
	*pixdst = color;
	return 0;
}

//! internal fast routine. don't verify anything

/*!
 * \param ptr
 * \param y
 * \param ix
 * \param ox
 * \param  unsignedcolor
 *
 * \return void
*/

void
agl_bdrhlinecolr(agl_pix * ptr, short y, short ix, short ox, unsigned long color)
{
	unsigned long *pixdst;
	int i, w;
	short dstrow;

	dstrow = ptr->row_size;
	pixdst = ptr->pixels + y * dstrow + ix;
	w = ox - ix;
	for (i = 0; i < w; i++)
		*pixdst++ = color;
}

//! internal fast routine. don't verify anything

/*!
 * \param ptr
 * \param x
 * \param iy
 * \param oy
 * \param  unsignedcolor
 *
 * \return void
*/

void
agl_bdrvlinecolr(agl_pix * ptr, short x, short iy, short oy, unsigned long color)
{
	unsigned long *pixdst;
	int i, w;
	short dstrow;

	dstrow = ptr->row_size;
	pixdst = ptr->pixels + iy * dstrow + x;
	w = oy - iy;
	for (i = 0; i < w; i++, pixdst += dstrow)
		*pixdst = color;
}

//!

/*!
 * \param ptr
 * \param rect
 * \param ipos
 * \param dpos
 * \param  unsignedcolor
 *
 * \return int
*/

int
agl_bdrlinecolr(agl_pix * ptr, agl_rect * rect, agl_pos * ipos, agl_pos * dpos, unsigned long color)
{
	unsigned long *pixdst;
	short nw, nh, tmp;
	short dstrow;
	short dx, dy, ex, ey, dw, dh, bx, by;
	short adx, ady;
	long trans;
	short cx, cy, ncx, ncy, len, r;
	int i;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(ipos != NULL, -1);

	agl_rvassert(dpos != NULL, -1);

	if (rect == NULL)
	{
		dw = ptr->size.width;
		dh = ptr->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = rect->width;
		dh = rect->height;
		dx = rect->x;
		dy = rect->y;
	}
	dstrow = ptr->row_size;

	bx = max(0, dx);
	ex = min(ptr->size.width, dx + dw);
	nw = ex - bx;
	by = max(0, dy);
	ey = min(ptr->size.height, dy + dh);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return 0;
	if (ipos->x > dpos->x)
	{
		tmp = ipos->x;
		ipos->x = dpos->x;
		dpos->x = tmp;
		tmp = ipos->y;
		ipos->y = dpos->y;
		dpos->y = tmp;
	}
	if (dpos->x <= bx)
		return 0;
	if (ipos->x >= ex)
		return 0;
	if (ipos->y < dpos->y)
	{
		if ((ipos->y >= ey) || (dpos->y <= by))
			return 0;
	}
	else
	{
		if ((dpos->y >= ey) || (ipos->y <= by))
			return 0;
	}
	adx = dpos->x - ipos->x;
	ady = dpos->y - ipos->y;
	trans = ady * (bx - ipos->x) - adx * (by - ipos->y);
	if (ipos->x > bx)
	{
		if (ipos->y <= by)
		{
			ipos->x += ((by - ipos->y) * adx) / ady;
			if (ipos->x >= ex)
				return 0;
			ipos->y = by;
		}
		else if (ipos->y >= ey)
		{
			ipos->x += ((ey - ipos->y) * adx) / ady;
			if (ipos->x >= ex)
				return 0;
			ipos->y = ey;
		}

	}
	else
	{
		if (ipos->y <= by)
		{
			if (trans > 0)
			{
				ipos->x += ((by - ipos->y) * adx) / ady;
				if (ipos->x >= ex)
					return 0;
				ipos->y = by;
			}
			else
			{
				ipos->y = (ady * (bx - ipos->x)) / adx;
				if ((ipos->y <= by) || (ipos->y >= ey))
					return 0;
				ipos->x = bx;
			}
		}
		else if (ipos->y >= ey)
		{
			if (trans > (nh * adx))
			{
				ipos->y = (ady * (bx - ipos->x)) / adx;
				if ((ipos->y <= by) || (ipos->y >= ey))
					return 0;
				ipos->x = bx;
			}
			else
			{
				ipos->x += ((ey - ipos->y) * adx) / ady;
				if (ipos->x >= ex)
					return 0;
				ipos->y = ey;
			}
		}
		else
		{
			ipos->y = (ady * (bx - ipos->x)) / adx;
			if ((ipos->y <= by) || (ipos->y >= ey))
				return 0;
			ipos->x = bx;
		}
	}
	pixdst = ptr->pixels + ipos->y * dstrow + ipos->x;
	if (((ady >= 0) && (adx > ady)) || ((ady < 0) && (adx > -ady)))
	{
		ady++;
		if (ipos->y < dpos->y)
		{
			if (dpos->y < ey)
				ey = dpos->y;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy <= ey)
			{
				for (i = 0; i < len; i++)
					*pixdst++ = color;
				cx = ncx;
				if (cx > ex)
					break;
				len = (r + adx) / ady;
				r = (r + adx) / ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				pixdst += dstrow;
				cy++;
			}
		}
		else
		{
			if (dpos->y > by)
				by = dpos->y;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy >= by)
			{
				for (i = 0; i < len; i++)
					*pixdst++ = color;
				cx = ncx;
				if (cx > ex)
					break;
				len = (r + adx) / ady;
				r = (r + adx) / ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				cy--;
				pixdst -= dstrow;
			}
		}
	}
	else
	{
		adx++;
		if (ipos->y < dpos->y)
		{
			if (dpos->x < ex)
				ex = dpos->x;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy + len;
			if (ncy >= ey)
			{
				ncy = ey;
				len = ncy - cy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				for (i = 0; i < len; i++, pixdst += dstrow)
					*pixdst = color;
				cy = ncy;
				if (cy >= ey)
					break;
				len = (r + ady) / adx;
				r = (r + ady) / adx;
				ncy = cy + len;
				if (ncy > ey)
				{
					ncy = ey;
					len = ncy - cy;
				}
				cx++;
				pixdst++;
			}
		}
		else
		{
			if (dpos->x < ex)
				bx = dpos->x;
			ady = -ady;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy - len;
			if (ncy < by)
			{
				ncy = by;
				len = cy - ncy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				for (i = 0; i < len; i++, pixdst -= dstrow)
					*pixdst = color;
				cy = ncy;
				if (cy <= by)
					break;
				len = (r + ady) / adx;
				r = (r + ady) / adx;
				ncy = cy - len;
				if (ncy > by)
				{
					ncy = by;
					len = cy - ncy;
				}
				cx++;
				pixdst++;
			}
		}
	}
	return 0;
}
