
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! this is an image management module

/*!\file agl_image.c
 * \ingroup video
 * internals.
 * \author Stephane APIOU
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "agl_private.h"

agl_image *
agl_imgnew(int width, int height)
{
	agl_image *image;

	image = (agl_image *) agl_malloc(sizeof(agl_image));
	if (image == NULL)
	{
		agl_memerr();
		return NULL;
	}

	image->pix = agl_pixnew(width, height);
	if (image->pix == NULL)
	{
		agl_memerr();
		agl_free(image);
		return NULL;
	}
	image->color_key = 0XFFFFFFFF;	/* no color key */
	image->image_name = NULL;
	image->alpha_name = NULL;
	return (image);
}

void
agl_imgfree(agl_image * image)
{
	agl_rassert(image != NULL);

	agl_pixfree(image->pix);

	if (image->image_name)
		agl_free(image->image_name);

	if (image->alpha_name)
		agl_free(image->alpha_name);

	agl_free(image);
}

agl_image *
agl_imgdup(agl_image * image)
{
	agl_image *cpy;


	agl_rvassert(image != NULL, NULL);

	cpy = (agl_image *) agl_malloc(sizeof(agl_image));
	if (cpy == NULL)
	{
		agl_memerr();
		return NULL;
	}

	cpy->pix = agl_pixdup(image->pix);
	if (cpy->pix == NULL)
	{
		agl_memerr();
		agl_free(cpy);
		return NULL;
	}
	cpy->color_key = image->color_key;

	if (image->image_name)
	{
		cpy->image_name = (char *) malloc(strlen(image->image_name) + 1);
		if (cpy->image_name == NULL)
		{
			agl_memerr();
			agl_pixfree(cpy->pix);
			agl_free(cpy);
		}
		strcpy(cpy->image_name, image->image_name);
	}
	else
		cpy->image_name = NULL;
	if (image->alpha_name)
	{
		cpy->alpha_name = (char *) malloc(strlen(image->alpha_name) + 1);
		if (cpy->alpha_name == NULL)
		{
			agl_memerr();
			agl_pixfree(cpy->pix);
			agl_free(cpy->image_name);
			agl_free(cpy);
		}
		strcpy(cpy->alpha_name, image->alpha_name);
	}
	else
		cpy->alpha_name = NULL;

	return cpy;
}
