
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#include <stdarg.h>
#include "config.h"
#include "agl_private.h"

int
agl_putc(agl_pix * pix, agl_font * font, agl_pos * pos, unsigned char c)
{
	agl_glyph *glyph;
	agl_pos npos;

	agl_rvassert(font != NULL, -1);
	agl_rvassert(pix != NULL, -1);
	glyph = (font->glyphs) + c;
	if (!glyph->created)
		if (font->pfio->load(font, c) < 0)
		{
			agl_error("unable to find character in character set");
			return -1;
		}
	if (pos != NULL)
	{
		npos.x = pos->x + glyph->bbox.x;
		npos.y = pos->y - glyph->bbox.y - glyph->yoffset;
	}
	else
	{
		npos.x = glyph->bbox.x;
		npos.y = -glyph->bbox.y - glyph->yoffset;
	}
	if (font->pfio->blit(pix, font, c, &npos) < 0)
	{
		agl_error("unable to draw character");
		return -1;
	}
	return 0;
}

int
agl_puts(agl_pix * pix, agl_font * font, agl_pos * pos, unsigned char *str)
{
	unsigned char *ptr;
	agl_pos prt_pos;
	agl_pos npos;
	agl_glyph *glyph;
	int count;

	agl_rvassert(pix != NULL, -1);
	agl_rvassert(font != NULL, -1);
	agl_rvassert(str != NULL, -1);

	if (pos != NULL)
	{
		prt_pos.x = pos->x;
		prt_pos.y = pos->y;
	}
	else
	{
		prt_pos.x = 0;
		prt_pos.y = 0;
	}
	for (ptr = str, count = 0; *ptr; ptr++, count++)
	{
		switch (*ptr)
		{
		case '\n':
			prt_pos.x = pos->x;
			prt_pos.y += font->height;
			break;
		case '\r':
		case '\t':
			break;
		default:
			glyph = (font->glyphs) + *ptr;
			if (!glyph->created)
				if (font->pfio->load(font, *ptr) < 0)
				{
					agl_error("unable to find character in character set");
					return -1;
				}
			npos.x = prt_pos.x + glyph->bbox.x;
			npos.y = prt_pos.y - glyph->bbox.y - glyph->yoffset;
			if (font->pfio->blit(pix, font, *ptr, &npos) < 0)
			{
				agl_error("unable to draw character");
				return -1;
			}
			prt_pos.x += glyph->advance;
		}
	}
	return count;

}

int
agl_printf(agl_pix * pix, agl_font * font, agl_pos * pos, char *fmt, ...)
{
	unsigned char *str;
	va_list ap;
	int count;

	agl_rvassert(pix != NULL, -1);
	agl_rvassert(font != NULL, -1);
	agl_rvassert(fmt != NULL, -1);
	va_start(ap, fmt);
	str = NULL;
	count = vsnprintf(str, 0, fmt, ap);
	agl_rvassert(count >= 0, -1);
	str = agl_malloc((count + 1) * sizeof(char));
	if (str == NULL)
	{
		agl_memerr();
		return -1;
	}

	if (vsnprintf(str, count + 1, fmt, ap) != count)
	{
		agl_error("string troncated");
		agl_free(str);
		return -1;
	}

	va_end(ap);
	if ((count = agl_puts(pix, font, pos, str)) < 0)
	{
		agl_error("printing failed");
		agl_free(str);
		return -1;
	}
	agl_free(str);
	return count;
}
