
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "config.h"
#include "agl.h"

#define	ARG_NONE	0
#define	ARG_INT		1
#define	ARG_DBL		2
#define	ARG_STR		3
#define ARG_TYPE	0X03
#define ARG_MAND	0X08
#define ARG_REMAIND	0X10

struct option
{
	char *long_opt;
	char short_opt;
	char *descr;
	int flags;
};

struct application
{
	char *descr;
	struct option *opt_tab;
	int idx;
	unsigned long done;
	long ival;
	double dbl;
	char *str;
};

static struct option options[] = {
	{"help", 'h', "get help and exit", ARG_NONE},
	{"version", 'V', "output version informations and exit", ARG_NONE},
	{"type", 't', "number of test to pass", ARG_INT},
	{NULL, 0, "image", ARG_STR | ARG_MAND | ARG_REMAIND},
	{NULL, 1, "font", ARG_STR | ARG_MAND | ARG_REMAIND},
	{NULL, 0, NULL, 0}
};

static struct application appli = {
	"test image loading for adgali",
	options,
	0, 0, 0, 0.0, NULL
};


static void
usage(char *name, struct application *palli, int status)
{
	struct option *popt;

	printf("\n%s -  %s\n", name, palli->descr);
	printf("\nUsage: %s [OPTIONS] ... ", name);
	for (popt = palli->opt_tab; popt->descr != NULL; popt++)
		if ((popt->flags & ARG_REMAIND) != 0)
		{
			if ((popt->flags & ARG_MAND) == 0)
				printf("[ %s ] ", popt->descr);
			else
				printf("%s ", popt->descr);
		}
	printf("\n\nOptions:\n");
	for (popt = palli->opt_tab; popt->descr != NULL; popt++)
		if ((popt->flags & ARG_REMAIND) == 0)
		{
			if (popt->short_opt != 0)
				printf("-%c\t", popt->short_opt);
			else
				printf("\t");
			if (popt->long_opt != NULL)
				printf("--%s\t", popt->long_opt);
			else
				printf("\t");
			printf("%s\n", popt->descr);
		}
	printf("\n");
	exit(status);
}

int
get_opts(struct application *app, int argc, char **argv)
{
	struct option *popt;

	unsigned long mand_nb;
	char *ptr;
	int i;
	int ret;


	if (app->idx >= (argc - 1))
	{
		mand_nb = 0;
		for (i = 0, popt = app->opt_tab; popt->descr != NULL; i++, popt++)
			if ((popt->flags & ARG_MAND) != 0)
				mand_nb |= 1 << i;
		if ((mand_nb & app->done) != mand_nb)
		{
			printf("Mandadory parameters not given\n");
			usage(argv[0], app, 1);
			return -1;
		}
		else
			return 0;
	}
	app->idx++;
	ret = -1;
	if (argv[app->idx][0] == '-')
	{
		if (argv[app->idx][1] == '-')
		{
			for (i = 0, popt = app->opt_tab; popt->descr != NULL; i++, popt++)
				if ((popt->long_opt != NULL) && (strcmp(argv[app->idx] + 2, popt->long_opt) == 0))
					break;
			if (popt->descr == NULL)
			{
				printf("Unknown long option: %s\n", argv[app->idx]);
				usage(argv[0], app, 1);
				return -1;
			}
		}
		else
		{
			for (i = 0, popt = app->opt_tab; popt->descr != NULL; i++, popt++)
				if (argv[app->idx][1] == popt->short_opt)
					break;
			if (popt->descr == NULL)
			{
				printf("Unknown short option: %s\n", argv[app->idx]);
				usage(argv[0], app, 1);
				return -1;
			}
			if (argv[app->idx][2] != '\0')
			{
				printf("not a legal short option: %s\n", argv[app->idx]);
				usage(argv[0], app, 1);
				return -1;
			}
		}
		if ((popt->flags & ARG_TYPE) != ARG_NONE)
		{
			if (app->idx >= (argc - 1))
			{
				printf("No argument given for option: %s\n", argv[app->idx]);
				usage(argv[0], app, 1);
				return -1;
			}
			app->idx++;
		}
		ret = popt->short_opt;
	}
	else
	{
		for (i = 0, popt = app->opt_tab; popt->descr != NULL; i++, popt++)
			if (((popt->flags & (ARG_REMAIND | ARG_MAND)) == (ARG_REMAIND | ARG_MAND)) && ((app->done & (1 << i)) == 0))
				break;
		if (popt->descr == NULL)
		{
			for (i = 0, popt = app->opt_tab; popt->descr != NULL; i++, popt++)
				if (((popt->flags & (ARG_REMAIND | ARG_MAND)) == ARG_REMAIND) && ((app->done & (1 << i)) == 0))
					break;
		}
		if (popt->descr == NULL)
		{
			printf("too many arguments: %s\n", argv[app->idx]);
			usage(argv[0], app, 1);
			return -1;
		}
		ret = 256 + popt->short_opt;

	}
	switch (popt->flags & ARG_TYPE)
	{
	case ARG_NONE:
		break;
	case ARG_INT:
		app->ival = strtol(argv[app->idx], &ptr, 0);
		if (ptr == argv[app->idx])
		{
			printf("argument %s is not a valid integer\n", argv[app->idx]);
			usage(argv[0], app, 1);
			return -1;
		}
		break;
	case ARG_DBL:
		app->dbl = strtod(argv[app->idx], &ptr);
		if (ptr == argv[app->idx])
		{
			printf("argument %s is not a valid double\n", argv[app->idx]);
			usage(argv[0], app, 1);
			return -1;
		}
		break;
	case ARG_STR:
		app->str = argv[app->idx];
		break;
	default:
		printf("Internal error: unknown argument type :%x\n", popt->flags);
		return -1;
	}
	app->done |= 1 << i;
	return ret;
}

#define benchbeg	{ struct timeval otv, ntv;  int delta; gettimeofday(&otv, NULL)
#define benchend	gettimeofday(&ntv, NULL); delta = (ntv.tv_sec - otv.tv_sec) * 1000000 + ntv.tv_usec - otv.tv_usec; printf("bench time us: %d\n", delta); }

int
main(int argc, char **argv)
{
	int par, test;
	char *file, *fontstr;
	agl_handle handle;
	agl_image *pimg;
	agl_font *pfnt;
	agl_ios *pios;
	agl_pos pos;
	int *ph, *pw;

	file = NULL;
	test = 0;
	while ((par = get_opts(&appli, argc, argv)) > 0)
	{
		if (par < 256)
		{
			switch (par)
			{
			case 'h':
				usage(argv[0], &appli, 0);
				break;
			case 't':
				test = appli.ival;
				break;
			case 'V':
				printf("%s: Version: %s\n", argv[0], VERSION);
				exit(0);
			}
		}
		else
		{		/* remaind parameters */
			switch (par & 0XFF)
			{
			case 0:
				file = appli.str;
				break;
			case 1:
				fontstr = appli.str;
				break;
			}
		}
	}
	if (par < 0)
	{
		exit(1);
	}



	if (agl_init(&handle) < 0)
	{
		printf("error:\n%s", agl_geterror());
		goto end;
	}
	if (agl_sublayergetvidsz(handle, 0, &ph, &pw) < 0)
	{
		printf("error:\n%s", agl_geterror());
		goto end;
	}
	if ((ph != 0) && (pw != 0))
	{
		while (*ph != 0)
			printf("h:%d w:%d\n", *ph++, *pw++);
	}
	if (agl_sublayersetvid(handle, 256, 256, 0) < 0)
	{
		printf("error:\n%s", agl_geterror());
		goto end;
	}

	if ((pios = agl_ios_open(handle, file)) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		goto end;
	}
	if ((pimg = agl_imgioload(pios)) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		goto end;
	}
	agl_ios_close(pios);
	if ((pios = agl_ios_open(handle, fontstr)) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		goto end;
	}
	if ((pfnt = agl_fontioload(pios, 50)) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		goto end;
	}

	pos.x = 25;
	pos.y = 50;

	if (agl_printf(pimg->pix, pfnt, &pos, "Adgali\nHello boys") < 0)
	{
		fprintf(stderr, "(4)%s\n", agl_geterror());
		goto end;
	}

	pos.x = 0;
	pos.y = 0;
	if (agl_sublayerblit(handle, pimg->pix, NULL, NULL, &pos) < 0)
	{
		fprintf(stderr, "(1)%s\n", agl_geterror());
	}
	getchar();
	agl_imgfree(pimg);
	agl_fontfree(pfnt);
	agl_ios_close(pios);
      end:
	if (agl_exit(handle) < 0)
	{
		printf("error:\n%s", agl_geterror());
	}
	return 0;
}
