
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include "../config.h"
#include "agl.h"



static void usage(int status);

/* The name the program was run with, stripped of any leading path. */
char *program_name;

int
main(int argc, char **argv)
{
	int h, w, x, y, adv, upper, lower, base, next;
	int state, begin, val;
	agl_image *pimg;
	agl_ios *pios;
	agl_handle handle;
	unsigned long *p, color;
	FILE *textfile, *strfile;

	program_name = argv[0];

	if (argc <= 3)
		usage(1);

	if (agl_init(&handle) < 0)
	{
		printf("error:\n%s", agl_geterror());
		exit(1);
	}

	if ((pios = agl_ios_open(handle, argv[1])) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		exit(1);
	}
	if ((pimg = agl_imgioload(pios)) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		exit(1);
	}
	agl_ios_close(pios);

	if ((textfile = fopen(argv[3], "w")) == NULL)
	{
		printf("unable to open output file: %s\n", argv[2]);
		exit(1);
	}
	if (argv[2][0] == '-')
		strfile = stdin;
	else if ((strfile = fopen(argv[2], "r")) == NULL)
	{
		printf("unable to character file: %s\n", argv[2]);
		exit(1);
	}
	h = pimg->pix->size.height;
	w = pimg->pix->size.width;
	adv = pimg->pix->row_size;
	state = 0;
	color = base = lower = upper = 0;
	for (y = 0, p = pimg->pix->pixels; y < h; y++, p += adv)
	{
		switch (state)
		{
		case 0:
			color = *p;
			state++;
			break;
		case 1:
			if (*p == color)
			{
				upper = y;
				state++;
			}
			break;
		case 2:
			if (*p == color)
			{
				base = y;
				state++;
			}
			break;
		case 3:
			if (*p == color)
			{
				lower = y;
				state++;
			}
			break;
		case 4:
			if (*p == color)
			{
				next = y;
				state++;
			}
			break;
		default:
			break;
		}
	}
	if (state <= 4)
	{
		printf("verticals lines: upper line, baseline, lower line are not found\naborting...\n");
		exit(1);
	}
	fprintf(textfile, "BFONT\n# this file describes a bitmap font\n");
	fprintf(textfile, "\"%s\"\n%d,%d,%d,%d # upper line, lower line, base line, next line\n", argv[1], upper, lower, base, next);
	begin = 0;
	for (x = 1, p = pimg->pix->pixels + 1; x < w; x++, p++)
	{
		if (*p == color)
		{
			if (begin)
			{
				fprintf(textfile, "%d #space between characters\n\n", x - 1 - begin);
				begin = x;
				break;
			}
			else
				begin = x;
		}
	}
	if (x >= w)
	{
		printf("Space between characters not found\naborting...\n");
		exit(1);
	}
	fprintf(textfile, "# character hexa, X start, X stop\n");
	for (x++, p++; x < w; x++, p++)
	{
		if (*p == color)
		{
			val = fgetc(strfile);
			if (val == EOF)
				break;
			if (isgraph(val))
				fprintf(textfile, "%x,%d,%d #%c\n", val, begin, x - 1, val);
			else
				fprintf(textfile, "%x,%d,%d\n", val, begin, x - 1);
			begin = x;
		}
	}
	fclose(strfile);
	fclose(textfile);
	exit(0);
}

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  */

static void
usage(int status)
{
	printf("%s -  create a description file for a bitmap font\n", program_name);
	printf("Usage: %s bitmap_file string_file output_file\n", program_name);
	exit(status);
}
