
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "../config.h"
#include "agl.h"



static void usage(int status);

/* The name the program was run with, stripped of any leading path. */
char *program_name;

int
main(int argc, char **argv)
{
	int h, w, x, y, yc, xo, yo, wi, he;
	agl_image *image;
	unsigned long *pl, *p, *po, color, adv;
	agl_handle handle;
	FILE *textfile;
	agl_ios *pios;

	program_name = argv[0];

	if (argc <= 2)
		usage(1);

	if (agl_init(&handle) < 0)
	{
		printf("error:\n%s", agl_geterror());
		exit(1);
	}
	if ((pios = agl_ios_open(handle, argv[1])) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		exit(1);
	}
	if ((image = agl_imgioload(pios)) == NULL)
	{
		printf("error:\n%s", agl_geterror());
		exit(1);
	}
	agl_ios_close(pios);

	if ((textfile = fopen(argv[2], "w")) == NULL)
	{
		printf("unable to open output file: %s\n", argv[2]);
		exit(1);
	}
	fprintf(textfile, "IFILE\n# this file describes images zones in a bitmap\n");
	fprintf(textfile, "# X   Y  W   H\n");
	h = image->pix->size.height;
	w = image->pix->size.width;
	adv = image->pix->row_size;
	color = *image->pix->pixels & AGL_RGBA2LONG(0XFFFFFF00);
	for (y = 0, pl = p = image->pix->pixels; y < h; y++, pl += adv)
	{
		for (x = 0, p = pl; x < w;)
		{
			if (((*p & AGL_RGBA2LONG(0XFFFFFF00)) == color)
			    && ((x < (w - 1)) && ((p[1] & AGL_RGBA2LONG(0XFFFFFF00)) == color)) && ((y < (h - 1))
												    &&
												    ((p[adv] &
												      AGL_RGBA2LONG(0XFFFFFF00)) ==
												     color))
			    && ((p[adv + 1] & AGL_RGBA2LONG(0XFFFFFF00)) != color))
			{
				xo = x;
				yo = y;
				po = p;
				for (wi = 0, x++, p++; x < w; wi++, x++, p++)
				{
					if ((*p & AGL_RGBA2LONG(0XFFFFFF00)) != color)
						goto next;
					if ((p[adv] & AGL_RGBA2LONG(0XFFFFFF00)) == color)
						break;
				}
				if (wi == 0)
					goto next;
				for (he = 0, yc = y + 1, po += adv; yc < h; yc++, he++, po += adv)
				{
					if ((*po & AGL_RGBA2LONG(0XFFFFFF00)) != color)
						goto next;
					if ((po[1] & AGL_RGBA2LONG(0XFFFFFF00)) == color)
						break;
				}
				if (he == 0)
					goto next;
				fprintf(textfile, "%d %d %d %d\n", xo + 1, yo + 1, wi, he);
			}
			else
			{
				p++;
				x++;
			}
		      next:
		}
	}
	fclose(textfile);
	exit(0);
}

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  */

static void
usage(int status)
{
	printf("%s -  create a description file for a bitmap collection\n", program_name);
	printf("Usage: %s bitmap_file output_file\n", program_name);
	exit(status);
}
